/*************************************************************************
  program name: lm_auto.c
  
  Program to do posterior IBD probs 
  The state (X) consists of segregation indicators -- now generalized
     so that up to 8 gametes can be scored for IBD,
	
  X determines the IBD pattern among the specified descendantgametes.

  The data, Y, is the multilocus genotypes of observed individuals
     in the pedigree.

  Modified from autozyg.c, April 1997,  so that sampling can be done on a
  meiosis by meiosis basis. (EAT)
  Modified to include general penetrance routine; August 1997 (SCH/EAT).
  Modified to score the Iceland states, October 1997 (EAT)
 
  Many improvements made May-July 1998, for GAW 11 analyses (EAT)
     Now scores multiple gamates (up to 8) at multiple loci (up to 24)
           on several component pedigrees (up to 5)
     Check the header files for size limits.
  Until dynamic allocation is implemented, beware of exceeding local
     stacksize default limits.

  Modified to allow use of the L-sampler, July 1999 (SCH)
  Modified to incorporate sex-specific maps, 
      to use L-sampler and M-sampler new library routines,  March 2000 (EAT)

  Converted to MORGAN_V2.3                   MAJ June 2000
**************************************************************************/

#define PNAM  "lm_auto"

#include "autozyg.h"

#undef CHECK

    void chkprnt1 (void);
    void chkprnt2a (int);
    void chkprnt2b (int);
    void chkprnt5 (void);
    void chkprnt6 (int);
    void chkprnt7 (int, int, int, int, int);
    void chkprnt8 (int, int, int);
    void chkprnt9 (void);

    int   nloc;
    int   flag_t;

    int ***y;				/* full genotypic array  */

    double prob[NC][NLOC];
    int    gaga[G6];


int main (int argc, char *argv[])
{
    int   i, j, k, l;
    int **dodo = NULL;
    int   t, c, h, nh, pt, err;
    int   jj;
    int   cfound, found[5];
    int   ngenes = 0, nfou;
    int   nsegs, nsw, sx;
    int  *ind_sw = 0, *sex_sw = 0;	/*  switch arrays        */
    int   posp;
    int   recounts[NC][NLOC][10];
    int   npt[NC];			/* numbers of patterns */
    int   index[NC][CD6], label[NC][PT6];

    double **quant = 0;
    double sss[NLOC], ss, avcou;
                                        /* sums of counts and scores */
    double tot_log_prob_seg[NC][NH], tot_log_prob_pen[NC][NH];


    double seg_pen2 (), full_seg_prob ();
    void  fwd_compute (), back_smpl (), set_lsmpl (), do_msmpl ();
    void  score_genes ();
    int   do_lsmpl ();

    FILE *pedF, *oScorF, *oSeedF;


    atexit (print_tail);

    print_head (PNAM);

    if (!proc_pars (argc, argv, PNAM))	/* process input parameter      */
	exit (EXIT_FAILURE);		/* statements                   */

    if (!proc_auto ())			/* check statements for         */
	exit (EXIT_FAILURE);		/* completeness and consistency */
                                        /* and prepare controls for     */
                                        /* lm_auto                      */

    if (!proc_seeds ())                 /* set up seed                  */
	exit (EXIT_FAILURE);		/* parameters                   */


    if (!proc_fils ())			/* set up pedigree file         */
	exit (EXIT_FAILURE);		/* parameters                   */

    setSeeds (SAMPL);  			/* initializes seeds for rans() */

    if (!chk_pedfil())                  /* checks pedigee file for      */
        give_up(ERR_qit);               /* errors                       */

    if (!mak_opedfil())                 /* writes new pedigree file,    */
        give_up(ERR_qit);               /* if changes                   */

    if (!(pedF = open_ped()))           /* reopens pedigree file        */
        give_up(ERR_qit);         

    alloc_ped ();
    input_ped (pedF, 0, dodo, quant);   /*input pedigree and trait data */

    nloc = NLoci - 1;

    nfou = count_founders ();
    printf (" Number of founders:  %d\n\n", nfou);
    printf (" Number of components:  %d\n\n", n_compnts);

    qdex = malloc((pedsize + 1) * sizeof(int));
    alloc_chk(WHERE, qdex);
    for (i=0; i<=pedsize; ++i) qdex[i]=0;
    trtphen = malloc((pedsize + 1) * sizeof(int));
    alloc_chk(WHERE, qdex);
    for (i=0; i<=pedsize; ++i) trtphen[i]=0;

/*  now allocate space for genotypes y for our values of pedsize, NMarks */

    alloc_y(&y, nloc, pedsize);


/*  now set marker data on observed individuals 	*/
/*  marker loci are labelled 1,..., nloc         	*/

    set_markdat(y, nloc, n_compnts);


/*  set data for the trait                              */

    set_traitdat(y, qdex, trtphen, pedsize, n_compnts);

    allocate (pedsize, nloc);

/*  set up initial segs config */
/*  ** removed **              */


/*  now set and index switches  */

    ind_sw = calloc (2 * pedsize, sizeof (int));
    alloc_chk (WHERE, ind_sw);
    sex_sw = calloc (2 * pedsize, sizeof (int));
    alloc_chk (WHERE, sex_sw);
    nsw = list_switches (nloc, ind_sw, sex_sw);
    printf (" Number of meiosis switches to sample is %d\n\n", nsw);
    fflush (stdout);

/*  set up probands for gene scoring */

    set_probgam(n_compnts);


/*  set up recombination patterns and initial lods */

    nh = 0;
                                                /* L-sampler set-up */
    set_lsmpl (Rs, nloc, NAll, Qq, y);    	/* EAT  March 2000  */

/*  set genes from initial switches */
    for (j = 0; j <= nloc; ++j)
    {
	ngenes = set_genes (j);

        chkprnt2a (j);

	for (c = 0; c < n_compnts; ++c)
	    prob[c][j] = seg_pen2 (c, y, Qq[j], j, ngenes, &err);

        chkprnt2b (j);
    }


/*  probability now computed separately by component  */

    for (c = 0; c < n_compnts; ++c)
    {
	for (h = 0; h <= nh; ++h)
	{
	    tot_log_prob_seg[c][h] = 0.0;
	    tot_log_prob_pen[c][h] = 0.0;
	}
	for (j = 0; j <= nloc; ++j)
	{
	    count[c][j] = 0;
            for (jj = 0; jj < 10; ++jj)
         	recounts[c][j][jj] = 0;
	    for (pt = 0; pt < PT6; ++pt)
		score[c][j][pt] = 0.0;
	}
    }

    chkprnt5 ();

/*  count total number of patterns */

    for (c = 0; c < n_compnts; ++c)
    {
	/* for 3-sibs + grandma problem npt=4,                       */
	/* for 3-sibs problem npt=4, in scoring 6 genes              */
        /* npt[c] = index_ibd(NPr[c], 2, 0, index[c], label[c], 
                                                      list[c], pedF);*/
        /* for JV pedigree npt =15; 6 genes in 4 groups              */
        /* npt[c] = index_ibd(NPr[c], 1, 4, index[c], label[c], 
                                                      list[c], pedF);*/
        /* for JV pedigree homozygosity of 1 individual              */
        /* for JV pedigree npt =15; specified as 4 gametes           */
        /* npt[c] = index_ibd(NPr[c], 0, 0, index[c], label[c], 
                                                      list[c], pedF);*/

        npt[c] = index_ibd (NPr[c], 0, 0, index[c], label[c], 
                                                    list[c], pedF);

	/* for iceland; 6 gametes with no reduction, but
	   with only partial list of states scored                   */
	/* use same indexing for iceland, and icesim data            */
        /* npt[c] = index_ibd(NPr[c], -2, 0, index[c], label[c], 
                                             list[c], pedF);         */
         
	printf (" Number of pattern sets to be scored is %d on "
                      "component %d\n\n", npt[c], c + 1);
    }


/*  close data file and open output file */

    fclose (pedF);

    if (!(oScorF = open_fil(OSCORE)))
        give_up(ERR_qit);


/*  now test switches and do MC sampling */

    flag_t = -1;

    for (t = 0; t < NIt; ++t)
    {
        chkprnt6 (t);

	if (rans () < LSamplProb)
	{					/*** Do L-sample step ***/
            if (!((t + 1) % WIt))   
	    {
	        printf (" Starting L-sampler iteration %d\n", t + 1);
	        fflush (stdout);
	    }
	    jj = do_lsmpl (nloc, 0, 0);
	    (void) set_genes (jj);

	    for (c = 0; c < n_compnts; ++c)
	    {
		prob[c][jj] = seg_pen2 (c, y, Qq[jj], jj, ngenes, &err);
		j = revchro[0][jj];

		/* scoring for complete gamete set  */
		score_genes (j, jj, c, NPr[c], gaga, index[c],
			     Probands[c], Proband_gamete[c]);
        	/* cumulate total log-prob on all compts */
                tot_log_prob_seg[c][0] += 
                        full_seg_prob (nloc, c + 1, 0, recounts[c]) / nloc;
                for (j = 0; j <= nloc; ++j)
                    tot_log_prob_pen[c][0] += log (prob[c][j]) / (nloc + 1);
                for (h = 1; h <= nh; ++h)
                tot_log_prob_seg[c][h] += 
                        full_seg_prob (nloc, c + 1, h, recounts) / nloc;
	    }
	}
	else
	{					/*** Do M-sample step ***/
            if (!((t + 1) % WIt))   
	    {
	        printf (" Starting M-sampler iteration %d\n", t + 1);
	        fflush (stdout);
	    }
	    					/*  choose a random meiosis  */
	    l = intran (0, nsw - 1) + 1;
	    i = ind_sw[l];
	    sx = sex_sw[l];
	    c = nghd[i].component - 1;

            chkprnt7 (t, l, i, sx, c);

	    /* cumulate total log-prob on compt c */
            tot_log_prob_seg[c][0] += 
                               full_seg_prob (nloc, c + 1, 0, recounts[c]);
	    for (j = 0; j <= nloc; ++j)
	        tot_log_prob_pen[c][0] += log (prob[c][j]);
	    for (h = 1; h <= nh; ++h)
	        tot_log_prob_seg[c][h] += 
                               full_seg_prob (nloc, c + 1, h, recounts);
 
	    do_msmpl (nloc, c, i, sx, prob, y, ngenes, flag_t, t);

	    for (j = nloc; j >= 0; --j)
	    {
		jj = chromo[0][j];	/* scoring for full gamete set */
		score_genes (j, jj, c, NPr[c], gaga, index[c],
			     Probands[c], Proband_gamete[c]);

                chkprnt8 (c, t, jj);
	    }
	}
        if (!((t + 1) % FWIt))   
	    for (c = 0; c < n_compnts; ++c)
	    {
		for (j = 0; j <= nloc; ++j)
		{
		    jj = chromo[0][j];
                  
		    fprintf (oScorF, "\n %d  %d  ", jj, count[c][jj]);
		    for (pt = 0; pt < npt[c]; ++pt)
			fprintf (oScorF, " %5.1f", score[c][jj][list[c][pt]]);
		        /*fprintf (oScorF, " %e", score[c][jj][list[c][pt]]);*/
		    fflush (oScorF);
		}
	    }
    }
    for (c = 0; c < n_compnts; ++c)
    {
	avcou = 0.0;
	for (j = 0; j <= nloc; ++j)
	    avcou += count[c][j];
	avcou = avcou / (nloc + 1);

        printf ("\n\n Output summary for component %d:\n\n", c + 1);
	printf ("    Average count over loci  = %7.3f\n", avcou);
	printf ("    Average total log-probability of segregations = %7.3f\n",
	        	tot_log_prob_seg[c][0] / avcou);
	printf ("    Average total log-probability of penetrances  = %7.3f\n\n",
	        	tot_log_prob_pen[c][0] / avcou);

        /* print out recombination counts  */
        printf ("    Recombination counts:\n" );
        printf ("       map interval   M:  total  recomb non-rec    "
                                      "F:  total  recomb non-rec\n");
        for (j = 0; j < nloc; ++j)
        {
            printf ("              %d        ", j + 1);
            for (i = 0; i < 6; ++i)
            {
                if (i == 3)
                    printf ("     ");
                printf ("%8d", recounts[c][j][i]);
            }
            printf ("\n");
        }
        printf ("\n");
        fflush (stdout);
 
        printf("    Probabilities of ibd patterns:\n");
	for (k = 0; k < npt[c]; ++k)
	{
	    ss = 0.0;
	    pt = list[c][k];
	    for (j = 0; j <= nloc; ++j)
	    {
		sss[j] = score[c][j][pt] / count[c][j];
		ss += sss[j];
	    }
	    if (ss > 0.001)
	    {
		invert_label (label[c][k], NPr[c], gaga);
                printf("      ");
		for (i = 0; i < NPr[c]; ++i)
		    printf (" %d", gaga[i]);
		printf (" : ");
		for (j = 0; j <= nloc; ++j)
		    printf ("%7.4f", sss[j]);
		printf ("  (label %2d)\n", label[c][k]);
		fflush (stdout);
	    }
	}
    }
    printf("\n");

/*  now we are done  */

/*  Free up allocated memory */

    free_allocate ();
    free_gen_dat ();
    for (j = 0; j <= nloc; j++)
    {
	free (y[0][j]);
	free (y[1][j]);
    }
    free (y[0]);
    free (y[1]);
    free (y);
    if (ind_sw)
	free (ind_sw);
    if (sex_sw)
	free (sex_sw);
    free_allocate_loci ();
   
    chkprnt9 ();

    if (SamplSeedsOut)
    {
        getSeeds(SAMPL);
        if (!(oSeedF = open_fil(OSEED)))
            give_up(ERR_qit);
        writSeeds(oSeedF);
        fclose(oSeedF);
    }

    exit (EXIT_SUCCESS);
}

/*********************************************************************/

void chkprnt1 (void)
{
#ifdef CHECK
    int i, j;
/*  data check  */
    for (i = 0; i < pedsize; ++i)
    {
	printf (" ind = %d", nghd[i].self);
	for (j = 0; j <= nloc; ++j)
	    printf ("    %d %d", y[0][j][i], y[1][j][i]);
        printf("\n");
    }
    printf("\n");  fflush (stdout); 
#endif /*CHECK*/
    return;
}
void chkprnt2a (int j)
{
#ifdef CHECK
    printf (" Set genes at locus %d\n", j);  fflush (stdout);
    print_obs_genes(j, y);   fflush(stdout);          
    printf (" Scored prob at locus %d; giving", j);  
    fflush (stdout);        
#endif /*CHECK*/
    return;
}
void chkprnt2b (int j)
{
#ifdef CHECK
    int c;
    for (c = 0; c < n_compnts; ++c)
        printf (" %e", prob[c][j]);   
    printf("\n\n");
    fflush (stdout); 				    
#endif /*CHECK*/
    return;
}
void chkprnt5 (void)
{
#ifdef CHECK
    int j;

    printf (" Now done the data check\n");  
    fflush (stdout);
    for (j = 0; j <= nloc; ++j)
    {
	printf ("\n Locus %d set-up \n", j);
	print_set_up (j);
	printf ("\n");  fflush (stdout);
    }                           
#endif /*CHECK*/
    return;
}
void chkprnt6 (int t)
{
#ifdef CHECK
    int c, j;

    for (c = 0; c < n_compnts; ++c)
        for (j = 0; j <= nloc; ++j)
	    if (prob[c][j] <= 0.0)
	    {
	        printf ("\n PROBLEM iteration %d", t + 1);
	        printf (" zero probability on compnt %d at locus %d\n",
		        c + 1, j);
	        exit (92);
	    }       
#endif /*CHECK*/
    return;
}
void chkprnt7 (int t, int l, int i, int sx, int c)
{
#ifdef CHECK
    if (t == flag_t)
    {
	printf ("\n sw %d, ind %d %d, cmp %d\n\n",
		l, nghd[i].self, sx, c);
	fflush (stdout);
    }	
#endif /*CHECK*/
    return;
}
void chkprnt8 (int c, int t, int jj)
{
#ifdef CHECK
    int k;

    if (prob[c][jj] <= 0.0)
    {
        printf ("\n ZERO PROBABILITY PROBLEM\n");
        printf (" zero prob at iter %d locus %d\n", t + 1, jj);
        printf (" chk gaga");
        for (k = 0; k < NPr[0]; ++k)
	    printf (" %d", gaga[k]);
	printf ("\n\n");  fflush (stdout);
        exit (93);
    }
#endif /*CHECK*/
    return;
}
void chkprnt9 (void)
{
#ifdef CHECK
    loki_peel_free();
    printf ("  Now we are done freeing \n\n");
    fflush (stdout); 
#endif /*CHECK*/
    return;
}
