/*************************************************************************
  program name: lm_lods.c
  
  Program to do lod scores (replaces homozyg.c from old Autozyg)
  The state (X) consists of segregation indicators -- so can generalize to
	cases of more than one observed/pedigree, and multiple loci.
  
  The data, Y, is the multilocus genotypes of observed individuals
     in the pedigree.

  Obtained by modification of jnt_auto.c.  December 1998.  (EAT)

Here is the plan!
    OK this is the PLAN  (12/15/98)

OK, now we have a general autozyg program, we need to go back to
lod estimation, for those as want.  We use the same program (modified).

For k markers, we will have (k+2) simulation points, and (k+2) evaluation
points  (but maybe it is not worth evaluating all for all).
To start we will just evaluate relative to a left-point and a right-point
for each simulation point.

The simulation points are mid-point of each interval, and r=0.1 to each
end of chromosome (0.1 can be adjusted later), and "unlinked" which is
0.5 to either end.

Since the data are locus-by-locus, the likelihood ratio contribution is
just ratio of P_theta(S)/P_sim_th(S).

This is very easy, once we get the chromosome set-ups fixed.

1) Trait locus data set-up -- assume trait-locus genotypes for observed
individuals -- done  12/20/98
2) Likelihood-ratio contributions routine,  -- done 12/18/98
3) Simulation and evaluation chromosomes for each location. -- done 12/23/98

4) Need a test run!! -- JV4C maybe.   working! 12/24/98

New routines:  (12/19/98)
       full_seg_prob() and full_pen_seg_lod() and lik_rat()
          (all in full_logprob.c)
           full_seg_prob modified from old routine -- returns log(prob P(S)) 
                of all meioses on given component pedigree
           full_pen_seg_lod computes full penetrance P(Y|S) for data y, and
                  given meiosis indicators, on given component pedigree.
               calls full_lod and returns full log-probability of (Y,S).
        [lod_next routine is not currently used ]

Now to integrate the L-sampler, using the new Lsampler and Msampler
      libraries and new routines;        EAT March 2000

Converted to MORGAN_V2.3                 MAJ June 2000
**************************************************************************/

#define PNAM  "lm_lods"

#include "autozyg.h"

#undef CHECK

    void chkprnt1 (void);
    void chkprnt2a (int); 
    void chkprnt2b (int); 
    void chkprnt5 (void); 
    void chkprnt6 (void); 
    void chkprnt7 (int); 
    void chkprnt8 (void);
    void chkprnt9 (int);
    void chkprnt19 (int, int, int);
    void chkprnt10 (void);
    void chkprnt11 (int, int, int);
    void chkprnt12 (int, int);
    void chkprnt13 (int);
    void chkprnt14 (int, int);
    void chkprnt15 (int, int);
    void chkprnt16 (int);
    void chkprnt17 (void);

    int ***y;				/* full genotypic array 	*/
    int    dummy[NLOC][10];
    double l_ratio[NC][NH];


int   main (int argc, char *argv[])
{
    int   i, j, k, l;
    int **dodo = NULL;
    double **quant = 0;
    int   t, c, h, nh, ix, err, flag_t;
    int   ngenes = 0, nfou, nloc;
    int   nsegs, nsw, sx;

    int  *ind_sw, *sex_sw;		/*  switch arrays  		*/

    int   recounts[NC][NLOC][10];
    double prob[NC][NLOC], avcou, ltemp;
    double cumulik[NC][NH];
    double tot_log_prob_seg[NC][NH], tot_log_prob_pen[NC][NH];
                                         /* likelihoods etc.   	     	*/
    int   jj;
    double rec0[NH];
    int loc0[NH];

    double seg_pen2 (), full_seg_prob (), full_pen_seg_lod (),
          lik_rat ();

    void  fwd_compute (), back_smpl (), set_lsmpl (),
          do_msmpl ();
    void  score_genes (), set_positions ();
    int   do_lsmpl ();

    FILE *pedF, *oScorF, *oSeedF;


    atexit (print_tail);

    print_head (PNAM);

    if (!proc_pars (argc, argv, PNAM))	/* process input parameter      */
	exit (EXIT_FAILURE);		/* statements                   */

    if (!proc_auto ())			/* check statements for         */
	exit (EXIT_FAILURE);		/* completeness and consistency */
                                        /* and prepare controls for     */
                                        /* lm_auto                      */

    if (!proc_seeds ())			/* set up seed                  */
	exit (EXIT_FAILURE);		/* parameters                   */

    if (!proc_fils ())			/* set up pedigree file         */
	exit (EXIT_FAILURE);		/* parameters                   */

    setSeeds (SAMPL);                   /* initializes seeds for rans() */

    if (!chk_pedfil())                  /* checks pedigree file for     */
        give_up(ERR_qit);               /* errors.                      */

    if (!mak_opedfil())                 /* writes new pedigree file     */
        give_up(ERR_qit);               /* if changes                   */

    if (!(pedF = open_ped ()))          /*opens pedigree file           */
        give_up(ERR_qit);

    alloc_ped ();
    input_ped (pedF, 0, dodo, quant);   /*input pedigree and trait data */

    nloc = NLoci - 1;

    nfou = count_founders ();
    printf (" Number of founders:  %d\n\n", nfou);
    printf (" Number of components:  %d\n\n", n_compnts);

/* now allocate space for genotypes y for our values of pedsize, nloc   */

    alloc_y (&y, nloc, pedsize);

    qdex = malloc((pedsize + 1) * sizeof(int));
    alloc_chk(WHERE, qdex);
    for (i=0; i<=pedsize; ++i) qdex[i]=0;
    trtphen = malloc((pedsize + 1) * sizeof(int));
    alloc_chk(WHERE, qdex);
    for (i=0; i<=pedsize; ++i) trtphen[i]=0;

/* now set marker data for observed individuals  */

    set_markdat (y, nloc, n_compnts);

/* now set trait data                            */

    set_traitdat (y, qdex, trtphen, pedsize, n_compnts);

    chkprnt1 ();

    allocate (pedsize, nloc);

/*  set up initial segs config ** removed **   */

/* now set and index switches  */

    ind_sw = calloc (2 * pedsize, sizeof (int));
    alloc_chk (WHERE, ind_sw);
    sex_sw = calloc (2 * pedsize, sizeof (int));
    alloc_chk (WHERE, sex_sw);
    nsw = list_switches (nloc, ind_sw, sex_sw);
    printf (" Number of meiosis switches to sample is %d\n", nsw);
    fflush (stdout);

/* for simplicity we specify trait locus unlinked to start
   NO--THIS CAUSES BUG!!!  EAT April 2000
   so can same set-up as lm_auto program -- but need Marek */

/* L-sampler set-up */
    set_lsmpl (Rs, nloc, NAll, Qq, y);	/* EAT  March 2000	 	*/

    for (j = 0; j <= nloc; ++j)
    {
	ngenes = set_genes (j);

        chkprnt2a (j);

	for (c = 0; c < n_compnts; ++c)
	    prob[c][j] = seg_pen2 (c, y, Qq[j], j, ngenes, &err);

        chkprnt2b (j);
    }


/* now start BIG loop */

    for (k = 0; k < NTPos; ++k)
    {
	printf ("\n\n CHROMOSOME SET %d\n\n", k);
	fflush (stdout);

	nh = 2;       			/* two evaluation sets for each	*/
                                        /* simulation chromo & two sexs */
        printf(" Recombinations:\n");

	for (sx = FEM; sx <= MAL; ++sx)
	{
	    loc0[0] = IntVal[k];
	    rec0[0] = RecVal[sx][k];

	    if (k == 0)
	    {
		loc0[1] = nloc;
                rec0[1] = RecVal[sx][NTPos-1];
	    }
	    else
	    {
		loc0[1] = IntVal[k - 1];
		rec0[1] = RecVal[sx][k - 1];
	    }

	    loc0[2] = IntVal[k + 1];
	    rec0[2] = RecVal[sx][k + 1];

	    set_chromosomes (nloc, Marek[sx], nh, loc0, rec0, 2 - sx);

	    printf ("    Sim,   ");
            if (sx == MAL)
	        printf ("M:");
            if (sx == FEM)
	        printf ("F:");
	    print_chromosome (nloc, 0, 2 - sx);

	    for (i = 1; i <= nh; ++i)
	    {
                if (i == 1)
		    printf ("    Left,  ");
                else
		    printf ("    Right, ");
                if (sx == MAL)
	            printf ("M:");
                if (sx == FEM)
	            printf ("F:");
		print_chromosome (nloc, i, 2 - sx);
	    }
	}
        printf ("\n");

        chkprnt5 ();

/* L-sampler set-up */

	set_positions (nloc);		/* EAT  March 2000 		*/

        chkprnt6 ();

/* now should do an L-sample step on the trait locus to reset  */

	jj = do_lsmpl (nloc, 1, 0);
	(void) set_genes (jj);

	for (c = 0; c < n_compnts; ++c)
	    prob[c][jj] = seg_pen2 (c, y, Qq[jj], jj, ngenes, &err);
	j = revchro[0][jj];

        chkprnt7 (j);

        for (c = 0; c < n_compnts; ++c)
            prob[c][0] = seg_pen2 (c, y , Qq[0], 0, ngenes, &err);

	for (j = 0; j <= nloc; ++j)
	{
	    ngenes = set_genes (j);

            chkprnt2a (j);

	    for (c = 0; c < n_compnts; ++c)
		prob[c][j] = seg_pen2 (c, y, Qq[j], j, ngenes, &err);

            chkprnt2b (j);
	}

        chkprnt8 (); 

	flag_t = -1;

	for (c = 0; c < n_compnts; ++c)
	{
	    for (h = 0; h <= nh; ++h)
	    {
		l_ratio[c][h] = 1.0;
		cumulik[c][h] = 0.0;
		tot_log_prob_seg[c][h] = 0.0;
		tot_log_prob_pen[c][h] = 0.0;
	    }
	    for (j = 0; j <= nloc; ++j)
	    {
		count[c][j] = 0;
		for (jj = 0; jj < 10; ++jj)
		    recounts[c][j][jj] = 0;
	    }
	}

	for (j = 0; j <= nloc; ++j)
	    for (jj = 0; jj < 10; ++jj)
		dummy[j][jj] = 0.0;

        chkprnt9 (k);

        for (c = 0; c < n_compnts; ++c)
            for (h = 1; h <= nh; ++h)
            {
	        l_ratio[c][h] = (full_seg_prob (nloc, c + 1, h, dummy)
		       	       - full_seg_prob (nloc, c + 1, 0, dummy));
                l_ratio[c][h] = exp (l_ratio[c][h]);
	    }

        chkprnt19 (n_compnts, nh, nloc);

/* now do MCMC resampling by meiosis or by locus */

	for (t = 0; t < NIt; ++t)
	{

            chkprnt10 ();

	    if (rans () < LSamplProb)
	    {			       	/****    Do L-sample step    ****/
                if (!((t + 1) % WIt))
		{
		    printf (" Starting L-sampler iteration %d\n", t + 1);
		    fflush (stdout);
		}

		jj = do_lsmpl (nloc, 0, 0);
		(void) set_genes (jj);

		for (c = 0; c < n_compnts; ++c)
		    prob[c][jj] = seg_pen2 (c, y, Qq[jj], jj, ngenes, &err);
		j = revchro[0][jj];

                /* cumulate total log-prob for locus j */
/*commented out*/
/******         for (c = 0; c < n_compnts; ++c)                    ******
                    ++count[c][j];
                for (h = 0; h <= nh; ++h)  
                { 
                    tot_log_prob_seg[c][h] += 
                            full_seg_prob(nloc, c + 1, h, recounts[c]);
                    for (j = 0; j <= nloc; ++j)  
                        tot_log_prob_pen[c][h] += log(prob[c][j]); 
*******         }                                                  ******/

		for (c = 0; c < n_compnts; ++c)
		{
		    ltemp = full_seg_prob (nloc, c + 1, 0, recounts[c]);
		    for (h = 1; h <= nh; ++h)
		    {
			l_ratio[c][h] = 
                          (full_seg_prob (nloc, c + 1, h, dummy) - ltemp);
			l_ratio[c][h] = exp (l_ratio[c][h]);
		    }
		}

	    }
	    else
	    {			       	/****    Do M-sample step    ****/
                if (!((t + 1) % WIt))
		{
		    printf (" Starting M-sampler iteration %d\n", t + 1);
		    fflush (stdout);
		}
		                    	/*choose a random meiosis	*/
		l = intran (0, nsw - 1) + 1;
		i = ind_sw[l];
		sx = sex_sw[l];
		c = nghd[i].component - 1;

                chkprnt11 (l, sx, c);
 
                chkprnt12 (c, nh);

		/* adjust meiosis to be updated out of l_ratio */
            
                chkprnt13 (c);

		for (h = 1; h <= nh; ++h)
		    l_ratio[c][h] /= lik_rat (nloc, i, sx, h, 0);

                chkprnt14 (i, sx);

		/*  do-msmpl */
		do_msmpl (nloc, c, i, sx, prob, y, ngenes, flag_t, t);

		/* put updated meiosis back in to l_ratio and
		 * update  the cumulative ratio                */

                chkprnt15 (i, sx);

		for (h = 1; h <= nh; ++h)
		{
		    l_ratio[c][h] *= lik_rat (nloc, i, sx, h, 0);
		    cumulik[c][h] += l_ratio[c][h];
		}

		/* cumulate total log-prob on compt c */
		tot_log_prob_seg[c][0] += 
                             full_seg_prob (nloc, c + 1, 0, recounts[c]);
		for (j = 0; j <= nloc; ++j)
		    tot_log_prob_pen[c][0] += log (prob[c][j]);
		for (h = 1; h <= nh; ++h)
		    tot_log_prob_seg[c][h] += 
                                full_seg_prob (nloc, c + 1, h, dummy);
		for (j = 0; j <= nloc; ++j)
		    ++count[c][j];

                chkprnt16 (t);
	    }
	}

/* now we are done with this simulation/estimation for this trait locus
   position  */
/* output results for this simulation.estimation set */

	for (c = 0; c < n_compnts; ++c)
	{
	    avcou = 0.0;
	    for (j = 0; j <= nloc; ++j)
		avcou += count[c][j];
	    if (avcou <= 0.0)
		avcou += 1;
	    avcou = avcou / (nloc + 1);
	    printf ("\n Component %d:\n\n", c + 1);
	    printf ("    Cumulative likelihood ratios:\n", k);
	    printf ("       Estimated ratios   left:sim = %6.4f,"
                                          "  right:sim = %6.4f\n",
	 	              cumulik[c][1] / avcou, cumulik[c][2] / avcou);
	    printf ("       Average count over loci = %7.3f\n", avcou);
	    printf ("       Average total log-probability of segregations\n"
                    "                          left = %6.3f,  sim = %6.3f,  right = %6.3f\n",
    		                  tot_log_prob_seg[c][1] / avcou, 
                                  tot_log_prob_seg[c][0] / avcou,
		                  tot_log_prob_seg[c][2] / avcou);
	    printf ("       Average total log-probability of penetrances "
                                   "(at sim) = %6.3f\n\n",
		                  tot_log_prob_pen[c][0] / avcou);
	    fflush (stdout);

	    /* print out recombination counts  */
            printf ("    Recombination counts:\n" );
            printf ("       map interval   M:  total  recomb non-rec    "
                                          "F:  total  recomb non-rec\n"); 
	    for (j = 0; j < nloc; ++j)
	    {
		printf ("              %d        ", j + 1);
		for (i = 0; i < 6; ++i)
                {    
                    if (i == 3)
                        printf ("     ");
		    printf ("%8d", recounts[c][j][i]);
                }
                printf ("\n");
	    }
            fflush (stdout);
	}
    }					/*****    end of BIG loop   *****/

/* now we are done with simulating for all nloc+2 simulation sets */

    fflush (stdout);

					/* Free up allocated memory */
    free_allocate ();
    free_gen_dat ();
    for (j = 0; j <= nloc; j++)
    {
	free (y[0][j]);
	free (y[1][j]);
    }
    free (y[0]);
    free (y[1]);
    free (y);
    if (ind_sw)
	free (ind_sw);
    if (sex_sw)
	free (sex_sw);
    free_allocate_loci ();

    chkprnt17 ();

					/* now we are done  */

    if (SamplSeedsOut)
    {
        getSeeds(SAMPL);
        if (!(oSeedF = open_fil(OSEED)))
            give_up(ERR_qit);
        writSeeds(oSeedF);
        fclose(oSeedF);
    }

    exit (EXIT_SUCCESS);

}

/****************************************************************/

void chkprnt1 (void)
{
#ifdef CHECK
    int i,j;
/* Check out all the data -- trait genotypes and markers   */
    for (i = 0; i < pedsize; ++i)
    {
	printf (" ind = %d", nghd[i].self);
	for (j = 0; j <= nloc; ++j)
	    printf ("    %d %d", y[0][j][i], y[1][j][i]);
        printf("\n");
    }
    printf("\n");  fflush (stdout); 
#endif /*CHECK*/
    return;
} 
void chkprnt2a (int j) 
{
#ifdef CHECK
    printf(" Set genes at locus %d\n",j); fflush(stdout);
    print_obs_genes(j, y);   fflush(stdout);      	
    printf (" Scored prob at locus %d; giving", j);  
    fflush (stdout);                          
#endif /*CHECK*/
    return;
}
void chkprnt2b (int j) 
{
#ifdef CHECK
    int c;
    for (c = 0; c < n_compnts; ++c)
        printf (" %7.4f", prob[c][j]);     
    fflush(stdout);
#endif /*CHECK*/
    return;
}
void chkprnt5 (void) 
{
#ifdef CHECK
    printf (" Now to call set_positions\n\n");
    fflush (stdout);
#endif /*CHECK*/
     return;
}
void chkprnt6 (void) 
{
#ifdef CHECK
    printf (" Now did call set_lsampler\n\n");  fflush (stdout);
#endif /*CHECK*/
    return;
}
void chkprnt7 (int j) 
{
#ifdef CHECK
    printf (" Reset trait locus meioses, position %d\n\n", j);
    fflush (stdout);
#endif /*CHECK*/
    return;
}
void chkprnt8 (void)
{
#ifdef CHECK
    printf (" Done set-up %d \n\n", k);
    fflush (stdout);
#endif /*CHECK*/ 
    return; 
} 
void chkprnt9 (int k)
{
#ifdef CHECK
    printf (" Initial seg lhood-ratios for simulation chromosome %d\n\n", k);
    fflush (stdout);
#endif /*CHECK*/ 
    return; 
}
void chkprnt19 (int n_compnts, int nh, int nloc)
{
#ifdef CHECK
    int c;
    for (c = 0; c < n_compnts; ++c)
    {
        printf (" Component %d", c + 1);
	printf ("    left point %7.4f,    right point %7.4f\n\n",
		    l_ratio[c][1], l_ratio[c][2]);
	fflush (stdout);
    }
#endif /*CHECK*/ 
    return; 
} 
void chkprnt10 (void)
{
#ifdef CHECK                /* print set-up if needed for chk */
    int j;
    for (j = 0; j <= nloc; ++j)
    { 
        printf(" Locus %d position %d set-up \n", j, chromo[j][0]);
        print_set_up (j); printf("\n");  fflush(stdout); 
    }
    printf("\n");  fflush (stdout);
#endif /*CHECK*/
    return;
}
void chkprnt11 (int l, int sx, int c)
{
#ifdef CHECK    
     printf (" sw %d, ind %d %d, cmp %d\n\n", 
                      l, nghd[i].self, sx, c);   
     fflush (stdout);
#endif /*CHECK*/
    return;
}
void chkprnt12 (int c, int nh)
{
#ifdef CHECK    
    int j, h;
    for (j = 0; j <= nloc; ++j)  
         printf("\n log_pen_prob locus %d is %7.4f", j, log(prob[c][j])); 
    for (h = 0; h <= nh; ++h)
         printf ("\n chromoset %d log_segs  %7.4f, cumulative totals %7.4f", 
                        h, full_seg_prob(nloc,c+1,h, dummy), 
                           tot_log_prob_seg[c][h]);
     fflush (stdout);
#endif /*CHECK*/
    return;
}
void chkprnt13 (int c)
{
#ifdef CHECK    

    printf ("\n chk l-ratio = %7.4f %7.4f", l_ratio[c][1], 
                        l_ratio[c][2]);        
     fflush (stdout);
#endif /*CHECK*/
    return;
}
void chkprnt14 (int i, int sx)
{
#ifdef CHECK    
    printf ("\n chk adj l-ratio by %7.4f %7.4f", lik_rat(nloc, i, sx, 1, 0), 
                                           lik_rat(nloc, i, sx, 2, 0));
     fflush (stdout);
#endif /*CHECK*/
    return;
}
void chkprnt15 (int i, int sx)
{
#ifdef CHECK    
    printf (" chk adj l-ratio back by %7.4f %7.4f\n",
                              lik_rat(nloc, i, sx, 1, 0), 
                              lik_rat(nloc, i, sx, 2, 0));
     fflush (stdout);
#endif /*CHECK*/
    return;
}
void chkprnt16 (int t)
{
#ifdef CHECK    
    printf (" iter %d:  lrat left %7.4f, lrat rght %7.4f\n\n",
                          t+1, l_ratio[c][1], l_ratio[c][2]);
     fflush (stdout);
#endif /*CHECK*/
    return;
}
void chkprnt17 (void)
{
#ifdef CHECK    
   loki_peel_free();
   printf(" \n Now we are done freeing\n\n ");
   fflush(stdout); 
#endif /*CHECK*/
    return;
}
