
/*:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::*/
/*       globs.h:  header file defining global variables               */
/*                 -  global in scope                                  */
/*                 -  included by main programs and function libraries */
/*                 -  contains modelling or simulation parameters      */
/*                    derived from the input statements                */ 
/*                 -  values are stored by Pars/ functions             */
/*:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::*/

#ifndef  GLOBS_H		
#define  GLOBS_H

/***  scalar variables, etc, derived from the parameter statements  ****/

GLOBAL int    PedSize;     

GLOBAL int    NInts,        NReals,       NFix,         NNInts;

GLOBAL int    NStdIn,       StdIn[5],     NStdOut,      StdOut[5];

GLOBAL int    DiscardReals;

GLOBAL int    NLoci,        NMarks,       NTraits,      NUnlnkd;

GLOBAL int    SexIn,        SexOp,        SexOut;

GLOBAL int    ObsIn,        ObsOp,        ObsOut;

GLOBAL int    AllVarOut,    ChronOut;

GLOBAL int    MarkSeedSet,  TraitSeedSet, MarkSeedsOut, TraitSeedsOut;

GLOBAL int    SamplSeedSet,  SamplSeedsOut;

GLOBAL uint32 MarkSeedC,    MarkSeedT,    TraitSeedC,   TraitSeedT;

GLOBAL uint32 SamplSeedC,   SamplSeedT;

       enum  sEnums {NAM, PA, MA, SEX, OBS};

/*****     the following variables are used by Nghds/ functions     *****/

GLOBAL int    MaxNames;    		/*Largest name for individual   */

GLOBAL int    NColData;	        	/*(No. of integers) + 1 to      */
                                        /*selectively read from ped file*/

GLOBAL int    MaxNCmps;   		/*No. of components allocated	*/

GLOBAL int    SelTrait[FMAX_REAL];	/*Selection/nonselection of 	*/
					/*reals (traits) from ped file	*/

GLOBAL int    SelInt[FMAX_INT];   	/*Selection/nonselection of     */
					/*integers from ped file  	*/

 
/**** the following variables and tables are used by Quant functions ****/

GLOBAL int    FENo[FIXEFF];		/*For fixed effects selected,	*/
					/*fixed effect #		*/

GLOBAL int    TrNo[MAX_TRAIT];		/*For traits selected,		*/
					/*trait #			*/

GLOBAL int    FESex;                    /*Gender a fixed effect?        */

GLOBAL int    FfDex[MAX_TRAIT][FIXEFF];         /*Presence/absence of   */
                                        /*fixed effects j in trait i    */

GLOBAL int    IFix[FIXEFF]; 	    	/*Number of levels of fixed	*/
					/* effects 			*/

GLOBAL int    IOCor;			/*IOCor == TRUE:  output        */
                                        /*correlations                  */

GLOBAL int    ChkEigC;			/*flag for printing progress in	*/
					/*computing eigenvalues		*/
  

/*** the following variables are used by Quant/ and PolyEM functions ****/

GLOBAL double Mufti[MAX_FEP][FIXEFF][LEVELS+1];  /*Fixed effect params; */
                                        /*Up to MAX_FEP combinations    */

GLOBAL double SigAdd, SigErr;           /* variance parameters          */
                                        /* (univariate)                 */

GLOBAL double TauSq[MAX_TRAIT*(MAX_TRAIT+1)/2],    /* covariances	*/
              SigSq[MAX_TRAIT*(MAX_TRAIT+1)/2];


/*** the following structure, _filtab, is used by Stuff/ functions      */
/*   and by other MORGAN functions                                      */

#define PAR      "par"
#define PED      "ped"
#define EIGEN    "eig"
#define PEEL     "peel"
#define SEED     "seed"
#define OPED     "oped"
#define OSCORE   "oscor"
#define OSEED    "oseed"

struct _filtab  {
  char *type;			        /*file type (com. line keyword) */
  int command;			  	/*flag for named in command line*/
  int param;				/*flag for named in par statemnt*/
  int done;                             /*flag for file has been read   */
  int parstat;				/*flag for contains param stmnts*/
  char *pcods;				/*progs which use this file type*/
  char *fop;				/*fopen mode "r", "w" or "a"	*/
  char *fname;				/*file name			*/
  char *fdesc;				/*description of file		*/
                };
typedef struct _filtab * FILTAB_PTR;

#ifdef  DEFINE_DATA
  struct _filtab Filtab[] = {
  { PAR   , N, N, N, Y, "pubgmdal", "r", NULL, "parameter file"       },
  { PED   , N, N, N, Y, "pubgmdal", "r", NULL, "input pedigree file"  },
  { EIGEN , N, N, N, N, "bgm"     , "r", NULL, "eigenvalue file"      },
  { PEEL  , N, N, N, N, "u"       , "r", NULL, "peeling file"         },
  { SEED  , N, N, N, Y, "dal"     , "r", NULL, "seed file"            },
  { OPED  , N, N, N, N, "pubgmdal", "w", NULL, "output pedigree file" },
  { OSCORE, N, N, N, N, "a"       , "w", NULL, "output score file"    },
  { OSEED , N, N, N, N, "dal"     , "a", NULL, "output seed file"     },
  { ""    , N, N, N, N, ""        , "",  NULL, ""                     }}; 
#else
  extern struct _filtab Filtab[];
#endif  /*DEFINE_DATA*/


/**** the following variables and tables are used by PolyEM programs ****/

                                        /*See limits.h for default	*/
					/*values                        */

GLOBAL int    ItMo;			/*Breeding-value iterations	*/
GLOBAL int    ItEM;			/*Number of EM iterations	*/
GLOBAL int    ItWt;			/*How often to write out estmts */
GLOBAL int    IPtEM;			/*iem =1 partitioned EM		*/
GLOBAL int    IEn;			/*ien =1 fit environmental model*/
GLOBAL double Diff;			/*Convergence criterion for	*/
					/*breeding values iteration	*/
GLOBAL int    ICh;			/*ich =1 -- call bigcheig	*/

GLOBAL int    ChkEig;			/*flag for printing eigenvals	*/
GLOBAL int    ChkGnv;			/*flag for printing G-iverse	*/
GLOBAL int    ChkGmt;			/*flag for printing G-matrix    */
GLOBAL int    ChkTrc;			/*flag for printing trace	*/

GLOBAL int    TauModel[MAX_TRAIT*(MAX_TRAIT+1)/2],      /* fit/no fit   */
              SigModel[MAX_TRAIT*(MAX_TRAIT+1)/2];      /* covariances  */
                                                        /* (multivar)   */

          
/******     the following structures are used by genedrop          ******/
                        
                        /* macros for recombMap->recombTyp below        */

#define FM 1			/* first marker (on chromosome)		*/
#define OM 2			/* other than first marker		*/
#define FT 3			/* trait is first locus on chromosome   */
#define IT 4			/* intermediate trait - has preceding	*/
				/* marker/trait locus & following marker*/
#define ET 5			/* trait with no following markers	*/
#define UT 6			/* unlinked trait			*/

			/*recombMap:  identifies neighbors and          */
                        /*            gender-specific probabilities     */
			/*            used in recombination algorithms	*/
struct recombMap {		
    int id;				/* markers & traits are in order*/
    int recombTyp;			/* of chrom, locus;  unlinked	*/
    int iA;				/* traits are last		*/
    int iB; 
    double probA[2];
    double probB[2];
                 };
typedef struct recombMap RMAP;
typedef RMAP * RMAP_PTR;
GLOBAL RMAP_PTR PRecombMap NINIT;

			/*traitPars:  parameters used for computing	*/
   			/*            the additive and residual 	*/
    			/*            of the phenotype			*/
struct traitPars {
    int id;				/* ids are in recombMap order	*/
    double means[3];
    double addi_term_found;
    double addi_term;
    double resi_term;
                 };
typedef struct traitPars TPARS;
typedef TPARS * TPARS_PTR;
GLOBAL TPARS_PTR PTraitPars NINIT;

			/*traitOut:  traits in numerical order 		*/
struct traitOut {
    int id;
    int wid;
    int dec;
    int ixG;
    int ixPh;
                };
typedef struct traitOut TOUT;
typedef TOUT * TOUT_PTR;
GLOBAL TOUT_PTR PTraitOut NINIT;

			/*alleleFreq:  allele frequencies for each locus*/
struct alleleFreq{		
    int nAlleles;			/*no. of alleles at this locus	*/
    double * pCumFreq;			/*ptr to list of cumulative frqs*/
                 };		
typedef struct alleleFreq ALLELE;
typedef ALLELE * ALLELE_PTR;
GLOBAL ALLELE_PTR PAlleleFreq NINIT;

GLOBAL int WidAlle;                     /*no. characters for printing   */
                                        /*alleles                       */
GLOBAL double * peigen;



/**** the following variables and tables are used by Autozyg programs ***/

GLOBAL int NIt,	  WIt,   FWIt;

GLOBAL int NTPos;	      /* no. of positions for trait on chrom   */
                              /* for lods scoring                      */

GLOBAL int NObs;              /* no. observed                          */

                              /* indices of loci in ordered chromosome,*/ 
                              /* for each set of recombination values  */
                              /* chromo[][0]...chromo[][NLoci-1]       */
GLOBAL int chromo[NH][NLOC]; 

                              /* positions of loci ordered chrmosomes  */
                              /* -- "inverse" of chromo[][]            */
                              /* ( chromo[][pos] = nam                 */
                              /*  revchro[][nam] = pos )               */ 
                              /* revchro[][0]...revchro[][NLoci-1]     */
GLOBAL int revchro[NH][NLOC];   

                              /* sex-valued recombinations in ordered  */
                              /* chromosome:   0 = sex-averaged,       */
                              /* 1 = male, 2 = female                  */
                              /* xecomo[0][][0]...xecomo[0][][NLoci-1] */
                              /*               ...xecomo[2][][NLoci-1] */
GLOBAL double xecomo[3][NH][NLOC]; 

                              /* no. of alleles at each locus          */
                              /* NAll[0]...NAll[Nloci-1]               */
                              /* ordered by name (chromo[][])          */
GLOBAL int NAll[NLOC];

                              /* allele frequencies at each locus      */
                              /* Qq[0][1]....Qq[0][NAll[0]]            */
                              /*          ...Qq[NLoci-1][Nall[NLoci-1]]*/ 
GLOBAL double **Qq;    

GLOBAL int NObs;              /* no. observed                          */
GLOBAL int IObs[NOBS];        /* names or indices of observed individs */
GLOBAL int Dat1[NLOC][NOBS], Dat2[NLOC][NOBS];
                              /* observed markers                      */
GLOBAL int NPr[NC];           /* no. probands for the component        */
GLOBAL int Probands[NC][G6];  /* indices for the individuals           */
                              /* & gametes to be scored for each       */
                              /* component                             */
GLOBAL int Proband_gamete[NC][G6];

GLOBAL int Rs;                /* = 0 for l sampler setup               */ 
GLOBAL int TraitPheno;        /* T for phenotype, F for genotype data  */ 
GLOBAL double LSamplProb;     /* locus sampler probability             */

GLOBAL double ** Marek;       /* lm_lods:  recomb fractions for markers*/
                              /*           (was [3][NLOC])             */
GLOBAL int * IntVal;          /*           recomb loci for trait       */ 
                              /*           (was [NP])                  */
GLOBAL double ** RecVal;      /*           recomb fractions for trait  */
                              /*           (was [3][NP])               */
#endif  /*GLOBS_H*/

