/****************************************************************************
 *                                                                          *
 *     Loki - Programs for genetic analysis of complex traits using MCMC    *
 *                                                                          *
 *             Simon Heath - University of Washington                       *
 *                         - Rockefeller University                         *
 *                                                                          *
 *                        August 1997                                       *
 *                                                                          *
 * Modified to work with Morgan libraries: SCH, July 1998                   *
 *                                                                          *
 * Corrections made, and modified for sex-specific maps: EAT, Feb, 2000     *
 *                                                                          *
 * lsample_peel.c:                                                          *
 *                                                                          *
 * Perform peeling calculations                                             *
 *                                                                          *
 ****************************************************************************/

#include "lsample.h"

extern double recomb(double);

static double *qval,*pval,*mval,*freq,*peel_famval,**complex_freq,**complex_pen,*complex_out_p;
static double *res,sige,konst;
static struct R_Func *rf;
static struct Peelseq_Head **peel_list,*peelseq_head;
static unsigned long **a_set,mask,*req_set[3],*idtemp[2],*peel_pt1;
static int n_idx,n_all,nn_all,*rfp,comp,*lumped[2],*nhaps[2],*ngens,*allele_trans,*peel_alls;
static int n_bits,n_terms,hash_mode,*peel_alls,*complex_mem;
static size_t hb_size=2048,mb_size=4096;
static struct fset *peel_fs;
int peel_trace;
static struct hash_element *hashtable[HASHTABLE_SIZE];
static struct hash_block *first_hash_block,*hash_block;

#define MRK_MBLOCK 0
#define TRT_MBLOCK 1
static struct peel_mem_block *first_mem_block[2],*mem_block[2];
static lk_ulong old_index;
static struct hash_element *old_element;
static double **seg_prob,**seg_prob1;
static int seg_mode;

int check_seg_mode(void) 
{
  return seg_mode;
}

void set_seg_mode(int i) 
{
  seg_mode=i;
}

void recessive(double *val,const Ind *id,const struct Locus *locus,double lam)
      { int i,j,obs,k;
        if (n_all != 2) 
            printf ("\n WARNING: trait locus has %d alleles\n",n_all);
	obs=qdex[id->index];
        if (obs>0)
           {  k=trtphen[id->index];
        if (k==4)
          { val[0]=0.0; val[1]=0.0; val[2]=0.0; val[3]=1.0; }
        else if (k==3)
          { val[0]=0.0; val[1]=1.0; val[2]=1.0; val[3]=0.0; }
        else if (k==2)
          { val[0]=1.0; val[1]=1.0; val[2]=1.0; val[3]=0.0; }
        else if (k==1)
          { val[0]=1.0; val[1]=0.0; val[2]=0.0; val[3]=0.0; }
        else
           printf ("\n UNRECOGNIZED TRAIT PHENOTYPE of %d is %d",
                id->self, k);
            }
        return;
}

void penetrance(double *val,const Ind *id,const struct Locus *locus,double lam)
{
	double *g_mu,z,y;
	int i,j,k,obs;
	
	if(CHK_PEEL(TRACE_LEVEL_1)) printf("Inserting penetrance probabilities for %d\n",id->self);
	obs=qdex[id->index];
	if(obs>0) { 
		g_mu=locus->g_mu;
		y=res[obs];
                k=0;
		for(i=0;i<n_all;i++) for(j=0;j<=i;j++) {
			z=y-lam*g_mu[k++];
			z=exp(-.5*z*z/sige)*konst;
			val[i*n_all+j]*=z;
			if(i!=j) val[j*n_all+i]*=z;
		}
	}
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "get_new_element"
/* Returns storage for storing a new non-zero element.  Elements are allocated
 * in blocks of size hb_size */
static struct hash_element *get_new_element(void)
{
	struct hash_element *element;
	
	while(hash_block->ptr>=hash_block->size) {
		if(!hash_block->next)
		  if(!(hash_block->next=calloc(1,sizeof(struct hash_block)))) ABT_FUNC(MMsg);
		hash_block=hash_block->next;
		if(!hash_block->elements) {
			if(!(hash_block->elements=malloc(sizeof(struct hash_element)*hb_size))) ABT_FUNC(MMsg);
			hash_block->size=hb_size;
			hash_block->ptr=0;
		}
	}
	element=hash_block->elements+hash_block->ptr++;
	element->left=element->right=0;
	element->index=0;
	element->p=0.0;
	n_terms++;
	return element;
}

double normalize(double *val,int n)
{
	int k;
	double p1=0.0;
	
	for(k=0;k<n;k++) p1+=val[k];
	if(p1>0.0) for(k=0;k<n;k++) val[k]/=p1;
	return p1;
}

/* Memory allocation routines for R-Functions */
#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "peel_mem_block"
static struct peel_mem_block *get_new_memblock(size_t size,int flag)
{
	struct peel_mem_block *p;
	
	if(!(p=malloc(sizeof(struct peel_mem_block)))) ABT_FUNC(MMsg);
	p->index=0;
	p->val=0;
	p->size=size;
	p->ptr=0;
	p->next=0;
	if(flag==MRK_MBLOCK) if(!(p->index=malloc(sizeof(lk_ulong)*size))) ABT_FUNC(MMsg);
	if(!(p->val=malloc(sizeof(double)*size))) ABT_FUNC(MMsg);
	return p;
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "get_rf_memory"
static void get_rf_memory(struct R_Func *rf,size_t size,int flag)
{
	size_t i;
	struct peel_mem_block *p,*p1;
	
	p1=mem_block[flag];
	i=p1->size-p1->ptr;
	if(size>i) {
		p=p1;
		while(p->next) {
			if(size<=p->next->size) break;
			p=p->next;
		}
		if(!(p->next)) {
			hb_size*=1.2;
			i=(size>hb_size)?size:hb_size;
			p=get_new_memblock(i,flag);
		} else {
			p1=p->next;
			p->next=p1->next;
			p=p1;
			p->ptr=0;
			p1=mem_block[flag];
		}
		p->next=p1->next;
		p1->next=p;
		p1=p;
	}
	i=p1->ptr;
	rf->index=p1->index+i;
	rf->p=p1->val+i;
	p1->ptr+=size;
	mem_block[flag]=p1;
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "loki_peel_alloc"
/* Allocate various structures required by the peeling routines */
/* Called by loki_peel_init(), do not call directly */
void loki_peel_alloc(struct Locus *locus,int nloci,int *comp_npeel)
{
	int i,j,k=2,k1=2,n_all,n_bits;

	for(i=0;i<nloci;i++) {
		j=locus[i].n_all;
		if(j>k) k=j;
		if(locus[i].locus_type==TRAIT_LOCUS && j>k1) k1=j;
	}
	if(!k) return;
	if(!(lumped[0]=malloc(sizeof(int)*pedsize*5))) ABT_FUNC(MMsg);
	lumped[1]=lumped[0]+pedsize;
	nhaps[0]=lumped[1]+pedsize;
	nhaps[1]=nhaps[0]+pedsize;
	ngens=nhaps[1]+pedsize;
	if(!(idtemp[0]=malloc(sizeof(unsigned long)*pedsize*2))) ABT_FUNC(MMsg);
	idtemp[1]=idtemp[0]+pedsize;
	if(!(freq=malloc(sizeof(double)*k))) ABT_FUNC(MMsg);
	if(!(seg_prob=malloc(sizeof(void *)*2*pedsize))) ABT_FUNC(MMsg);
	if(!(seg_prob[0]=malloc(sizeof(double)*8*pedsize))) ABT_FUNC(MMsg);
	for(i=1;i<pedsize*2;i++) seg_prob[i]=seg_prob[i-1]+4;
	seg_prob1=seg_prob+pedsize;
	n_all=k;
	n_bits=num_bits(n_all);
	j=k1*k1;
	if(!(peel_famval=malloc(sizeof(double)*j*j))) ABT_FUNC(MMsg);
	if(!(peel_alls=malloc(sizeof(int)*(pedsize+k)))) ABT_FUNC(MMsg);
	rfp=peel_alls+k;
	j=k*k;
	if(!(peel_fs=malloc(sizeof(struct fset)*j*j))) ABT_FUNC(MMsg);
	j=1<<(n_bits+n_bits);
	if(!(pval=malloc(sizeof(double)*j*(2+j)))) ABT_FUNC(MMsg);
	mval=pval+j;
	qval=mval+j;
	for(k=i=0;i<n_compnts;i++)	{
		j=comp_npeel[i];
		if(j>k) k=j;
	}
	if(k) if(!(peel_list=malloc(sizeof(void *)*k))) ABT_FUNC(MMsg);
	if(max_peel_off) 
	  if(!(peel_pt1=malloc(sizeof(long)*max_peel_off*2))) ABT_FUNC(MMsg);
	first_mem_block[MRK_MBLOCK]=get_new_memblock(mb_size,MRK_MBLOCK);
	first_mem_block[TRT_MBLOCK]=get_new_memblock(mb_size,TRT_MBLOCK);
}

/* Clean up memory used */
static void free_hash_blocks(void)
{
	struct hash_block *hb1;
	
	hash_block=first_hash_block;
	while(hash_block)	{
		if(hash_block->elements) free(hash_block->elements);
		hb1=hash_block->next;
		free(hash_block);
		hash_block=hb1;
	}
	first_hash_block=0;
}

/* Free the structures allocated by loki_peel_alloc() */
/* This routine is called by loki_peel_free(), do not call directly */
void peel_free(void)
{
	int k;
	struct peel_mem_block *p,*p1;
	
	for(k=0;k<2;k++) {
		p1=first_mem_block[k];
		while(p1) {
			p=p1->next;
			if(p1->index) free(p1->index);
			if(p1->val) free(p1->val);
			free(p1);
			p1=p;
		}
	}
	free_hash_blocks();
	if(seg_prob) {
		if(seg_prob[0]) free(seg_prob[0]);
		free(seg_prob);
	}
	if(complex_pen) {
		if(complex_pen[0]) free(complex_pen[0]);
		free(complex_pen);
	}
	if(complex_freq) {
		if(complex_freq[0]) free(complex_freq[0]);
		free(complex_freq);
		complex_freq=0;
	}
	if(complex_mem) {
		free(complex_mem);
		complex_mem=0;
	}
	if(complex_out_p) {
		free(complex_out_p);
		complex_out_p=0;
	}
	if(lumped[0]) free(lumped[0]);
	if(idtemp[0]) free(idtemp[0]);
	if(pval) free(pval);
	if(peel_pt1) free(peel_pt1);
	if(peel_alls) free(peel_alls);
	if(peel_list) free(peel_list);
	if(peel_fs) free(peel_fs);
	if(peel_famval) free(peel_famval);
	if(freq) free(freq);
}

/* Given index x (from R-Function), returns corresponding n allele types in gt */
static void get_gts(lk_ulong x,const int n,int *gt)
{
	int i;
	lk_ulong x1;
	unsigned long a;
	
	x1=x;
	a=(1<<n_bits)-1;
	i=0;
	while(x)	{
		gt[i++]=1+(int)(x&a);
		x>>=n_bits;
	}
	for(;i<n;i++) gt[i]=1;
	for(i=0;i<n;i++) if(gt[i]>n_all) {
		printf("OOOK!\n");
	}
}

/* The inverse of get_gts() - converts the n allele types in gt into an index */
static lk_ulong get_index1(const int n,const int *gt)
{
	int i;
	lk_ulong x;
	
	x=0;
	for(i=n-1;i>=0;i--) {
		x<<=n_bits;
		x|=gt[i]-1;
	}
	return x;
}

static void trans_all(FILE *fptr,int x)
{
	int i,j;
	
	i=allele_trans[x-1];
	if(i<0) {
		for(i=j=0;i<nn_all;i++) if(peel_alls[i]==n_all-1) {
			fputc(j++?',':'<',fptr);
			fprintf(fptr,"%d",i+1);
		}
		fputc('>',fptr);
	} else fprintf(fptr,"%d",i+1);
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "print_rfunction"
static void print_rfunction(FILE *fptr,const struct R_Func *rf,const int *inv,const int flag)
{
	int i,j,k,k1,k2,n,*gt;
	lk_ulong x;
	
	fputs("R-Function: ",fptr);
	n=rf->n_ind;
	if(!(gt=malloc(sizeof(int)*n))) ABT_FUNC(MMsg);
	for(j=0;j<n;j++) {
		fputc(j?',':'(',fptr);
		print_orig_allele_id(fptr,inv[j]);
	}
	fputs(")\n",fptr);
	for(i=0;i<rf->n_terms;i++) {
		fprintf(fptr,"%10g - ",rf->p[i]);
		x=(flag==MARKER_LOCUS?rf->index[i]:(lk_ulong)i);
		get_gts(x,n,gt);
		if(flag==MARKER_LOCUS) {
			for(j=0;j<n;j++) {
				k=abs(inv[j]-1);
				if(inv[j]<0) {
					x=req_set[X_PAT][k];
					k2=lumped[X_PAT][k];
				} else {
					x=req_set[X_MAT][k];	
					k2=lumped[X_MAT][k];
				}
				if(x && gt[j]==k2+1) {
					for(k2=k1=0;k1<n_all;k1++) if(x&(1<<k1)) {
						fputc(k2++?',':'[',fptr);
						trans_all(fptr,k1+1);
					}
					fputc(']',fptr);
				} else trans_all(fptr,gt[j]);
				fputc(' ',fptr);
			}
		} else for(j=0;j<n;j++) fprintf(fptr,"%d ",gt[j]);
		fputc('\n',stdout);
	}
	free(gt);
}
							
static int cmp_position(const void *s1,const void *s2)
{
	double x1,x2;
	
	x1=(*((struct Locus **)s1))->fempos;
	x2=(*((struct Locus **)s2))->fempos;
	if(x1<x2) return -1;
	if(x1>x2) return 1;
	return 0;
}

/* Transmission probs. are set up on an individual basis to
 * allow for linked loci to be taken into account */
#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "set transmission"
void set_transmission(struct Locus *locus,int index,int n_loci)
{
	int i,j,k,chr,nl,ix;
	double x[2],r,Mtp[2],Ptp[2],z,z1;
	struct Locus **t_ord;
	
	if(!locus || !locus[index].chromosome) { /* Set up for unlinked case */
		for(i=0;i<pedsize;i++) for(j=0;j<4;j++) {
			seg_prob[i][j]=.25;
			seg_prob1[i][j]=.5;
		}
	} else {
		if(!(t_ord=malloc(sizeof(struct Locus *)*n_loci))) ABT_FUNC(MMsg);
		chr=locus[index].chromosome;
		x[0]=locus[index].fempos;
		x[1]=locus[index].malepos;
		/* Collect all loci on the same chromosome as index */
		for(nl=i=0;i<n_loci;i++) if(locus[i].chromosome==chr) t_ord[nl++]=locus+i;
		/* If necessary, sort them on position */
		if(nl>1) qsort(t_ord,nl,sizeof(struct Locus *),cmp_position);
		/* Find index locus in order */
		for(ix=0;ix<nl;ix++) if(locus+index==t_ord[ix]) break;
		if(ix==nl) ABT_FUNC("Internal error - locus not found in linkage group\n");
	   for(i=0;i<pedsize;i++) {
			Mtp[0]=Mtp[1]=Ptp[0]=Ptp[1]=1.0;
			k=nghd[i].ma->index;
			if(k!=INVALID_INDEX) {
				/* Get nearest informative maternal seg on left */
				for(j=ix-1;j>=0;j--) {
					k=t_ord[j]->seg[X_MAT][i];
					if(k>=0) {
						r=recomb(0.01*(x[0]-t_ord[j]->fempos));
						Mtp[k]=1.0-r;
						Mtp[1-k]=r;
						break;
					}
				}
				/* Get nearest informative maternal seg on right */
				for(j=ix+1;j<nl;j++) {
					k=t_ord[j]->seg[X_MAT][i];
					if(k>=0) {
						r=recomb(0.01*(t_ord[j]->fempos-x[0]));
						Mtp[k]*=1.0-r;
						Mtp[1-k]*=r;
						break;
					}
				}
			}
			k=nghd[i].pa->index;
			if(k!=INVALID_INDEX) {
				/* Get nearest informative paternal seg on left */
				for(j=ix-1;j>=0;j--) {
					k=t_ord[j]->seg[X_PAT][i];
					if(k>=0) {
						r=recomb(0.01*(x[1]-t_ord[j]->malepos));
						Ptp[k]=1.0-r;
						Ptp[1-k]=r;
						break;
					}
				}
				/* Get nearest informative paternal seg on right */
				for(j=ix+1;j<nl;j++) {
					k=t_ord[j]->seg[X_PAT][i];
					if(k>=0) {
						r=recomb(0.01*(t_ord[j]->malepos-x[1]));
						Ptp[k]*=1.0-r;
						Ptp[1-k]*=r;
						break;
					}
				}
			}
			z=Mtp[0]+Mtp[1];
			z1=Ptp[0]+Ptp[1];
			for(k=0;k<2;k++) {
				Mtp[k]/=z;
				Ptp[k]/=z1;
				seg_prob1[i][X_MAT*2+k]=Mtp[k];
				seg_prob1[i][X_PAT*2+k]=Ptp[k];
			}
			seg_prob[i][X_MM_PM]=Mtp[X_MAT]*Ptp[X_MAT];
			seg_prob[i][X_MM_PP]=Mtp[X_MAT]*Ptp[X_PAT];
			seg_prob[i][X_MP_PM]=Mtp[X_PAT]*Ptp[X_MAT];
			seg_prob[i][X_MP_PP]=Mtp[X_PAT]*Ptp[X_PAT];
		}
		free(t_ord);
	}
}

/* Insert element with index idx into the binary tree hanging off *node */
static struct hash_element *insert_node(struct hash_element *node,lk_ulong idx)
{
	if(idx<node->index) {
		if(node->left) return insert_node(node->left,idx);
		return node->left=get_new_element();
	} else if(idx>node->index) {
		if(node->right) return insert_node(node->right,idx);
		return node->right=get_new_element();
	}
	return node;
}

/* Recursive routine for getting all elements from binary tree hanging off *node and storing
 * the values and indices from them in *tp and *tl */
static void get_nodes(struct hash_element *node,double *tp,lk_ulong *tl,int *j)
{
	if(node->left)	get_nodes(node->left,tp,tl,j);
	tp[*j]=node->p;
	tl[(*j)++]=node->index;
	if(node->right) get_nodes(node->right,tp,tl,j);
}

/* Similar to get_nodes, but stops when the cumulative probability >= z.  Allows sampling
 * from distribution */
static int sample_comb(struct hash_element *node,const double z,double *p,lk_ulong *idx)
{
	if(node->left) if(sample_comb(node->left,z,p,idx)) return 1;
	if(node->p>0.0) {
		*p+=node->p;
		if(*p>=z) {
			*idx=node->index;
			return 1;
		}
	}
	if(node->right) if(sample_comb(node->right,z,p,idx)) return 1;
	return 0;
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "add_to_term"
/* Adds p to output element defined by the n_out allele types in gt_store.  If the element
 * does not exist, it is created. */
static void add_to_term(const int n_out,int *gt_store,const double p)
{
	lk_ulong idx;
	int i;
	struct hash_element *element;

	idx=get_index1(n_out,gt_store);
	if(hash_mode) {/* If all possible combinations can fit into hashtable, we don't need to search */
		i=(int)idx;
		if(!hashtable[i])	{
			hashtable[i]=get_new_element();
			hashtable[i]->p=p;
		} else hashtable[i]->p+=p;
	} else {/* ...else we need to search */
		if(old_element) if(idx==old_index) {/* Avoid search if the index is the same as the last one */
			old_element->p+=p;
			return;
		}
		i=(int)(idx%HASHTABLE_SIZE); /* Get hash index */
		if(hashtable[i]) element=insert_node(hashtable[i],idx);
		else element=hashtable[i]=get_new_element();
		element->p+=p;
		element->index=idx;
		old_index=idx;
		old_element=element;
	}
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "setup_complex_peel"
static void setup_complex_peel(const struct Complex_Element *element, const int sampling,int *temp[],int *n_other)
{
	int i,j,k,k1,n_peel,n_inv,n_rf,*inv;
	static int complex_mem_size;
	
	n_inv=element->n_involved;
	n_peel=element->n_peel;
	n_rf=element->n_rfuncs;
	inv=element->involved;
	
	i=8*n_inv+n_rf+2*n_peel;
	if(!complex_mem) {
		complex_mem_size=i;
		if(!(complex_mem=calloc((size_t)i,sizeof(int)))) ABT_FUNC(MMsg);
	} else {
		if(i>complex_mem_size) {
			complex_mem_size=i;
			if(!(complex_mem=realloc(complex_mem,i*sizeof(int)))) ABT_FUNC(MMsg);
		}
		memset(complex_mem,0,i*sizeof(int));
	}
	temp[0]=complex_mem;
	for(i=1;i<9;i++) temp[i]=temp[i-1]+n_inv;
	for(;i<11;i++) temp[i]=temp[i-1]+n_peel;
	/* 'other' alleles are alleles not in R-Functions and not already sampled
	 * (if on reverse sampling pass) */
	k1=sampling?n_peel:n_inv; /* When we are sampling, the output alleles have already been sampled */
	for(*n_other=i=0;i<k1;i++) if(!(element->flags[i]&IN_RF)) {
		if(element->flags[i]&HAP_JNT) {
			k= -inv[i];
			for(j=0;j<n_inv;j++) if(inv[j]==k) temp[4][*n_other]=j;
		}
		temp[2][(*n_other)++]=i;
	}
	/* Make a list of alleles that we need transmission probs for */
	for(i=0;i<n_inv;i++)	{
		if(element->flags[i]&HAD_P) {
			k=nghd[-inv[i]-1].pa->index+1;
			for(j=0;j<n_inv;j++) {
				if(inv[j]==k) temp[6+X_MAT][i]=j;
				else if(-inv[j]==k) temp[6+X_PAT][i]=j;
			}
		}
		if(element->flags[i]&HAD_M) {
			k=nghd[inv[i]-1].ma->index+1;
			for(j=0;j<n_inv;j++) {
				if(inv[j]==k) temp[6+X_MAT][i]=j;
				else if(-inv[j]==k) temp[6+X_PAT][i]=j;
			}
		}
	}
	/* Form list of alleles not in R-Functions */
	if(sampling) {
		for(i=0;i<n_peel;i++) temp[0][i]=0;
		for(;i<n_inv;i++) temp[0][i]=1;
	} else for(i=0;i<n_inv;i++) temp[0][i]=0;	
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "loki_complex_peelop"
/* Perform general peeling operation defined in element (i.e., as opposed to a 'simple' 
 * nuclear family based peeling operation) */
static double loki_complex_peelop(const struct Complex_Element *element,const struct Locus *locus,const double lam,const int s_flag)
{
	int i,j,k,k1,k2,k3,k4,n_out,n_peel,n_inv,*inv,n_rf,n_ind,n_other,ef,ef1,sampling=0;
	int *gt_store,*gt_store1,*gt_store2,*other_ptr,*other_list,*rf_ptr,*jnt_list,ht_size=0,*gt[2];
	int *off_index[2],*fnd_list,n_fnd=0,all,*pen_list,n_pen,*temp_p[11],sparse;
	double max_terms,*tp,prob=0.0,z,p1,Konst=0.0,*p_rf=0;
	lk_ulong a,b,m,*tl;
	struct hash_element *elem;
	static int max_fnd,max_all,max_pen,max_pen1,out_p_size;
	
	/* Get details about peeling operation */
	n_inv=element->n_involved; /* No. alleles involved in op */
	n_peel=element->n_peel; /* No. alleles to peel out (absorb) */
	n_out=n_inv-n_peel; /* No. to appear in output R-Function */
	n_rf=element->n_rfuncs; /* No. input R-Functions */
	inv=element->involved; /* List of involved alleles (alleles to peel come first) */
	/* if s_flag is non-zero then we are doing a sampling run */
	/* if s_flag&OP_SAMPLING then we are on the reverse (sampling) phase */
	/* In any case, if !n_out and s_flag then we can sample */
	if(s_flag) {
		if(s_flag&OP_SAMPLING) sampling=1;
		else if(!n_out) sampling=1;
	}
	sparse=locus->sparse_flag;
	if(sparse) {
		/* Allocate first hash_block, if not already done so */
		if(!first_hash_block) {
			if(!(first_hash_block=malloc(sizeof(struct hash_block)))) ABT_FUNC(MMsg);
			if(!(first_hash_block->elements=malloc(sizeof(struct hash_element)*hb_size))) ABT_FUNC(MMsg);
			first_hash_block->next=0;
			first_hash_block->size=hb_size;
		}
		/* Reset pointers to all hash_blocks */
		hash_block=first_hash_block;
		while(hash_block)	{
			hash_block->ptr=0;
			hash_block=hash_block->next;
		}
		hash_block=first_hash_block;
	}
	n_terms=0; /* No. non-zero terms in output R-Function */
	/* if s_flag is non-zero then we are doing a sampling run */
	/* if s_flag&OP_SAMPLING then we are on the reverse (sampling) phase */
	/* In any case, if !n_out and s_flag then we can sample */
	if(s_flag) {
		if(s_flag&OP_SAMPLING) sampling=1;
		else if(!n_out) sampling=1;
	}
	if(CHK_PEEL(TRACE_LEVEL_1)) {
		fputs("Marker, complex ",stdout);
		fputs(sampling?"sample, ":"peel, ",stdout);
		fputs(sparse?"sparse: ":"full: ",stdout);
		print_complex_element(stdout,element,rf);
		fputc('\n',stdout);
	}
	/* Check size of function we are assembling.  Check index for output function will
	 * fit into an lk_ulong.  Compute max_terms, the number of output terms if they are all non-zero.
	 * See if we need to do hash search, or if output will fit into hashtable */
	/* If sampling then we assemble function on the peeled alleles */
	if(sampling) {
		max_terms=log((double)n_all)*(double)n_peel;
		if(sparse) {
			if(n_bits*n_peel>(int)LK_LONG_BIT) ABT_FUNC("Too many individuals in output R-Function - need a larger long\n");
			hash_mode=(log(2.0)*n_bits*n_peel<log((double)HASHTABLE_SIZE));
			ht_size=hash_mode?1<<(n_bits*n_peel):HASHTABLE_SIZE;
		}
	} else {/* Otherwise we assemble function on output alleles */
		max_terms=log((double)n_all)*(double)n_out;
		if(sparse) {
			if(n_bits*n_out>(int)LK_LONG_BIT) ABT_FUNC("Too many individuals in output R-Function - need a larger long\n");
			hash_mode=(log(2.0)*n_bits*n_out<log((double)HASHTABLE_SIZE));
			ht_size=hash_mode?1<<(n_bits*n_out):HASHTABLE_SIZE;
		}
	}
	gt[0]=locus->gt[0];
	gt[1]=locus->gt[1];
	if(sparse) {
		old_element=0;
		for(i=0;i<ht_size;i++) hashtable[i]=0;
	} else {
		i=n_terms=(int)(.5+exp(max_terms));
		if(!complex_out_p) {
			out_p_size=i+n_rf;
			if(!(complex_out_p=malloc(sizeof(double)*out_p_size))) ABT_FUNC(MMsg);
		}
		if(i+n_rf>out_p_size) {
			out_p_size=i+n_rf;
			if(!(complex_out_p=realloc(complex_out_p,sizeof(double)*out_p_size))) {
				ABT_FUNC(MMsg);
			}
		}
		p_rf=complex_out_p+i;
		for(j=0;j<i;j++) complex_out_p[j]=0.0;
	}
	setup_complex_peel(element,sampling,temp_p,&n_other);
	gt_store=temp_p[0];
	gt_store2=temp_p[1];
	other_list=temp_p[2];
	other_ptr=temp_p[3];
	jnt_list=temp_p[4];
	gt_store1=temp_p[5];
	off_index[0]=temp_p[6];
	off_index[1]=temp_p[7];
	pen_list=temp_p[8];
	fnd_list=temp_p[9];
	rf_ptr=temp_p[10];
	/* Compute masks used for finding mutually 
	 * consistent terms from multiple input R-Functions */
	if(n_rf) {
		for(i=0;i<n_rf;i++) {
			j=element->index[i];
			n_ind=rf[j].n_ind;
			if(sparse) {
				a=0;
				k2=0;
				b=(1<<n_bits)-1;
				for(k=n_ind-1;k>=0;k--) {
					a<<=n_bits;
					k1=rf[j].id_list[k];
					if(gt_store[k1]) a|=b;
					else {
						k2|=1<<k1;
						gt_store[k1]=1;
					}
				}
				rf[j].mask[sampling]=a;
				rf[j].mask1[sampling]=k2;
			} else {
				k2=0;
				k3=1;
				for(k=n_ind-1;k>=0;k--) {
					k1=rf[j].id_list[k];
					if(!gt_store[k1]) {
						k2|=1<<k1;
						k3*=n_all;
						gt_store[k1]=1;
					}
				}
				rf[j].mask1[sampling]=k2;
				rf[j].n_terms=k3;
			}
		}
	}
	/* Make a list of alleles we need penetrances and/or founder probs for */
	n_pen=0;
	if(n_peel) {
		for(k=0;k<n_peel;k++) {
			j=abs(inv[k])-1;
			k1=nghd[j].pa->index;
			if(k1!=INVALID_INDEX) continue;
			n_fnd++;
		}
		if(n_fnd) {
			if(!complex_freq) {
				if(!(complex_freq=malloc(sizeof(void *)*n_fnd))) ABT_FUNC(MMsg);
				max_all=n_all*n_fnd;
				if(!(complex_freq[0]=malloc(sizeof(double)*max_all))) ABT_FUNC(MMsg);
				max_fnd=n_fnd;
			} else {
				if(n_fnd>max_fnd) {
					max_fnd=n_fnd;
					tp=complex_freq[0];
					if(!(complex_freq=realloc(complex_freq,sizeof(void *)*n_fnd))) ABT_FUNC(MMsg);
					complex_freq[0]=tp;
				}
				if(n_fnd*n_all>max_all)	{
					max_all=n_all*n_fnd;
					if(!(complex_freq[0]=realloc(complex_freq[0],sizeof(double)*max_all))) ABT_FUNC(MMsg);
				}
			}
			for(i=1;i<n_fnd;i++) complex_freq[i]=complex_freq[i-1]+n_all;
			n_fnd=0;
			for(k=0;k<n_peel;k++) {
				j=abs(inv[k])-1;
				k1=nghd[j].pa->index;
				if(k1!=INVALID_INDEX) continue;
				if(inv[k]<0) {
					a=req_set[X_PAT][j];
					k2=lumped[X_PAT][j];
				} else {
					a=req_set[X_MAT][j];	
					k2=lumped[X_MAT][j];
				}
				for(k1=0;k1<n_all;k1++)	complex_freq[n_fnd][k1]=0.0;
				for(k1=0;k1<n_all;k1++)	{
					if(a&(1<<k1)) complex_freq[n_fnd][k2]+=freq[k1];
					else complex_freq[n_fnd][k1]=freq[k1];
				}
				fnd_list[n_fnd++]=k;
			}
		}
		if(locus->penetrance) {
			for(k=0;k<n_peel;k++) {
				for(k1=k+1;k1<n_inv;k1++) if(inv[k1]== -inv[k]) break;
				if(k1==n_inv) continue;
				k1=abs(inv[k])-1;
				if(qdex[k1]>0) pen_list[n_pen++]=k1;
			}
			if(n_pen) {
				if(!complex_pen) {
					if(!(complex_pen=malloc(sizeof(void *)*n_pen))) ABT_FUNC(MMsg);
					max_pen1=n_pen*n_idx;
					if(!(complex_pen[0]=malloc(sizeof(double)*max_pen1))) ABT_FUNC(MMsg);
					max_pen=n_pen;
				} else {
					if(n_pen>max_pen)	{
						max_pen=n_pen;
						tp=complex_pen[0];
						if(!(complex_pen=realloc(complex_pen,sizeof(void *)*n_pen))) ABT_FUNC(MMsg);
						complex_pen[0]=tp;
					}
					if(n_pen*n_idx>max_pen1) {
						max_pen1=n_pen*n_idx;
						if(!(complex_pen[0]=realloc(complex_pen[0],sizeof(double)*max_pen1))) ABT_FUNC(MMsg);
					}
				}
				for(k1=0;k1<n_pen*n_idx;k1++) complex_pen[0][k1]=1.0;
				for(k1=1;k1<n_pen;k1++) complex_pen[k1]=complex_pen[k1-1]+n_idx;
				for(k=0;k<n_pen;k++)	{
					locus->penetrance(complex_pen[k],nghd+pen_list[k],locus,lam);
					z=normalize(complex_pen[k],n_idx);
					if(z<=0.0) return -DBL_MAX;
					Konst+=log(z);
				}
			}
		}
	}
	for(i=0;i<n_inv;i++) gt_store[i]=0;
	ef1=0;
	i=0;
	/* Main loop.
	 *
	 * (1) If sampling then fix allele types of already sampled alleles.
	 * (2) Find mutually conistent configurations from input R-Functions
	 * (3) Find consistent configurations for 'other' alleles
	 * (4) Compute transmission/founder/penetrance probs
	 * (5) If still possible then add to output function
	 * (6) Cycle until tired
	 */
	for(;;) {
		/* Fix allele types of sampled alleles */
		if(sampling) for(j=0;j<n_out;j++) {
			k=inv[j+n_peel];
			if(k>0) gt_store[j+n_peel]=gt[X_MAT][k-1];
			else gt_store[j+n_peel]=gt[X_PAT][-k-1];
		}
		/* Cycle through R-Functions */
		for(;i<n_rf;i++) {
			j=element->index[i];
			ef=0;
			n_ind=rf[j].n_ind;
			/* Find term consistent with what's already been set up */
			if(sparse) {
				a=rf[j].mask[sampling];
				if(a)	{
					b=0;
					for(k=n_ind-1;k>=0;k--) {
						b<<=n_bits;
						k1=rf[j].id_list[k];
						if(gt_store[k1]) b|=gt_store[k1]-1;
					}
					b&=a;
					k1=rf[j].n_terms;
					tl=rf[j].index;
					for(k=rf_ptr[i];k<k1;k++) if((tl[k]&a)==b) break;
					rf_ptr[i]=k;
				}
				/* If consistent term found then get corresponding allele types */
				if(rf_ptr[i]<rf[j].n_terms) {
					get_gts(rf[j].index[rf_ptr[i]],n_ind,gt_store1);
					for(k=0;k<n_ind;k++)	{
						k1=rf[j].id_list[k];
						gt_store[k1]=gt_store1[k];
					}
				} else {/* Inconsistency found - try new combination */
					k2=rf[j].mask1[sampling];
					k=0;
					while(k2) {
						if(k2&1) gt_store[k]=0;
						k++;
						k2>>=1;
					}
					do	{
						if(!i) {
							ef1=1;
							break;
						}
						rf_ptr[i--]=0;
						rf_ptr[i]++;
						j=element->index[i];
						k2=rf[j].mask1[sampling];
						k=0;
						while(k2) {
							if(k2&1) gt_store[k]=0;
							k++;
							k2>>=1;
						}
					} while(rf_ptr[i]>=rf[j].n_terms);
					i--;
				}
			} else { /* full stored */
				k3=rf_ptr[i];
				k2=0;
				for(k=n_ind-1;k>=0;k--)	{
					k1=rf[j].id_list[k];
					if(!gt_store[k1]) {
						gt_store[k1]=1+(k3%n_all);
						k3/=n_all;
					}
					k2=k2*n_all+gt_store[k1]-1;
				}
				p_rf[i]=rf[j].p[k2];
				if(p_rf[i]<=0.0) { /* Inconsistent - find the next combination */
					for(;;) {
						k2=rf[j].mask1[sampling];
						k=0;
						while(k2) {
							if(k2&1) gt_store[k]=0;
							k++;
							k2>>=1;
						}
						if(++rf_ptr[i]<rf[j].n_terms) break;
						if(!i) {
							ef1=1;
							break;
						}
						rf_ptr[i--]=0;
						j=element->index[i];
					}
					i--;
				}
			}
			if(ef1) break;
		}
		if(ef1) break;
		/* We have a consistent configuration from the R-Functions, now find 1 for the 'other' alleles */
		for(ef=i=0;i<=n_other;i++)	{
			if(i==n_other)	{
				if(!ef) {
					for(k=0;k<n_inv;k++)	{
						if(inv[k]>0) {
							if(element->flags[k]&HAP_JNT)	{
								j=inv[k]-1;
								for(k1=0;k1<n_inv;k1++) if(inv[k1]== -inv[k]) {
									if(!(a_set[gt_store[k]-1][j]&(1<<(gt_store[k1]-1)))) ef=1;
									break;
								}
							}
						}
						if(ef) break;
					}
					/* Add Parent-off transmission probs. */
					p1=1.0;
					if(!ef) for(k=0;k<n_inv;k++) {
						if(element->flags[k]&HAD_P) { /* Sire-offspring */
							j= -inv[k]-1;
							tp=seg_prob1[j];
							m=1<<(gt_store[k]-1);
							a=req_set[X_PAT][j];
							if(a&m) m=a;
							z=0.0;
							all=gt_store[off_index[X_MAT][k]];
							if(m&(1<<(all-1))) z=tp[X_PAT*2+X_MAT];
							all=gt_store[off_index[X_PAT][k]];
							if(m&(1<<(all-1))) z+=tp[X_PAT*2+X_PAT];
							p1*=z;
							if(p1==0.0)	{
								ef=1;
								break;
							}
						} else if(element->flags[k]&HAD_M) {/* Dam-offspring */
							j=inv[k]-1;
							tp=seg_prob1[j];
							m=1<<(gt_store[k]-1);
							a=req_set[X_MAT][j];
							if(a&m) m=a;
							z=0.0;
							all=gt_store[off_index[X_MAT][k]];
							if(m&(1<<(all-1))) z=tp[X_MAT*2+X_MAT];
							all=gt_store[off_index[X_PAT][k]];
							if(m&(1<<(all-1))) z+=tp[X_MAT*2+X_PAT];
							p1*=z;
							if(p1==0.0)	{
								ef=1;
								break;
							}
							if(element->flags[k]&HAP_JNT)	{
								for(k1=0;k1<n_inv;k1++) if(inv[k1]== -inv[k]) {
									j=inv[k]-1;
									if(!(a_set[gt_store[k]-1][j]&(1<<(gt_store[k1]-1)))) ef=1;
									break;
								}
								if(ef) break;
							}
						}
					}
					/* Add penetrances and founder probs and store result */
					if(!ef) {
						if(sparse) for(k=0;k<n_rf;k++) p1*=rf[element->index[k]].p[rf_ptr[k]];
						else for(k=0;k<n_rf;k++) p1*=p_rf[k];
						for(k1=0;k1<n_fnd;k1++) {/* Add founder frequencies */
							k=fnd_list[k1];
							p1*=complex_freq[k1][gt_store[k]-1];
						}
						/* Add penetrances for those being peeled if required */
 						if(n_pen) {
							for(k=0;k<n_inv;k++)	{
								k1=inv[k];
								if(k1<0) gt[X_PAT][-k1-1]=gt_store[k];
								else gt[X_MAT][k1-1]=gt_store[k];
							}
							for(k=0;k<n_pen;k++)	{
								k1=pen_list[k];
								k2=((gt[X_PAT][k1]-1)<<n_bits)|(gt[X_MAT][k1]-1);
								p1*=complex_pen[k][k2];
							}
						}
						prob+=p1;
						if(sparse) {
							if(sampling) add_to_term(n_peel,gt_store,p1);
							else add_to_term(n_out,gt_store+n_peel,p1);
						} else {
							k1=0;
							if(sampling) for(k=n_peel-1;k>=0;k--) k1=k1*n_all+gt_store[k]-1;
							else for(k=n_inv-1;k>=n_peel;k--) k1=k1*n_all+gt_store[k]-1;
							complex_out_p[k1]+=p1;
						}
						ef=1;
					}
				}
			}
			/* Find next combination */
			if(ef) {
				i--;
				while(i>=0)	{
					other_ptr[i]++;
					gt_store[other_list[i]]=0;
					if(other_ptr[i]<n_all) break;
					other_ptr[i--]=0;
				}
				i--;
				if(i<-1) break;
				ef=0;
				continue;
			}
			k1=other_list[i];
			if(gt_store[k1]) continue;
			j=abs(inv[k1])-1;
			if(element->flags[k1]&HAP_JNT) {
				k2=jnt_list[i];
				if(inv[k1]>0) {
					if(gt_store[k2]) {
						b=1<<(gt_store[k2]-1);
						for(k=other_ptr[i];k<n_all;k++) if(a_set[k][j]&b) {
							gt_store[k1]=k+1;
							break;
						}
						if((other_ptr[i]=k)==n_all) ef=1;
					} else {
						for(k=0;k<n_other;k++) if(other_list[k]==k2) break;
						if(k==n_other) ABT_FUNC("Internal error - no other?\n");
						if(i<k) {
							for(k3=other_ptr[i];k3<n_all;k3++) {
								a=a_set[k3][j];
								if(!a) continue;
								for(k4=other_ptr[k];k4<n_all;k4++) if(a&(1<<k4)) {
									gt_store[k1]=k3+1;
									gt_store[k2]=k4+1;
									break;
								}
								if(k4==n_all) other_ptr[k]=0;
								else {
									other_ptr[k]=k4;
									break;
								}
							}
							if((other_ptr[i]=k3)==n_all) ef=1;
						} else {
							for(k4=other_ptr[k];k4<n_all;k4++) {
								a=1<<k4;
								if(idtemp[X_PAT][j]&a)	{
									for(k3=other_ptr[i];k3<n_all;k3++) if(a_set[k3][j]&a) {
										gt_store[k1]=k3+1;
										gt_store[k2]=k4+1;
										break;
									}
									if(k3==n_all) other_ptr[i]=0;
									else {
										other_ptr[i]=k3;
										break;
									}
								} else other_ptr[i]=0;
							}
							if((other_ptr[k]=k4)==n_all) ef=1;
						}
					}
				} else {
					if(gt_store[k2]) {
						b=a_set[gt_store[k2]-1][j];
						for(k=other_ptr[i];k<n_all;k++) if(b&(1<<k))	{
							gt_store[k1]=k+1;
							break;
						}
						if((other_ptr[i]=k)==n_all) ef=1;
					} else {
						for(k=i+1;k<n_other;k++) if(other_list[k]==k2) break;
						if(k==n_other) ABT_FUNC("Internal error - why for this happen?\n");
						if(i<k) {
							for(k3=other_ptr[i];k3<n_all;k3++) {
								a=1<<k3;
								if(idtemp[X_PAT][j]&a)	{
									for(k4=other_ptr[k];k4<n_all;k4++) if(a_set[k4][j]&a)	{
										gt_store[k1]=k3+1;
										gt_store[k2]=k4+1;
										break;
									}
									if(k4==n_all) other_ptr[k]=0;
									else {
										other_ptr[k]=k4;
										break;
									}
								} else other_ptr[k]=0;
							}
							if((other_ptr[i]=k3)==n_all) ef=1;
						} else {
							for(k4=other_ptr[k];k4<n_all;k4++) {
								a=a_set[k4][j];
								if(!a) continue;
								for(k3=other_ptr[i];k3<n_all;k3++) if(a&(1<<k3)) {
									gt_store[k1]=k3+1;
									gt_store[k2]=k4+1;
									break;
								}
								if(k3==n_all) other_ptr[i]=0;
								else {
									other_ptr[i]=k3;
									break;
								}
							}
							if((other_ptr[k]=k4)==n_all) ef=1;
						}
					}
				}
			} else {
				if(inv[k1]>0) {
					for(k=other_ptr[i];k<n_all;k++) if(a_set[k][j]) {
						gt_store[k1]=k+1;
						break;
					}
					if((other_ptr[i]=k)==n_all) ef=1;
				} else {
					a=idtemp[X_PAT][j];
					for(k=other_ptr[i];k<n_all;k++) if(a&(1<<k))	{
						gt_store[k1]=k+1;
						break;
					}
					if((other_ptr[i]=k)==n_all) ef=1;
				}
			}
		}
		for(i=0;i<n_other;i++) gt_store[other_list[i]]=0;
		/* Cycle through R-Functions */
		if(n_rf)	{
			i=n_rf-1;
			do	{
				rf_ptr[i]++;
				j=element->index[i];
				k2=rf[j].mask1[sampling];
				if(k2) {
					k=0;
					while(k2) {
						if(k2&1) gt_store[k]=0;
						k++;
						k2>>=1;
					}
					if(rf_ptr[i]<rf[j].n_terms) break;
				}
				rf_ptr[i--]=0;
			}
			while(i>=0);
			if(i>=0) ef=0;
		}
		else i=0;
		if(ef) break;
	}
	/* All valid combinations have been visited.  Have we found any ? */
	if(sparse) {
		if(!n_terms) return -DBL_MAX;
	} else {
		if(prob<=0.0) return -DBL_MAX;
	}
	/* If sampling then sample from output function */
	if(sampling) {
		z=rans()*prob;
		p1=0.0;
		if(sparse) {
			for(j=k=0;k<ht_size;k++) {
				elem=hashtable[k];
				if(!elem) continue;
				if(hash_mode) {
					if(elem->p>0.0) {
						p1+=elem->p;
						if(z<=p1) {
							a=(lk_ulong)k;
							break;
						}
					}
				} else if(sample_comb(elem,z,&p1,&a)) break;
			}
			if(k==ht_size)	ABT_FUNC("Internal error: No combination found when sampling?\n");
			get_gts(a,n_peel,gt_store);
		} else {
			for(k=0;k<n_terms;k++) {
				if(complex_out_p[k]>0.0) {
					p1+=complex_out_p[k];
					if(z<=p1) break;
				}
			}
			if(k==n_terms)	ABT_FUNC("Internal error: No combination found when sampling?\n");
			for(k1=0;k1<n_peel;k1++) {
				gt_store[k1]=(k%n_all)+1;
				k/=n_all;
			}
		}
		for(i=0;i<n_peel;i++) {
			j=inv[i];
			if(j>0) {
				gt[X_MAT][j-1]=gt_store[i];
				idflag[j-1]|=SAMPLED_MAT;
			} else {
				gt[X_PAT][-j-1]=gt_store[i];
				idflag[-j-1]|=SAMPLED_PAT;
			}
		}
	} else {/* Otherwise normalize and store */
		i=element->out_index;
		if(i>=0) {
			if(sparse) {
				rf[i].n_terms=n_terms;
				get_rf_memory(rf+i,n_terms,MRK_MBLOCK);
				tl=rf[i].index;
				tp=rf[i].p;
				for(j=k=0;k<ht_size;k++) {
					elem=hashtable[k];
					if(!elem) continue;
					if(hash_mode) {
						tp[j]=elem->p;
						tl[j++]=(lk_ulong)k;
					} else get_nodes(elem,tp,tl,&j);
				}
				for(j=0;j<n_terms;j++) tp[j]/=prob;
			} else {
				get_rf_memory(rf+i,n_terms,TRT_MBLOCK);
				for(j=0;j<n_terms;j++) rf[i].p[j]=complex_out_p[j]/prob;
				rf[i].n_terms=n_terms;
			}
			if(CHK_PEEL(TRACE_LEVEL_3)) print_rfunction(stdout,rf+i,inv+n_peel,sparse?MARKER_LOCUS:TRAIT_LOCUS);
		}
	}
	prob=log(prob)+Konst;
	if(CHK_PEEL(TRACE_LEVEL_2)) printf("Returning prob from %s() = %g\n",FUNC_NAME,prob);
	return prob;
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "loki_trait_complex_peelop"
/* Similar to loki_complex_peelop, but for a trait locus */
static double loki_trait_complex_peelop(const struct Complex_Element *element,const struct Locus *locus,const double lam,const int s_flag)
{
	int i,j,k,k1,k2,k3,n_out,n_peel,n_inv,*inv,n_rf,n_ind,n_other,ef,ef1,sampling=0;
	int *gt_store,*gt_store1,*other_ptr,*other_list,*rf_ptr,*jnt_list;
	int *off_index[2],*fnd_list,n_fnd=0,all,*pen_list,n_pen,*temp_p[11],*gt[2];
	double max_terms,*tp,prob=0.0,z,p1,Kon=0.0,*p_rf;
	static int max_pen,max_pen1,out_p_size;
	
	/* Get details about peeling operation */
	n_inv=element->n_involved; /* No. alleles involved in op */
	n_peel=element->n_peel; /* No. allles to peel out (absorb) */
	n_out=n_inv-n_peel; /* No. to appear in output R-Function */
	n_rf=element->n_rfuncs; /* No. input R-Functions */
	inv=element->involved; /* List of involved alleles (alleles to peel come first) */
	/* if s_flag is non-zero then we are doing a sampling run */
	/* if s_flag&OP_SAMPLING then we are on the reverse (sampling) phase */
	/* In any case, if !n_out and s_flag then we can sample */
	if(s_flag) {
		if(s_flag&OP_SAMPLING) sampling=1;
		else if(!n_out) sampling=1;
	}
	/* In a sampling operation we sample the peeled alleles conditional on
	 * R-Functions and on the pivot (output) alleles which have all already been sampled */
	if(CHK_PEEL(TRACE_LEVEL_1)) {
		if(sampling) fputs("Trait locus, complex sample: ",stdout);
		else fputs("Trait locus, complex peel: ",stdout);
		print_complex_element(stdout,element,rf);
		fputc('\n',stdout);
	}
	gt[0]=locus->gt[0];
	gt[1]=locus->gt[1];
	if(sampling) max_terms=log((double)n_all)*(double)n_peel;
	else max_terms=log((double)n_all)*(double)n_out;
	i=(int)(.5+exp(max_terms));
	if(!complex_out_p) {
		out_p_size=i+n_rf;
		if(!(complex_out_p=malloc(sizeof(double)*out_p_size))) ABT_FUNC(MMsg);
	}
	if(i+n_rf>out_p_size) {
		out_p_size=i+n_rf;
		if(!(complex_out_p=realloc(complex_out_p,sizeof(double)*out_p_size))) ABT_FUNC(MMsg);
	}
	p_rf=complex_out_p+i;
	for(j=0;j<i;j++) complex_out_p[j]=0.0;
	setup_complex_peel(element,sampling,temp_p,&n_other);
	gt_store=temp_p[0];
	other_list=temp_p[2];
	other_ptr=temp_p[3];
	jnt_list=temp_p[4];
	gt_store1=temp_p[5];
	off_index[0]=temp_p[6];
	off_index[1]=temp_p[7];
	pen_list=temp_p[8];
	fnd_list=temp_p[9];
	rf_ptr=temp_p[10];
	/* Compute masks - used for finding mutually 
	 * consistent terms from multiple input R-Functions */
	if(n_rf)	{
		for(i=0;i<n_rf;i++) {
			j=element->index[i];
			n_ind=rf[j].n_ind;
			k2=0;
			k3=1;
			for(k=n_ind-1;k>=0;k--) {
				k1=rf[j].id_list[k];
				if(!gt_store[k1]) {
					k2|=1<<k1;
					k3*=n_all;
					gt_store[k1]=1;
				}
			}
			rf[j].mask1[sampling]=k2;
			rf[j].n_terms=k3;
		}
	}
	/* Make a list of alleles we need penetrances and/or founder probs for */
	n_pen=0;
	if(n_peel) {
		for(k=0;k<n_peel;k++) {
			j=abs(inv[k])-1;
			k1=nghd[j].pa->index;
			if(k1!=INVALID_INDEX) continue;
			n_fnd++;
		}
		if(n_fnd) {
			n_fnd=0;
			for(k=0;k<n_peel;k++) {
				j=abs(inv[k])-1;
				k1=nghd[j].pa->index;
				if(k1!=INVALID_INDEX) continue;
				fnd_list[n_fnd++]=k;
			}
		}
		for(k=0;k<n_peel;k++) {
			for(k1=k+1;k1<n_inv;k1++) if(inv[k1]== -inv[k]) break;
			if(k1==n_inv) continue;
			k1=abs(inv[k])-1;
			if(qdex[k1]>0) pen_list[n_pen++]=k1;
		}
		if(n_pen) {
			if(!complex_pen) {
				if(!(complex_pen=malloc(sizeof(void *)*n_pen))) ABT_FUNC(MMsg);
				max_pen1=n_pen*n_idx;
				if(!(complex_pen[0]=malloc(sizeof(double)*max_pen1))) ABT_FUNC(MMsg);
				max_pen=n_pen;
			} else {
				if(n_pen>max_pen) {
					max_pen=n_pen;
					tp=complex_pen[0];
					if(!(complex_pen=realloc(complex_pen,sizeof(void *)*n_pen))) ABT_FUNC(MMsg);
					complex_pen[0]=tp;
				}
				if(n_pen*n_idx>max_pen1) {
					max_pen1=n_pen*n_idx;
					if(!(complex_pen[0]=realloc(complex_pen[0],sizeof(double)*max_pen1))) ABT_FUNC(MMsg);
				}
			}
			for(k1=0;k1<n_pen*n_idx;k1++) complex_pen[0][k1]=1.0;
			for(k1=1;k1<n_pen;k1++) complex_pen[k1]=complex_pen[k1-1]+n_idx;
			for(k=0;k<n_pen;k++) {
				locus->penetrance(complex_pen[k],nghd+pen_list[k],locus,lam);
				z=normalize(complex_pen[k],n_idx);
				if(z<=0.0) return -DBL_MAX;
				Kon+=log(z);
			}
		}
	}
	for(i=0;i<n_inv;i++) gt_store[i]=0;
	ef1=0;
	i=0;
	/* Main loop. */
	for(;;) {
		/* Fix allele types of sampled alleles */
		if(sampling) for(j=0;j<n_out;j++) {
			k=inv[j+n_peel];
			if(k>0) gt_store[j+n_peel]=gt[X_MAT][k-1];
			else gt_store[j+n_peel]=gt[X_PAT][-k-1];
		}
		/* Cycle through R-Functions */
		for(;i<n_rf;i++) {
			j=element->index[i];
			ef=0;
			n_ind=rf[j].n_ind;
			/* Find term consistent with what's already been set up */
			k3=rf_ptr[i];
			k2=0;
			for(k=n_ind-1;k>=0;k--)	{
				k1=rf[j].id_list[k];
				if(!gt_store[k1]) {
					gt_store[k1]=1+(k3%n_all);
					k3/=n_all;
				}
				k2=k2*n_all+gt_store[k1]-1;
			}
			p_rf[i]=rf[j].p[k2];
			if(p_rf[i]<=0.0) {/* Inconsistent - find the next combination */
				for(;;) {
					k2=rf[j].mask1[sampling];
					k=0;
					while(k2) {
						if(k2&1) gt_store[k]=0;
						k++;
						k2>>=1;
					}
					if(++rf_ptr[i]<rf[j].n_terms) break;
					if(!i) {
						ef1=1;
						break;
					}
					rf_ptr[i--]=0;
					j=element->index[i];
				}
				i--;
			}
			if(ef1) break;
		}
		if(ef1) break;
		for(ef=i=0;i<=n_other;i++)	{
			if(i==n_other) {
				if(!ef) {
					/* Add Parent-off transmission probs. */
					p1=1.0;
					if(!ef) for(k=0;k<n_inv;k++) {
						if(element->flags[k]&HAD_P) {/* Sire-offspring */
							j= -inv[k]-1;
							tp=seg_prob1[j];
							k3=gt_store[k];
							z=0.0;
							all=gt_store[off_index[X_MAT][k]];
							if(k3==all) z=tp[X_PAT*2+X_MAT];
							all=gt_store[off_index[X_PAT][k]];
							if(k3==all) z+=tp[X_PAT*2+X_PAT];
							p1*=z;
							if(p1==0.0) {
								ef=1;
								break;
							}
						} else if(element->flags[k]&HAD_M) {/* Dam-offspring */
							j=inv[k]-1;
							tp=seg_prob1[j];
							k3=gt_store[k];
							z=0.0;
							all=gt_store[off_index[X_MAT][k]];
							if(k3==all) z=tp[X_MAT*2+X_MAT];
							all=gt_store[off_index[X_PAT][k]];
							if(k3==all) z+=tp[X_MAT*2+X_PAT];
							p1*=z;
							if(p1==0.0) {
								ef=1;
								break;
							}
						}
					}
					/* Add penetrances and founder probs and store result */
					if(!ef) {
						for(k=0;k<n_rf;k++) p1*=p_rf[k];
						for(k1=0;k1<n_fnd;k1++) {/* Add founder frequencies */
							k=fnd_list[k1];
							j=abs(inv[k])-1;
							p1*=freq[gt_store[k]-1];
						}
						/* Add penetrances for those being peeled if required */
 						if(n_pen) {
							for(k=0;k<n_inv;k++)	{
								k1=inv[k];
								if(k1<0) gt[X_PAT][-k1-1]=gt_store[k];
								else gt[X_MAT][k1-1]=gt_store[k];
							}
							for(k=0;k<n_pen;k++)	{
								k1=pen_list[k];
								k2=((gt[X_PAT][k1]-1)<<n_bits)|(gt[X_MAT][k1]-1);
								p1*=complex_pen[k][k2];
							}
						}
						prob+=p1;
						k1=0;
						if(sampling) for(k=n_peel-1;k>=0;k--) k1=k1*n_all+gt_store[k]-1;
						else for(k=n_inv-1;k>=n_peel;k--) k1=k1*n_all+gt_store[k]-1;
						complex_out_p[k1]+=p1;
						ef=1;
					}
				}
			}
			/* Find next combination */
			if(ef) {
				i--;
				while(i>=0) {
					other_ptr[i]++;
					gt_store[other_list[i]]=0;
					if(other_ptr[i]<n_all) break;
					other_ptr[i--]=0;
				}
				i--;
				if(i<-1) break;
				ef=0;
			} else gt_store[other_list[i]]=other_ptr[i]+1;
		}
		for(i=0;i<n_other;i++) gt_store[other_list[i]]=0;
		/* Cycle through R-Functions */
		if(n_rf)	{
			i=n_rf-1;
			do	{
				rf_ptr[i]++;
				j=element->index[i];
				k2=rf[j].mask1[sampling];
				if(k2) {
					k=0;
					while(k2) {
						if(k2&1) gt_store[k]=0;
						k++;
						k2>>=1;
					}
					if(rf_ptr[i]<rf[j].n_terms) break;
				}
				rf_ptr[i--]=0;
			}	while(i>=0);
			if(i>=0) ef=0;
		} else i=0;
		if(ef) break;
	}
	/* All valid combinations have been visited.  Have we found any ? */
	if(prob<=0.0) return -DBL_MAX;
	n_terms=(int)(.5+exp(max_terms));
	/* If sampling then sample from output function */
	if(sampling) {
		z=rans()*prob;
		p1=0.0;
		for(k=0;k<n_terms;k++) {
			if(complex_out_p[k]>0.0) {
				p1+=complex_out_p[k];
				if(z<=p1) break;
			}
		}
		if(k==n_terms)	ABT_FUNC("Internal error: No combination found when sampling?\n");
		for(k1=0;k1<n_peel;k1++) {
			gt_store[k1]=(k%n_all)+1;
			k/=n_all;
		}
		for(i=0;i<n_peel;i++) {
			j=inv[i];
			if(j>0) {
				gt[X_MAT][j-1]=gt_store[i];
				idflag[j-1]|=SAMPLED_MAT;
			} else {
				gt[X_PAT][-j-1]=gt_store[i];
				idflag[-j-1]|=SAMPLED_PAT;
			}
		}
	} else { /* Otherwise normalize and store */
		i=element->out_index;
		if(i>=0) {
			get_rf_memory(rf+i,n_terms,TRT_MBLOCK);
			for(j=0;j<n_terms;j++) rf[i].p[j]=complex_out_p[j]/prob;
			rf[i].n_terms=n_terms;
			if(CHK_PEEL(TRACE_LEVEL_3)) print_rfunction(stdout,rf+i,inv+n_peel,TRAIT_LOCUS);
		}
	}
	prob=log(prob)+Kon;
	if(CHK_PEEL(TRACE_LEVEL_2)) printf("Returning prob from %s() = %g\n",FUNC_NAME,prob);
	return prob;
}

/* Similar to get_par_probs, but for a trait locus */
static double get_trait_par_probs(double *val,const Ind *id,const struct Locus *locus,const double lam)
{
	int i,j,k,fflag=0,l1,l2;
	double f;
	struct R_Func *rf1=0;

	if(id->pa->index==INVALID_INDEX) fflag=1;
	if(fflag && CHK_PEEL(TRACE_LEVEL_2)) printf("Inserting founder genotype probabilities for %d\n",id->self);
	i=id->index;
	k=rfp[i];
	if(k) {
		rf1=rf+k-1; 	/* Previously computed R_Func */
		if(CHK_PEEL(TRACE_LEVEL_2)) printf("Inserting previous R-Function for %d\n",id->self);
	}
	if(!fflag) {
		if(k) for(j=0;j<n_idx;j++) val[j]=rf1->p[j];
		else for(j=0;j<n_idx;j++) val[j]=1.0;
	} else for(j=l1=0;l1<n_all;l1++) {
		f=freq[l1];
		for(l2=0;l2<n_all;l2++)	{
			if(k) val[j]=rf1->p[j]*freq[l2]*f;
			else val[j]=freq[l2]*f;
			j++;
		}
	}
	if(locus->penetrance && qdex[i]>0) locus->penetrance(val,id,locus,lam);
	f=normalize(val,n_idx);
	if(CHK_PEEL(TRACE_LEVEL_2)) printf("Returning prob from get_trait_par_probs() = %g\n",log(f));
	return f;
}

/* Get parental probability dist., put into *val */
static double get_par_probs(double *val,const Ind *id,const struct Locus *locus,const double lam)
{
	int i,j,k,fflag=0;
	double f,f1,lf[2];
	unsigned long cm[2],a;
	lk_ulong l,l1,l2;
	struct R_Func *rf1;
	
	if(id->pa->index==INVALID_INDEX) fflag=1;
	if(fflag && (CHK_PEEL(TRACE_LEVEL_2))) printf("Inserting founder genotype probabilities for %d\n",id->self);
	i=id->index;
	/* If founder, calculate founder probs */
	if(fflag) for(j=0;j<2;j++) {
		lf[j]=0.0;
		cm[j]=req_set[j][i];
		if(cm[j]) {
			a=cm[j];
			for(k=0;k<n_all;k++) if(a&(1<<k)) lf[j]+=freq[k];
		}
	}
	for(k=0;k<n_idx;k++) val[k]=0.0;
	k=rfp[i];
	if(k) { /* Insert Previously computed R_Func */
		rf1=rf+k-1;
		if(CHK_PEEL(TRACE_LEVEL_2)) 
                   printf("Inserting previous R-Function for %d\n",id->self);
		if(locus->sparse_flag) {
			if(!fflag) for(j=0;j<rf1->n_terms;j++) 
                           val[rf1->index[j]]=rf1->p[j]; /* Non-founder */
			else for(j=0;j<rf1->n_terms;j++) { /* Founder */
				l=rf1->index[j];
				l1=l&mask;
				l2=(l>>n_bits)&mask;
				if((cm[X_MAT])&(1<<l1)) f=lf[X_MAT];
				else f=freq[l1];
				if((cm[X_PAT])&(1<<l2)) f1=lf[X_PAT];
				else f1=freq[l2];
/* BUG fix from SCH, installed by EAT, 8 Feb, 2000 */
				val[l]=rf1->p[j]*f*f1;
			}
		} else {
			if(!fflag) for(j=0;j<n_idx;j++) val[j]=rf1->p[j]; 
                                                           /* Non-founder */
			else for(j=l1=0;l1<(unsigned int)n_all;l1++) {
				if((cm[X_MAT])&(1<<l1)) f=lf[X_MAT];
				f=freq[l1];
				for(l2=0;l2<(unsigned int)n_all;l2++) { /* Founder */
					if((cm[X_PAT])&(1<<l2)) f1=lf[X_PAT];
					else f1=freq[l2];
					/**** val[j++]=f*f1*rf1->p[j]; **May 2000**/
					val[j]=f*f1*rf1->p[j]; j++;
                                        
				}
			}
		}
	} else { /* No R-Function - remove illegal genotypes */
		if(!fflag) for(j=0;j<n_all;j++) {
			a=a_set[j][i];
			if(locus->sparse_flag) {
				if(a) for(k=0;k<n_all;k++) if(a&(1<<k)) {
					l=(lk_ulong)((k<<n_bits)|j);
					val[l]=1.0;
				}
			} else if(a) for(k=0;k<n_all;k++) if(a&(1<<k)) val[j+k*n_all]=1.0;
		} else for(j=0;j<n_all;j++) {
			a=a_set[j][i];
			if(!a) continue;
			if((cm[X_MAT])&(1<<j)) f=lf[X_MAT];
			else f=freq[j];
			for(k=0;k<n_all;k++) if(a&(1<<k)) {
				if((cm[X_PAT])&(1<<k)) f1=lf[X_PAT];
				else f1=freq[k];
				l=locus->sparse_flag?(lk_ulong)((k<<n_bits)|j):k*n_all+j;
				val[l]=f*f1;
			}
		}
	}
	/* Add penetrances */
	if(locus->penetrance && qdex[i]) locus->penetrance(val,id,locus,lam);
	f=normalize(val,n_idx);
	if(CHK_PEEL(TRACE_LEVEL_2)) printf("Returning prob from get_par_probs() = %g\n",log(f));
	return f;
}

static void check_for_trait_rf(double *val,Ind *id)
{
	int j,k;
	struct R_Func *rf1;
	
	k=rfp[id->index];
	if(k) {/* Insert Previously computed R_Func */
		rf1=rf+k-1;
		if(CHK_PEEL(TRACE_LEVEL_2)) printf("Inserting previous R-Function for %d\n",id->self);
		for(j=0;j<n_idx;j++) val[j]=rf1->p[j];
	} else for(j=0;j<n_idx;j++) val[j]=1.0;
}

static void check_for_rf(double *val,Ind *id,int sparse)
{
	int j,k,i;
	unsigned long a,l;
	struct R_Func *rf1;
	
	i=id->index;
	k=rfp[i];
	for(j=0;j<n_idx;j++) val[j]=0.0;
	if(k) {/* Insert Previously computed R_Func */
		rf1=rf+k-1;
		if(CHK_PEEL(TRACE_LEVEL_2)) printf("Inserting previous R-Function for %d\n",id->self);
		if(sparse) for(j=0;j<rf1->n_terms;j++) val[rf1->index[j]]=rf1->p[j];
		else for(j=0;j<n_idx;j++) val[j]=rf1->p[j];
	} else for(j=0;j<n_all;j++) {
		a=a_set[j][i];
		if(a) for(k=0;k<n_all;k++) if(a&(1<<k)) {
			l=(lk_ulong)((k<<n_bits)|j);
			val[l]=1.0;
		}		
	}
}

static void set_for_sampled(double *val,int idx,const struct Locus *locus)
{
	int j,k;
	
	for(j=0;j<n_idx;j++) val[j]=0.0;
	j=locus->gt[X_MAT][idx]-1;
	k=locus->gt[X_PAT][idx]-1;
	if(locus->sparse_flag) val[(k<<n_bits)|j]=1.0;
	else val[k*n_all+j]=1.0;
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "setup_parents"
static double setup_parents(double *val,const Ind *id,const struct Locus *locus,const double lam,int *fg)
{
	int idx;
	double p;
	
	idx=id->index;
	if(idflag[idx]&SAMPLED_MAT) {
		set_for_sampled(val,idx,locus);
		p=1.0;
		*fg=0;
	} else {
		if(locus->locus_type==MARKER_LOCUS) p=get_par_probs(val,id,locus,lam);
		else p=get_trait_par_probs(val,id,locus,lam);
		if(p<=0.0) return -DBL_MAX;
		*fg=1;
	}
	return p;
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "loki_trait_simple_sample"
/* Similar to loki_simple_sample, but for a trait locus */
static double loki_trait_simple_sample(const struct Simple_Element *element,const struct Locus *locus,const double lam)
{
	int i,j,k,l,n,m,*gt[2],k1,sidx,didx;
	double prob=0.0,*tp,p1,p2,pp[4],z;
	Ind *sire,*dam,*kid;
	Indlist *id_list;

	gt[0]=locus->gt[0];
	gt[1]=locus->gt[1];
	sire=element->sire;
	sidx=sire->index;
	dam=element->dam;
	didx=dam->index;
	if(CHK_PEEL(TRACE_LEVEL_1)) {
		fputs("Trait locus, family sample: ",stdout);
		print_simple_element(stdout,element,rf);
		fputc('\n',stdout);
	}
	p1=setup_parents(pval,sire,locus,lam,&i);
	if(p1<0.0) ABT_FUNC("Zero probability in sampling operation\n");
	prob+=log(p1);
	p1=setup_parents(mval,dam,locus,lam,&j);
	if(p1<0.0) ABT_FUNC("Zero probability in sampling operation\n");
	prob+=log(p1);
	if(i || j) {
		i=n_idx*n_idx;
		for(j=0;j<i;j++) peel_famval[j]=0.0;
		for(n=0,j=0;j<n_idx;j++) {
			p1=pval[j];
			for(m=0;m<n_idx;m++) peel_famval[n++]=p1*mval[m];
		}
		id_list=element->off;
		while(id_list) {
			kid=id_list->current;
			id_list=id_list->next;
			j=kid->index;
			if(idflag[j]&SAMPLED_MAT) set_for_sampled(qval,j,locus);
			else {
				check_for_trait_rf(qval,kid);
				if(locus->penetrance && qdex[j]) {
					locus->penetrance(qval,kid,locus,lam);
					p1=normalize(qval,n_idx);
					if(p1<=0.0) return -DBL_MAX;
					prob+=log(p1);
				}
			}
			tp=seg_prob[j];
			for(n=0,j=0;j<n_all;j++) {
				for(i=0;i<n_all;i++)	{
					for(l=0;l<n_all;l++)	{
						for(k=0;k<n_all;k++)	{
							peel_famval[n++]*=
							  tp[X_MM_PM]*qval[i*n_all+k]+
							  tp[X_MM_PP]*qval[j*n_all+k]+
							  tp[X_MP_PM]*qval[i*n_all+l]+
							  tp[X_MP_PP]*qval[j*n_all+l];
						}
					}
				}
			}
		}
		p1=normalize(peel_famval,n_idx*n_idx);
		prob+=log(p1);
		z=rans();
		p2=0.0;
		for(m=n=0,j=0;j<n_all;j++)	{
			for(i=0;i<n_all;i++) {
				for(l=0;l<n_all;l++)	{
					for(k=0;k<n_all;k++)	{
						p1=peel_famval[n++];
						if(p1<=0.0) continue;
						p2+=p1;
						if(p2>=z) {
							m=1;
							gt[X_MAT][sidx]=i+1;
							gt[X_PAT][sidx]=j+1;
							gt[X_MAT][didx]=k+1;
							gt[X_PAT][didx]=l+1;
							idflag[sidx]|=(SAMPLED_MAT|SAMPLED_PAT);
							idflag[didx]|=(SAMPLED_MAT|SAMPLED_PAT);
							i=j=k=l=n_all;
							break;
						}
					}
				}
			}
		}
		if(!m) ABT_FUNC("Internal error - no parental combination found\n");
	}
	id_list=element->off;
	while(id_list) {
		kid=id_list->current;
		id_list=id_list->next;
		k1=kid->index;
		if(idflag[k1]&SAMPLED_MAT) continue;
		check_for_trait_rf(qval,kid);
		if(locus->penetrance && qdex[k1]) {
			locus->penetrance(qval,kid,locus,lam);
			p1=normalize(qval,n_idx);
			prob+=log(p1);
		}
		/* transmission probs */
		tp=seg_prob[k1];
		i=gt[X_MAT][sidx]-1;
		j=gt[X_PAT][sidx]-1;
		k=gt[X_MAT][didx]-1;
		l=gt[X_PAT][didx]-1;
		p1=(pp[X_MM_PM]=tp[X_MM_PM]*qval[i*n_all+k]);
		p1+=(pp[X_MM_PP]=tp[X_MM_PP]*qval[j*n_all+k]);
		p1+=(pp[X_MP_PM]=tp[X_MP_PM]*qval[i*n_all+l]);
		p1+=(pp[X_MP_PP]=tp[X_MP_PP]*qval[j*n_all+l]);
		if(p1<=0.0) ABT_FUNC("Internal error - no offspring combination possible\n");
		z=rans()*p1;
		p2=0.0;
		for(n=0;n<4;n++) {
			if(pp[n]>0.0) {
				p2+=pp[n];
				if(z<=p2) break;
			}
		}
		gt[X_PAT][k1]=(n&1)?j+1:i+1;
		gt[X_MAT][k1]=(n&2)?l+1:k+1;
		idflag[k1]|=(SAMPLED_MAT|SAMPLED_PAT);
	}
	return prob;
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "peel_singleton"
static double peel_singleton(const Ind *id,const struct Locus *locus,const double lam,const int sample_flag) 
{
	double p,z,z1;
	int f,i,j;
	
	f=locus->locus_type==MARKER_LOCUS?1:0;
	if(CHK_PEEL(TRACE_LEVEL_1)) {
		if(f) fputs("Marker locus",stdout);
		else fputs("Trait locus",stdout);
		if(sample_flag) printf(", singleton peel/sample: %d",id->self);
		else printf(", singleton peel: %d",id->self);
		fputc('\n',stdout);
	}
	if(f) p=get_par_probs(qval,id,locus,lam);
	else p=get_trait_par_probs(qval,id,locus,lam);
	if(sample_flag) {
		z=rans();
		for(z1=0.0,i=0;i<n_idx;i++) if(qval[i]>0.0) {
			z1+=qval[i];
			if(z1>=z) break;
		}
		j=id->index;
		if(f) {
 			locus->gt[X_MAT][j]=1+(i&((1<<n_bits)-1));
			locus->gt[X_PAT][j]=1+(i>>n_bits);
		} else {
 			locus->gt[X_MAT][j]=1+(i%n_all);
			locus->gt[X_PAT][j]=1+(i/n_all);
		}
		locus->seg[0][j]=locus->seg[1][j]=-1;
		idflag[j]|=(SAMPLED_PAT|SAMPLED_MAT);
	}
	if(CHK_PEEL(TRACE_LEVEL_2)) printf("Returning prob from %s() = %g\n",FUNC_NAME,log(p));
	return log(p);
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "loki_trait_simple_peelop"
/* Similar to loki_simple_peelop, but for a trait locus
 * Notes:
 *   1) Only R-Functions on single individuals can be used,
 *      although joint individuals on pairs can be produced.
 *   2) For pivot and pivot1, acceptable configurations are:
 *        0,0
 *        dam,0
 *        sire,0
 *        kid,0
 *        dam,sire
 *        dam,kid
 *        sire,kid
 *        kid,kid
 *      Others (such as sire,dam kid,parent or 0,dam) are illegal and NOT
 *      checked for! (the peeling sequence generation routines
 *      in loki_peel_init.c will not give such illegal sequences)
 */
static double loki_trait_simple_peelop(const struct Simple_Element *element,const struct Locus *locus,const double lam,const int s_flag)
{
	int i,j,k,l,n,of,of1,ip[4],k1,k2=-1,inv[4];
	double prob=0.0,*tp,*tp1,p1;
	Ind *sire,*dam,*kid,*pivot,*pivot1;
	Indlist *id_list;
	
	pivot=element->pivot;
	if(!pivot && s_flag) loki_trait_simple_sample(element,locus,lam);
	pivot1=element->pivot1;
	sire=element->sire;
	dam=element->dam;
	if(CHK_PEEL(TRACE_LEVEL_1)) {
		fputs("Trait locus, family peel: ",stdout);
		print_simple_element(stdout,element,rf);
		fputc('\n',stdout);
	}
 	if(CHK_PEEL(TRACE_LEVEL_3)) {
		if(pivot) {
			inv[0]=pivot->index+1;
			inv[1]=-1-pivot->index;
		}
		if(pivot1) {
			inv[2]=pivot1->index+1;
			inv[3]=-1-pivot1->index;
		}
	}
	if(sire!=pivot && sire!=pivot1) {
		p1=get_trait_par_probs(pval,sire,locus,lam);
		if(p1<=0.0) return -DBL_MAX;
		prob+=log(p1);
	} else check_for_trait_rf(pval,sire);
		
	if(dam!=pivot)	{
		p1=get_trait_par_probs(mval,dam,locus,lam);
		if(p1<=0.0) return -DBL_MAX;
		prob+=log(p1);
	} else check_for_trait_rf(mval,dam);
	/* Set up joint function on parents */
	for(n=i=0;i<n_idx;i++) {
		p1=pval[i];
		for(k=0;k<n_idx;k++) peel_famval[n++]=p1*mval[k];
	}
	id_list=element->off;
	while(id_list)	{
		kid=id_list->current;
		id_list=id_list->next;
		if(kid==pivot || kid==pivot1) continue;
		check_for_trait_rf(qval,kid);
		j=kid->index;
		if(locus->penetrance && qdex[j])	{
			locus->penetrance(qval,kid,locus,lam);
			p1=normalize(qval,n_idx);
			if(p1<=0.0) return -DBL_MAX;
			prob+=log(p1);
		}
		tp=seg_prob[j];
		for(n=j=0;j<n_all;j++) {
			for(i=0;i<n_all;i++) {
				for(l=0;l<n_all;l++)	{
					for(k=0;k<n_all;k++)	{
						peel_famval[n++]*=
						  tp[X_MM_PM]*qval[i*n_all+k]+
						  tp[X_MM_PP]*qval[j*n_all+k]+
						  tp[X_MP_PM]*qval[i*n_all+l]+
						  tp[X_MP_PP]*qval[j*n_all+l];
					}
				}
			}
		}
	}
	of=of1=0;
	if(pivot && pivot!=sire && pivot!=dam) {
		tp=seg_prob[pivot->index];
		check_for_trait_rf(pval,pivot);
		of=1;
	} else tp=0;
	if(pivot1 && pivot1!=sire)	{
		tp1=seg_prob[pivot1->index];
		check_for_trait_rf(mval,pivot1);
		of1=1;
	} else tp1=0;
	if(pivot1) {/* Two pivot individuals */
		i=n_idx*n_idx;
		if(!of1) for(j=0;j<i;j++) qval[j]=peel_famval[j];
		else {
			for(j=0;j<i;j++) qval[j]=0.0;
			for(n=j=0;j<n_all;j++) {
				for(i=0;i<n_all;i++)	{
					for(l=0;l<n_all;l++)	{
						for(k=0;k<n_all;k++)	{
							p1=peel_famval[n++];
							ip[0]=i*n_all+k;
							ip[1]=j*n_all+k;
							ip[2]=i*n_all+l;
							ip[3]=j*n_all+l;
							if(of) {/* Peeling to 2 kids */
								for(k1=0;k1<4;k1++) for(k2=0;k2<4;k2++)
								  qval[ip[k1]*n_idx+ip[k2]]+=p1*pval[ip[k2]]*mval[ip[k1]]*tp1[k1]*tp[k2];
							} else {
								if(pivot==dam) k2=l*n_all+k;
								else if(pivot==sire) k2=j*n_all+i;
								else ABT_FUNC("Internal error - illegal pivot \n");
								for(k1=0;k1<4;k1++) qval[ip[k1]*n_idx+k2]+=p1*mval[ip[k1]]*tp1[k1];
							}
						}
					}
				}
			}
		}
		i=n_idx*n_idx;
		p1=normalize(qval,i);
		if(p1<=0.0) return -DBL_MAX;
		prob+=log(p1);
		k=element->out_index;
		rf[k].n_ind=4;
		rf[k].n_terms=i;
		get_rf_memory(rf+k,i,TRT_MBLOCK);
		for(j=0;j<i;j++) rf[k].p[j]=qval[j];
	} else if(pivot) {/* Single pivot */
		for(j=0;j<n_idx;j++) qval[j]=0.0;
		for(n=j=0;j<n_all;j++) {
			for(i=0;i<n_all;i++) {
				for(l=0;l<n_all;l++)	{
					for(k=0;k<n_all;k++)	{
						if(dam==pivot) qval[l*n_all+k]+=peel_famval[n];
						else if(sire==pivot) qval[j*n_all+i]+=peel_famval[n];
						else {
							p1=peel_famval[n];
							qval[i*n_all+k]+=tp[X_MM_PM]*p1;
							qval[j*n_all+k]+=tp[X_MM_PP]*p1;
							qval[i*n_all+l]+=tp[X_MP_PM]*p1;
							qval[j*n_all+l]+=tp[X_MP_PP]*p1;
						}
						n++;
					}
				}
			}
		}
		if(of) for(j=0;j<n_idx;j++) qval[j]*=pval[j];
		p1=normalize(qval,n_idx);
		if(p1<=0.0) return -DBL_MAX;
		prob+=log(p1);
		k=element->out_index;
		rfp[pivot->index]=k+1;
		rf[k].n_ind=2;
		rf[k].n_terms=n_idx;
		get_rf_memory(rf+k,n_idx,TRT_MBLOCK);
		for(j=0;j<n_idx;j++) rf[k].p[j]=qval[j];
	} else {
		i=n_idx*n_idx;
		for(p1=0.0,j=0;j<i;j++) p1+=peel_famval[j];
		prob+=log(p1);
	}
	if(CHK_PEEL(TRACE_LEVEL_3) && pivot) print_rfunction(stdout,rf+element->out_index,inv,TRAIT_LOCUS);
	if(CHK_PEEL(TRACE_LEVEL_2)) printf("Returning prob from %s() = %g\n",FUNC_NAME,prob);
	return prob;
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "loki_simple_sample"
/* Sample a nuclear family */
static double loki_simple_sample(const struct Simple_Element *element,const struct Locus *locus,const double lam)
{
	int ids,idd,i,j,k,l,m,n,fsp=0,n_off,lump[2],*gt[2],kd;
	double *tp,p1,p2,z,prob=0.0,pp[4];
	Ind *sire,*dam,*kid;
	Indlist *id_list;
	unsigned long a,b,cm[2],*cm1[2];
		  
	gt[0]=locus->gt[0];
	gt[1]=locus->gt[1];
	sire=element->sire;
	ids=sire->index;
	dam=element->dam;
	idd=dam->index;
	if(CHK_PEEL(TRACE_LEVEL_1)) {
		fputs("Marker locus, family sample: ",stdout);
		print_simple_element(stdout,element,rf);
		fputc('\n',stdout);
	}
	p1=setup_parents(pval,sire,locus,lam,&i);
	if(p1<0.0) ABT_FUNC("Zero probability in sampling operation\n");
	prob+=log(p1);
	p1=setup_parents(mval,dam,locus,lam,&j);
	if(p1<0.0) ABT_FUNC("Zero probability in sampling operation\n");
	prob+=log(p1);
	if(i||j) { /* If both parents are not already sampled then sample genotypes for them */
		cm1[0]=peel_pt1;
		cm1[1]=cm1[0]+max_peel_off;
 		id_list=element->off;
		m=0;
		while(id_list) {
			kid=id_list->current;
			id_list=id_list->next;
			kd=kid->index;
			cm[0]=req_set[0][kd];
			cm[1]=req_set[1][kd];
			if(idflag[kd]&SAMPLED_MAT) {
				cm1[X_MAT][m]=1<<(gt[X_MAT][kd]-1);
				cm1[X_PAT][m]=1<<(gt[X_PAT][kd]-1);
			} else {
				cm1[X_PAT][m]=idtemp[X_PAT][kd];
				cm1[X_MAT][m]=idtemp[X_MAT][kd];
			}
			if(cm[X_MAT]&cm1[X_MAT][m]) cm1[X_MAT][m]|=cm[X_MAT];
			if(cm[X_PAT]&cm1[X_PAT][m]) cm1[X_PAT][m]|=cm[X_PAT];
			m++;
		}
		n_off=m;
		for(i=0;i<n_all;i++) { /* Construct set of possible parental genotype combinations */
			a=a_set[i][ids];
			if(a) for(j=0;j<n_all;j++) if(a&(1<<j)) {
				cm[X_PAT]=(1<<i)|(1<<j);
				for(m=0;m<n_off;m++) if(!(cm1[X_PAT][m]&cm[X_PAT])) break; /* Check compatability with sampled kid genotypes */
				if(m<n_off) continue;
				for(k=0;k<n_all;k++)	{
					b=a_set[k][idd];
					if(b) for(l=0;l<n_all;l++) if(b&(1<<l)) {
						cm[X_MAT]=(1<<k)|(1<<l);
						for(m=0;m<n_off;m++) if(!(cm1[X_MAT][m]&cm[X_MAT])) break;
						if(m==n_off) {
							peel_fs[fsp].pat_gene[X_MAT]=i;
							peel_fs[fsp].pat_gene[X_PAT]=j;
							peel_fs[fsp].mat_gene[X_MAT]=k;
							peel_fs[fsp].mat_gene[X_PAT]=l;
							peel_fs[fsp++].p=pval[(j<<n_bits)|i]*mval[(l<<n_bits)|k];
						}
					}
				}
			}
		}
		if(!fsp) return -DBL_MAX;
		id_list=element->off; /* Add in contributions from offspring */
		while(id_list) {
			kid=id_list->current;
			id_list=id_list->next;
			kd=kid->index;
			if(idflag[kd]&SAMPLED_MAT) set_for_sampled(qval,kd,locus);
			else {
				check_for_rf(qval,kid,locus->sparse_flag);
				if(locus->penetrance && qdex[j]) {
					locus->penetrance(qval,kid,locus,lam);
					p1=normalize(qval,n_idx);
					prob+=log(p1);
				}
			}
			tp=seg_prob[kd];
			for(k=0;k<2;k++) {
				cm[k]=req_set[k][kd];
				lump[k]=lumped[k][kd];
			}
			for(n=0;n<fsp;n++) {
				i=peel_fs[n].pat_gene[X_MAT];
				j=peel_fs[n].pat_gene[X_PAT];
				k=peel_fs[n].mat_gene[X_MAT];
				l=peel_fs[n].mat_gene[X_PAT];
				if((1<<i)&cm[X_PAT]) i=lump[X_PAT];
				if((1<<j)&cm[X_PAT]) j=lump[X_PAT];
				if((1<<k)&cm[X_MAT]) k=lump[X_MAT];
				if((1<<l)&cm[X_MAT]) l=lump[X_MAT];
				peel_fs[n].p*=
				  tp[X_MM_PM]*qval[(i<<n_bits)|k]+
				  tp[X_MM_PP]*qval[(j<<n_bits)|k]+
				  tp[X_MP_PM]*qval[(i<<n_bits)|l]+
				  tp[X_MP_PP]*qval[(j<<n_bits)|l];
			}
		}
		/* And sample a parental genotype combination */
		p1=0.0;
		for(n=0;n<fsp;n++) p1+=peel_fs[n].p;
		prob+=log(p1);
		z=rans()*p1;
		p2=0.0;
		for(n=0;n<fsp;n++) {
			if(peel_fs[n].p>0.0)	{
				p2+=peel_fs[n].p;
				if(p2>=z) {
					gt[X_MAT][ids]=peel_fs[n].pat_gene[X_MAT]+1;
					gt[X_PAT][ids]=peel_fs[n].pat_gene[X_PAT]+1;
					gt[X_MAT][idd]=peel_fs[n].mat_gene[X_MAT]+1;
					gt[X_PAT][idd]=peel_fs[n].mat_gene[X_PAT]+1;
					idflag[ids]|=(SAMPLED_MAT|SAMPLED_PAT);
					idflag[idd]|=(SAMPLED_MAT|SAMPLED_PAT);
					break;
				}
			}
		}
	}
	/* Now sample offspring conditional on parental genotypes (if not already sampled) */
	id_list=element->off;
	while(id_list) {
		kid=id_list->current;
		id_list=id_list->next;
		kd=kid->index;
		if(idflag[kd]&SAMPLED_MAT) continue;
		check_for_rf(qval,kid,locus->sparse_flag);
		if(idflag[kd]&SAMPLED_MAT) continue;
		if(locus->penetrance && qdex[j]) {
			locus->penetrance(qval,kid,locus,lam);
			p1=normalize(qval,n_idx);
			prob+=log(p1);
		}
		/* transmission probs */
		tp=seg_prob[kd];
		/* lumping info */
		for(k=0;k<2;k++) {
			cm[k]=req_set[k][kd];
			lump[k]=lumped[k][kd];
		}
		i=gt[X_MAT][ids]-1;
		j=gt[X_PAT][ids]-1;
		k=gt[X_MAT][idd]-1;
		l=gt[X_PAT][idd]-1;
		if((1<<i)&cm[X_PAT]) i=lump[X_PAT];
		if((1<<j)&cm[X_PAT]) j=lump[X_PAT];
		if((1<<k)&cm[X_MAT]) k=lump[X_MAT];
		if((1<<l)&cm[X_MAT]) l=lump[X_MAT];
		p1=(pp[X_MM_PM]=tp[X_MM_PM]*qval[(i<<n_bits)|k]);
		p1+=(pp[X_MM_PP]=tp[X_MM_PP]*qval[(j<<n_bits)|k]);
		p1+=(pp[X_MP_PM]=tp[X_MP_PM]*qval[(i<<n_bits)|l]);
		p1+=(pp[X_MP_PP]=tp[X_MP_PP]*qval[(j<<n_bits)|l]);
		if(p1<=0.0) ABT_FUNC("Internal error - no offspring combination possible\n");
		z=rans()*p1;
		p2=0.0;
		for(n=0;n<4;n++) {
			if(pp[n]>0.0) {
				p2+=pp[n];
				if(z<=p2) break;
			}
		}
		gt[X_PAT][kd]=(n&1)?j+1:i+1;
		gt[X_MAT][kd]=(n&2)?l+1:k+1;
		idflag[kd]|=(SAMPLED_MAT|SAMPLED_PAT);
	}
	return prob;
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "loki_simple_peelop"
/* Performs simple (i.e., nuclear family based) peeling operation */
static double loki_simple_peelop(const struct Simple_Element *element,const struct Locus *locus,const double lam,const int s_flag)
{
	int i,ids,idd,j,k,l,n,n1,fsp=0,gt[4],of,of1,lump[2],lump1[2];
	int i1,j1,k1,k2=-1,l1,ip[4],ip1[4],pidx=-1,pidx1,inv[4];
	double prob=0.0,*tp,*tp1,p,p1;
	unsigned long a,b,cm[2],cm1[2];
	Ind *pivot,*pivot1,*sire,*dam,*kid;
	Indlist *id_list;
	
	tp=tp1=0;
	pivot=element->pivot;
	if(!pivot && s_flag) return loki_simple_sample(element,locus,lam);
	pivot1=element->pivot1;
	sire=element->sire;
	ids=sire->index;
	dam=element->dam;
	idd=dam->index;
	if(CHK_PEEL(TRACE_LEVEL_1)) {
		fputs("Marker locus, family peel: ",stdout);
		print_simple_element(stdout,element,rf);
		fputc('\n',stdout);
	}
 	if(CHK_PEEL(TRACE_LEVEL_3)) {
		if(pivot) {
			inv[0]=pivot->index+1;
			inv[1]=-1-pivot->index;
		}
		if(pivot1) {
			inv[2]=pivot1->index+1;
			inv[3]=-1-pivot1->index;
		}
	}
	if(sire!=pivot && sire!=pivot1) {
		p1=get_par_probs(pval,sire,locus,lam);
		if(p1<0.0) return DBL_MAX;
		prob+=log(p1);
	} else check_for_rf(pval,sire,locus->sparse_flag);
	if(dam!=pivot && dam!=pivot1) {
		p1=get_par_probs(mval,dam,locus,lam);
		if(p1<0.0) return DBL_MAX;
		prob+=log(p1);
	} else check_for_rf(mval,dam,locus->sparse_flag);
	/* Construct set of possible parental genotype combinations */
	for(i=0;i<n_all;i++)	{
		a=a_set[i][ids];
		if(a) for(j=0;j<n_all;j++) if(a&(1<<j)) {
			p1=pval[(j<<n_bits)|i];
			for(k=0;k<n_all;k++)	{
				b=a_set[k][idd];
				if(b) for(l=0;l<n_all;l++) if(b&(1<<l)) {
					peel_fs[fsp].pat_gene[X_MAT]=i;
					peel_fs[fsp].pat_gene[X_PAT]=j;
					peel_fs[fsp].mat_gene[X_MAT]=k;
					peel_fs[fsp].mat_gene[X_PAT]=l;
					peel_fs[fsp++].p=p1*mval[(l<<n_bits)|k];
				}
			}
		}
	}
	/* Add contributions from non-pivot offspring */
	id_list=element->off;
	while(id_list) {
		kid=id_list->current;
		id_list=id_list->next;
		if(kid==pivot || kid==pivot1) continue;
		check_for_rf(qval,kid,locus->sparse_flag);
		j=kid->index;
		if(locus->penetrance && qdex[j]) {
			locus->penetrance(qval,kid,locus,lam);
			p1=normalize(qval,n_idx);
			prob+=log(p1);
		}
		tp=seg_prob[j];
		for(k=0;k<2;k++) {
			cm[k]=req_set[k][j];
			lump[k]=lumped[k][j];
		}
		for(n=0;n<fsp;n++) {
			if(peel_fs[n].p<=0.0) continue;
			i=peel_fs[n].pat_gene[X_MAT];
			j=peel_fs[n].pat_gene[X_PAT];
			k=peel_fs[n].mat_gene[X_MAT];
			l=peel_fs[n].mat_gene[X_PAT];
			if((1<<i)&cm[X_PAT]) i=lump[X_PAT];
			if((1<<j)&cm[X_PAT]) j=lump[X_PAT];
			if((1<<k)&cm[X_MAT]) k=lump[X_MAT];
			if((1<<l)&cm[X_MAT]) l=lump[X_MAT];
			peel_fs[n].p*=
			  tp[X_MM_PM]*qval[(i<<n_bits)|k]+
			  tp[X_MM_PP]*qval[(j<<n_bits)|k]+
			  tp[X_MP_PM]*qval[(i<<n_bits)|l]+
			  tp[X_MP_PP]*qval[(j<<n_bits)|l];
		}
	}
	/* If pivot is an offspring, bring in previous R-Function and zero out
	 * illegal genotypes */
	of=of1=0;
	if(pivot) {
		pidx=pivot->index;
		for(k=0;k<2;k++) {
			cm[k]=req_set[k][pidx];
			lump[k]=lumped[k][pidx];
		}
		if(pivot!=sire && pivot!=dam) {
			tp=seg_prob[pidx];
			check_for_rf(mval,pivot,locus->sparse_flag);
			of=1;
		}
	}
	if(pivot1) {
		pidx1=pivot1->index;
		for(k=0;k<2;k++) {
			cm1[k]=req_set[k][pidx1];
			lump1[k]=lumped[k][pidx1];
		}
		if(pivot1!=sire) {
			tp1=seg_prob[pidx1];
			check_for_rf(pval,pivot1,locus->sparse_flag);
			of1=1;
		}
	}
	if(pivot1) { /* Two pivot individuals */
		if(!of1) { /* Peeling to both parents */
 			for(p1=0.0,n1=n=0;n<fsp;n++) {
				if(peel_fs[n].p>0.0) {
					p1+=peel_fs[n].p;
					n1++;
				}
			}
			prob+=log(p1);
			k=element->out_index;
			rf[k].n_ind=4;
			if(locus->sparse_flag) {
				rf[k].n_terms=n1;
				get_rf_memory(rf+k,n1,MRK_MBLOCK);
				for(n1=n=0;n<fsp;n++) {
					if(peel_fs[n].p>0.0) {
						gt[0]=peel_fs[n].mat_gene[X_MAT]+1;
						gt[1]=peel_fs[n].mat_gene[X_PAT]+1;
						gt[2]=peel_fs[n].pat_gene[X_MAT]+1;
						gt[3]=peel_fs[n].pat_gene[X_PAT]+1;
						rf[k].index[n1]=get_index1(4,gt);
						rf[k].p[n1++]=peel_fs[n].p/p1;
					}
				}
			} else {
				i=n_idx*n_idx;
				rf[k].n_terms=i;
				get_rf_memory(rf+k,i,TRT_MBLOCK);
				for(n=0;n<i;n++) rf[k].p[n]=0.0;
				for(n=0;n<fsp;n++) {
					if(peel_fs[n].p>0.0) {
						gt[0]=peel_fs[n].mat_gene[X_MAT]+1;
						gt[1]=peel_fs[n].mat_gene[X_PAT]+1;
						gt[2]=peel_fs[n].pat_gene[X_MAT]+1;
						gt[3]=peel_fs[n].pat_gene[X_PAT]+1;
						for(n1=0,j=3;j>=0;j--) n1=n1*n_all+gt[j]-1;
						rf[k].p[n1]=peel_fs[n].p/p1;
					}
				}
			}
		} else {
			for(n=0;n<n_idx*n_idx;n++) qval[n]=0.0;
			for(n1=n=0;n<fsp;n++) {
				if((p=peel_fs[n].p)<=0.0) continue;
				i=peel_fs[n].pat_gene[X_MAT];
				j=peel_fs[n].pat_gene[X_PAT];
				k=peel_fs[n].mat_gene[X_MAT];
				l=peel_fs[n].mat_gene[X_PAT];
				if((1<<i)&cm1[X_PAT]) i1=lump1[X_PAT];
				else i1=i;
				if((1<<j)&cm1[X_PAT]) j1=lump1[X_PAT];
				else j1=j;
				if((1<<k)&cm1[X_MAT]) k1=lump1[X_MAT];
				else k1=k;
				if((1<<l)&cm1[X_MAT]) l1=lump1[X_MAT];
				else l1=l;
				if((1<<i)&cm[X_PAT]) i=lump[X_PAT];
				if((1<<j)&cm[X_PAT]) j=lump[X_PAT];
				if((1<<k)&cm[X_MAT]) k=lump[X_MAT];
				if((1<<l)&cm[X_MAT]) l=lump[X_MAT];
				ip1[0]=(i1<<n_bits)|k1;
				ip1[1]=(j1<<n_bits)|k1;
				ip1[2]=(i1<<n_bits)|l1;
				ip1[3]=(j1<<n_bits)|l1;
				if(of) { /* Peeling to 2 kids */
					ip[0]=(i<<n_bits)|k;
					ip[1]=(j<<n_bits)|k;
					ip[2]=(i<<n_bits)|l;
					ip[3]=(j<<n_bits)|l;
					for(k1=0;k1<4;k1++) for(k2=0;k2<4;k2++) {
						p1=p*tp1[k1]*pval[ip1[k1]]*tp[k2]*mval[ip[k2]];
						if(p1>0.0) {
							qval[ip1[k1]*n_idx+ip[k2]]+=p1;
						}
					}
				} else { /* Peeling to 1 parent and 1 kid */
					if(pivot==dam) k2=(l<<n_bits)|k;
					else if(pivot==sire) k2=(j<<n_bits)|i;
					else ABT_FUNC("Internal error - illegal pivot\n");
					for(k1=0;k1<4;k1++) {
						p1=p*pval[ip1[k1]]*tp1[k1];
						if(p1>0.0) {
							qval[ip1[k1]*n_idx+k2]+=p1;
						}
					}
				}
			}
			for(p1=0.0,n1=n=0;n<n_idx*n_idx;n++) {
				p=qval[n];
				if(p>0.0) {
					p1+=p;
					n1++;
				}
			}
			prob+=log(p1);
			k=element->out_index;
			rf[k].n_ind=4;
			if(locus->sparse_flag) {
				rf[k].n_terms=n1;
				get_rf_memory(rf+k,n1,MRK_MBLOCK);
				for(n1=n=0;n<n_idx*n_idx;n++) {
					p=qval[n];
					if(p>0.0) {
						rf[k].index[n1]=n;
					rf[k].p[n1++]=p/p1;
					}
				}
			} else {
				n1=n_idx*n_idx;
				rf[k].n_terms=n1;
				get_rf_memory(rf+k,n1,TRT_MBLOCK);
				for(n=0;n<n1;n++) rf[k].p[n]=qval[n]/p1;
			}
		}
   } else if(pivot) { /* Single pivot */
		for(k=0;k<n_idx;k++) qval[k]=0.0;
		p1=0.0;
		for(n=0;n<fsp;n++) {
			if((p1=peel_fs[n].p)<=0.0) continue;
			i=peel_fs[n].pat_gene[X_MAT];
			j=peel_fs[n].pat_gene[X_PAT];
			k=peel_fs[n].mat_gene[X_MAT];
			l=peel_fs[n].mat_gene[X_PAT];
			if(dam==pivot) qval[(l<<n_bits)|k]+=p1;
			else if(sire==pivot) qval[(j<<n_bits)|i]+=p1;
			else {
				if((1<<i)&cm[X_PAT]) i=lump[X_PAT];
				if((1<<j)&cm[X_PAT]) j=lump[X_PAT];
				if((1<<k)&cm[X_MAT]) k=lump[X_MAT];
				if((1<<l)&cm[X_MAT]) l=lump[X_MAT];
				qval[(i<<n_bits)|k]+=tp[X_MM_PM]*p1;
				qval[(j<<n_bits)|k]+=tp[X_MM_PP]*p1;
				qval[(i<<n_bits)|l]+=tp[X_MP_PM]*p1;
				qval[(j<<n_bits)|l]+=tp[X_MP_PP]*p1;
			}
		}
		for(p1=0.0,n1=n=0;n<n_idx;n++) {
			p=qval[n];
			if(p>0.0) {
				n1++;
				p1+=qval[n];
			}
		}
		prob+=log(p1);
		k=element->out_index;
		rfp[pidx]=k+1;
		rf[k].n_ind=2;
		if(locus->sparse_flag) {
			rf[k].n_terms=n1;
			get_rf_memory(rf+k,n1,MRK_MBLOCK);
			for(n1=n=0;n<n_idx;n++) {
				p=qval[n];
				if(p>0.0) {
					rf[k].index[n1]=n;
					rf[k].p[n1++]=p/p1;
				}
			}
		} else {
			n1=n_idx*n_idx;
			rf[k].n_terms=n1;
			get_rf_memory(rf+k,n1,TRT_MBLOCK);
			for(n=0;n<n_idx;n++) rf[k].p[n]=qval[n]/p1;
		}
	} else {
		for(p1=0.0,n=0;n<fsp;n++) {
			if(peel_fs[n].p>0.0) {
				p1+=peel_fs[n].p;
			}
		}
		prob+=log(p1);
	}
	if(CHK_PEEL(TRACE_LEVEL_3) && pivot) print_rfunction(stdout,rf+element->out_index,inv,locus->sparse_flag?MARKER_LOCUS:TRAIT_LOCUS);
	if(CHK_PEEL(TRACE_LEVEL_2)) printf("Returning prob from %s() = %g\n",FUNC_NAME,prob);
	return prob;
}

static void get_seg(int i,int j,int *gt[2],int *seg[2],int x_flag,struct Locus *locus)
{
	int k,k1,l_flag;
	double trn[2],*tp,z;
	unsigned long a,b,c;
	
	l_flag=locus->locus_type;
	if(j!=INVALID_INDEX)	{
		if(n_all<2) {
			gt[x_flag][i]=1;
			trn[0]=trn[1]=1;
		} else {
			k=gt[x_flag][i];
			if(k<1 || k>n_all) {
				if(l_flag==MARKER_LOCUS) fprintf(stderr,"Marker locus %d",locus->index+1);
				else fprintf(stderr,"Trait locus %d",locus->index+1-n_markers);
				fprintf(stderr,", individual %d, genotypes %d %d\n",nghd[i].self,gt[0][i],gt[1][i]);
				ABT_FUNC("Internal error - bad sampled genotype\n");
			}
			if(l_flag==MARKER_LOCUS) {
				c=req_set[x_flag][i];
				a=1<<(k-1);
				if(a&c) a=c;
				for(k1=0;k1<2;k1++) {
					k=gt[k1][j]-1;
					b=1<<k;
					trn[k1]=(a&b)?1:0;
				}
			} else for(k1=0;k1<2;k1++) trn[k1]=(k==gt[k1][j])?1:0;
			if(!trn[0] && !trn[1]) {
				ABT_FUNC("Internal error - bad configuration?\n");
			}
		}
		if(trn[0] && trn[1])	{
		  if(seg_mode) seg[x_flag][i]=-2;
		  else {
		    tp=seg_prob1[i]+x_flag*2;
		    z=rans()*(tp[0]+tp[1]);
		    seg[x_flag][i]=(z<=tp[0])?0:1;
		  }
		} else seg[x_flag][i]=trn[0]?0:1;
	} else seg[x_flag][i]= -1;
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "loki_mjrpeel"
double loki_mjrpeel(double *ym,double ve,double lam,int sample_flag,struct Locus *locus)
{
	struct Peelseq_Head *pp;
	struct Simple_Element *simple_em;
   struct Complex_Element *complex_em; 
	int i,k,k1,n_rf,n_peel_ops,*gt[2],k1a,k2,k3,k2a;
	int *seg[2];
   double like=0.0,z;
	unsigned long a,b,b1;
	
	memset(idflag,0,sizeof(int)*pedsize);
	memset(rfp,0,sizeof(int)*pedsize);
	res=ym;
	sige=ve;
	if(locus->penetrance) konst=1.0/sqrt(PPI*sige);
	peelseq_head=locus->peel;
	for(i=0;i<2;i++) {
		gt[i]=locus->gt[i];
		seg[i]=locus->seg[i];
	}
	nn_all=locus->n_all;
	if(locus->locus_type==MARKER_LOCUS) {
		a_set=locus->all_set;
		for(i=0;i<3;i++) req_set[i]=locus->req_set[i];
		for(i=0;i<pedsize;i++) {
			for(k1=0;k1<2;k1++) {
				a=req_set[k1][i];
				if(a) {
					for(k=0;k<nn_all;k++) if(a&(1<<k)) break;
					lumped[k1][i]=k;
				} else lumped[k1][i]=0;
			}
			ngens[i]=nhaps[0][0]=nhaps[1][i]=0;
			k1=k1a=k2=k3=0;
			for(b=b1=0,k=0;k<nn_all;k++) {
				a=a_set[k][i];
				if(!a) continue;
				k1++;
				k1a=k;
				b|=a;
				b1|=(1<<k);
				while(a) {
					if(a&1) k3++;
					a>>=1;
				}
			}
			nhaps[X_MAT][i]=k1;
			if(k1==1) gt[X_MAT][i]=k1a+1;
			idtemp[X_PAT][i]=b;
			idtemp[X_MAT][i]=b1;
			k1a=k2a=0;
			while(b) {
				if(b&1) {
					k2++;
					k2a=k1a;
				}k1a++;
				b>>=1;
			}
			nhaps[X_PAT][i]=k2;
			if(k2==1) gt[X_PAT][i]=k2a+1;
			ngens[i]=k3;
		}
	} else for(i=0;i<nn_all;i++) {
		freq[i]=locus->freq[i];
		peel_alls[i]=i;
	}
	for(comp=0;comp<n_compnts;comp++) {
		for(i=0;i<2;i++) {
			mem_block[i]=first_mem_block[i];
			mem_block[i]->ptr=0;
		}
		n_all=locus->n_alls[comp];
		if(n_all<2) continue;
		if(locus->locus_type==MARKER_LOCUS) {
			allele_trans=locus->allele_trans[comp];
			for(i=0;i<nn_all;i++) {
				freq[i]=0.0;
				peel_alls[i]=n_all-1;
			}
			for(i=0;i<nn_all;i++) if((k1=allele_trans[i])>=0) peel_alls[k1]=i;
			for(i=0;i<nn_all;i++) freq[peel_alls[i]]+=locus->freq[i];
			n_bits=num_bits(n_all);
		}
		if(locus->sparse_flag) n_idx=1<<(n_bits+n_bits);
		else n_idx=n_all*n_all;
		n_peel_ops=0;
		n_rf= -1;
		rf=locus->r_func[comp];
		n_bits=num_bits(n_all);
		sample_flag&=~OP_SAMPLING;
		for(i=0;i<n_singles;i++) if(singles[i]==comp) {
			if(cmplist[comp+1].people->next) 
			  like+=peel_singleton(cmplist[comp+1].people->next->current,locus,lam,sample_flag);
			break;
		}
		/* Go through all operations in peeling sequence */
		pp=peelseq_head+comp;
		while(pp->type) {
			if(pp->type==PEEL_SIMPLE) {
				simple_em=pp->ptr.simple;
				if(sample_flag && simple_em->pivot) peel_list[n_peel_ops++]=pp;
				if(locus->locus_type==TRAIT_LOCUS) {
					if(simple_em->out_index>n_rf)	{
						n_rf=simple_em->out_index;
						rf[n_rf].p=0;
					}
					z=loki_trait_simple_peelop(simple_em,locus,lam,sample_flag);
					if(z== -DBL_MAX) {
						if(sample_flag&1) ABT_FUNC("Zero probability in peeling operation\n");
						/* Error return on likelihood (not sampling) run.  Clean up and return error */
						return z;
					}
					if(CHK_PEEL(TRACE_LEVEL_2)) printf("current like = %g + %g = %g\n",like,z,like+z);
					like+=z;
				} else {
					z=loki_simple_peelop(simple_em,locus,lam,sample_flag);
 					if(z== -DBL_MAX) {
						if(sample_flag&1) ABT_FUNC("Zero probability in peeling operation\n");
						/* Error return on likelihood (not sampling) run.  Clean up and return error */
						return z;
					}
					if(CHK_PEEL(TRACE_LEVEL_2)) printf("current like = %g + %g = %g\n",like,z,like+z);
					like+=z;
				}
				pp= &simple_em->next;
			} else {
				complex_em=pp->ptr.complex;
				if(sample_flag && (complex_em->n_involved-complex_em->n_peel)) peel_list[n_peel_ops++]=pp;
				if(locus->locus_type==TRAIT_LOCUS) {
					if(complex_em->out_index>n_rf) {
						n_rf=complex_em->out_index;
						rf[n_rf].p=0;
					}
					z=loki_trait_complex_peelop(complex_em,locus,lam,sample_flag);
 					if(z== -DBL_MAX) {
						if(sample_flag&1) ABT_FUNC("Zero probability in peeling operation\n");
						/* Error return on likelihood (not sampling) run.  Clean up and return error */
						return z;
					}
					if(CHK_PEEL(TRACE_LEVEL_2)) printf("current like = %g + %g = %g\n",like,z,like+z);
					like+=z;
				} else {
					z=loki_complex_peelop(complex_em,locus,lam,sample_flag);
 					if(z== -DBL_MAX) {
						if(sample_flag&1) ABT_FUNC("Zero probability in peeling operation\n");
						/* Error return on likelihood (not sampling) run.  Clean up and return error */
						return z;
					}
					if(CHK_PEEL(TRACE_LEVEL_2)) printf("current like = %g + %g = %g\n",like,z,like+z);
					like+=z;
				}
				pp= &complex_em->next;
			}
		}
		if(CHK_PEEL(TRACE_LEVEL_1)) printf("At component %d, like = %g\n",comp+1,like);
		/* If we're sampling, do them again in reverse */
		if(sample_flag) {
			sample_flag|=OP_SAMPLING;
			for(i=n_peel_ops-1;i>=0;i--) {
				pp=peel_list[i];
				if(pp->type==PEEL_SIMPLE) {
					simple_em=pp->ptr.simple;
					if(locus->locus_type==TRAIT_LOCUS) (void)loki_trait_simple_sample(simple_em,locus,lam);
					else (void)loki_simple_sample(simple_em,locus,lam);
				} else {
					complex_em=pp->ptr.complex;	
					if(locus->locus_type==TRAIT_LOCUS) (void)loki_trait_complex_peelop(complex_em,locus,lam,sample_flag);
					else (void)loki_complex_peelop(complex_em,locus,lam,sample_flag);
				}
			}
		}
	}
	/* If sampling, get new segregation pattern */
	if(sample_flag) {
		for(k=0,comp=0;comp<n_compnts;comp++) {
			for(i=0;i<cp_size[comp];i++) {
				k1=cplist[k++];
				n_all=locus->n_alls[comp];
				get_seg(k1,nghd[k1].ma->index,gt,seg,X_MAT,locus);
				get_seg(k1,nghd[k1].pa->index,gt,seg,X_PAT,locus);
			}
		}
	}
	return like;
}
