/*  Full log-probability of switches and data for a given pedigree component */
/*    --used to track mixing and provide diagnostics                         */

#include "msample.h"

double full_seg_prob (int nl, int c, int h, int recount[NLOC][10])
	             /*  Compute the full log-prob for given set of switches 
		         on a given component c of the pedigree     */
{
    int   i, jj, jl, jr;
    Indlist *iptr;
    double pp = 0.0;
    int   new_count[NLOC][10];

    for (jj = 0; jj < nl; ++jj)
	for (i = 0; i < 10; ++i)
	    new_count[jj][i] = 0;

    iptr = cmplist[c].people;
    iptr = iptr->next;
    while (iptr != NULL)
    {
	i = iptr->current->index;
	for (jj = 0; jj < nl; ++jj)
	{
	    if (xx0[jj][i] != -1)
	    {
		++new_count[jj][3];	/* total female count */
		jr = chromo[h][jj + 1];
		jl = chromo[h][jj];
		if (xx0[jl][i] == xx0[jr][i])
		    ++new_count[jj][5];	/*non-rec */
		else
		    ++new_count[jj][4];	/* recom  */
	    }
	}
	for (jj = 0; jj < nl; ++jj)
	{
	    if (xx1[jj][i] != -1)
	    {
		++new_count[jj][0];	/* total male count */
		jr = chromo[h][jj + 1];
		jl = chromo[h][jj];
		if (xx1[jl][i] == xx1[jr][i])
		    ++new_count[jj][2];
		/* non-rec */
		else
		    ++new_count[jj][1];	/* recom  */
	    }
	}
	iptr = iptr->next;
    }

    for (jj = 0; jj < nl; ++jj)
    {
	pp += new_count[jj][5] * log (1.0 - xecomo[2][h][jj]);
	pp += new_count[jj][4] * log (xecomo[2][h][jj]);
	pp += new_count[jj][2] * log (1.0 - xecomo[1][h][jj]);
	pp += new_count[jj][1] * log (xecomo[1][h][jj]);
    }

    for (jj = 0; jj < nl; ++jj)
	for (i = 0; i < 6; ++i)
	    recount[jj][i] += new_count[jj][i];

    return (pp);
}

double lik_rat (int nl, int ip, int sx, int h1, int h2)
  /* likelihood ratio for a single meiosis individual ip, sex sx,
     between two haplotype sets, h1, h2   
     counting only those intervals where the haplotypes differ */
{
    int   jj, jr, jl, ix;
    double pp;
    int **xx;
    if (sx == 0)
    {
	xx = xx0;
	ix = 2;
    }
    else
    {
	xx = xx1;
	ix = 1;
    }
    pp = 1.0;
    for (jj = 0; jj < nl; ++jj)
    {
	if ((chromo[h1][jj] != chromo[h2][jj]) ||
	    (chromo[h1][jj + 1] != chromo[h2][jj + 1]) ||
	    (xecomo[ix][h1][jj] != xecomo[ix][h2][jj]))
	{
	    jr = chromo[h1][jj + 1];
	    jl = chromo[h1][jj];
	    if (xx[jl][ip] == xx[jr][ip])
		pp *= (1.0 - xecomo[ix][h1][jj]);
	    else
		pp *= (xecomo[ix][h1][jj]);
	    jr = chromo[h2][jj + 1];
	    jl = chromo[h2][jj];
	    if (xx[jl][ip] == xx[jr][ip])
		pp /= (1.0 - xecomo[ix][h2][jj]);
	    else
		pp /= (xecomo[ix][h2][jj]);
	}
    }
    return (pp);
}

double full_pen_seg_lod (int nl, int c, int ngenes, int err, int ***y,
			 int recount[NLOC][10])
			      /* contribution of the data to log P(Y|S)  */
			      /* on pedigree component c                 */
{
    int   j;
    double psegp, ptot;
    psegp = 0.0;
    ptot = 0.0;
    for (j = 0; j <= nl; ++j)
    {
	psegp = seg_pen2 (c, y, Qq[j], j, ngenes, &err);
	if (!err)
	    ptot += log (psegp);
	else
	{
	    printf ("\n seg_pen error in full_seg_prob");
	    exit (3);
	}
    }

    ptot += log (full_seg_prob (nl, c, 0, recount));

    return (ptot);
}
