#include "msample.h"

/* set_chromosomes(), print_chromosome() 	*/

void  set_chromosomes (int nl, double *map, int nh, int *l0, double *r0,
		       int ix)
/* routine to set the array of chromosome recombinations and orders for
   simulations and likelihood evaluations;
   nl = number of marker loci, map = recombination fractions between;
   nh = number of alternative positions for disease -- 0-one is simulation one.
   ix = sex of map 0=avg, 1=male, 2=female
   (r0,l0) vector of pairs giving position of disease; 
   l0 is marker locus to the left 
   and recombination r0 is to that marker; note r0 <= r(l0).
   (if disease extreme left, l0=0, and r0 is recom to locus 1), 
 */
{
    int   i, l;

    for (i = 0; i <= nh; ++i)		/* set chromosome i */
    {
	if (l0[i] == 0)
	{                              	/* for trait first on chromosome*/
	    chromo[i][0] = 0;
	    xecomo[ix][i][0] = r0[i];
	    revchro[i][0] = 0;
	    for (l = 1; l <= nl; ++l)
	    {
		chromo[i][l] = l;
		xecomo[ix][i][l] = map[l];
		revchro[i][l] = l;
	    }
	}
	else if (l0[i] == nl)
	{				/* for trait last on chromosome*/
	    for (l = 1; l <= nl; ++l)
	    {
		chromo[i][l - 1] = l;
		xecomo[ix][i][l - 1] = map[l];
		revchro[i][l] = l - 1;
	    }
	    chromo[i][nl] = 0;
	    xecomo[ix][i][nl - 1] = r0[i];
	    revchro[i][0] = nl;
	}
	else				/* for trait between marker loci*/
	{                          	/* on chromosome		*/
	    for (l = 0; l < l0[i] - 1; ++l)
	    {
		chromo[i][l] = l + 1;
		xecomo[ix][i][l] = map[l + 1];
		revchro[i][l + 1] = l;
	    }
	    l = l0[i] - 1;
	    chromo[i][l] = l + 1;
	    xecomo[ix][i][l] = r0[i];
	    revchro[i][l + 1] = l;
	    ++l;
	    chromo[i][l] = 0;
	    revchro[i][0] = l;
	    xecomo[ix][i][l] = (map[l] - r0[i]) / (1.0 - 2.0 * r0[i]);

	    for (l = l0[i] + 1; l <= nl; ++l)
	    {
		chromo[i][l] = l;
		xecomo[ix][i][l] = map[l];
		revchro[i][l] = l;
	    }
	}
    }
}

/*
   Routine to set the sex-specific marker maps, and the trait locus 
   position for jnt_auto, and lm_auto programs; EAT March 2000   
 */

void  print_chromosome (int nl, int hh, int ix)
{
    int   j;
    
    for (j = 0; j < nl; ++j)
	printf ("  %d %6.3f", chromo[hh][j], xecomo[ix][hh][j]);
    printf ("  %d\n", chromo[hh][nl]);
    fflush (stdout);

    return;
}

/* set_maps removed       11/00  MAJ */
