
#include "parsedrop.h"

/*
 * proc_drop() has as its input the results of parsing the parameter
 * statements by proc_pars() for genedrop.  It checks this data for 
 * consistency and completeness and, if there are no errors, builds 
 * data structures used by genedrop.
 *							  MAJ Jun 1999 
 */

int proc_drop(void)
{
    int entry, entry2, sex;
    int lnkd, m_no, loc_no, t_no;
    int max_alle;
    SIM_PTR     p_s;
    DIST_PTR    p_d; 	
    UNLNKD_PTR  p_u;
    TLIST_PTR   p_t;
    PRNTMAP_PTR p_m;
    PRNTLOC_PTR p_l;
    int sxsp_maps = FALSE;
 

    if (!test_sim())			/* checks for simulation	*/
        if (proc_err())		        /* statement(s)			*/
            return FALSE;

    if (n_err)
        return proc_end();
		                        /* for each simulated chrom,	*/
    entry = FALSE;			/* checks marker map & trait map*/
    while (p_s = nxt_sim_chrom(&entry))
    {
        if (!test_mark_map(p_s, &sxsp_maps))
            if (proc_err())
                return FALSE;
        if (!test_trait_map(p_s, &sxsp_maps))	
            if (proc_err()) 	  
                return FALSE;
    }
		                        /* for each simulated chrom,	*/
					/* for each trait,              */
    entry = FALSE;			/* checks for duplications	*/
    lnkd = TRUE;		
    while (p_s = nxt_sim_chrom(&entry)) 
    {                             
        entry2 = FALSE;   
        while (p_d = nxt_mapd_trait(&entry2, AVG, p_s))
            if (!test_trait_list(lnkd, p_d->id))
                if (proc_err()) 
                    return FALSE;
    }
 			                /* for each unlinked trait,	*/
    entry = FALSE;			/* checks for duplications	*/
    lnkd = FALSE;	
    while (p_u = nxt_unlnkd_trait(&entry))     
        if (!test_trait_list(lnkd, p_u->id))
            if (proc_err())   
                return FALSE;
			                /* for each chromosome,		*/
					/* for each marker,		*/
    entry = FALSE;			/* checks for allele frequencies*/
    while (p_s = nxt_sim_chrom(&entry))   
    {
        entry2 = FALSE;	
        while (m_no = nxt_mark(&entry2, p_s))  
            if (!test_mark_loci(p_s, m_no))  
                if (proc_err())   
                    return FALSE;
    }
					/* for each trait,		*/
				        /* checks variances, genotype	*/
    entry = FALSE;			/* means and allele frequencies	*/
    while ( p_t = nxt_trait(&entry))
    {             
        if (!test_trait_vars(p_t->id))
            if (proc_err())
                return FALSE;
        if (!test_trait_genotyp(p_t->id)) 
            if (proc_err())
                return FALSE;
        if (!test_trait_loci(p_t->id))
            if (proc_err())
                return FALSE;
    }	
    if (n_err)
        return proc_end();
		                        /* for each simulated chrom,	*/
				        /* if no errors, prints out 	*/
                			/* simulation parameters	*/
      
		                  	/* for each chromosome, for each*/
    				        /* sex, puts markers into 'print*/
                                        /* map'                         */
                                        /* maps not present default to  */
    entry = FALSE;		        /* sex-averaged map, then F map */

    while (p_s = nxt_sim_chrom(&entry))
    {
        for (sex = FEM; sex <= AVG; ++sex)
        {                               /* adds marker to map for sex   */
            p_m = add_prntmap(p_s, sex);           
            if (p_s->no_mark == 1)	
                add_only_mark_locus(p_m);
            else
            {
                entry2 = FALSE;                 
                while (p_d = nxt_mapd_mark(&entry2, sex, p_s))  
                    add_mark_locus(p_m, p_d);              
            }
                  		        /* adds traits to map for sex   */
            entry2 = FALSE;	
            while (p_d = nxt_mapd_trait(&entry2, sex, p_s)) 
                   add_trait_locus(p_m, p_d);

			                /* finds distances between 	*/
            entry2 = FALSE; 	        /* adjacent loci		*/ 
            while (p_l = nxt_prntmap_locus(&entry2, p_m)) 
                fix_locus_dist(p_m, p_l);             
     
            if ((sxsp_maps && (sex == FEM || sex == MAL)) 
                      || (!sxsp_maps && sex == AVG)) 
            {
                prnt_map(p_m, sex);     /* prints the map		*/

                entry2 = FALSE;	        /* checks trait distances 	*/
                while (p_l = nxt_prntmap_locus(&entry2, p_m)) 
                    if(!test_trait_dist(p_m, p_l))
                        if (proc_err())
                            return FALSE;
            }
        }
        prnt_mark_freqs(p_s); 	        /* prints marker parameters	*/

        prnt_trait_freqs(p_m);  	/* prints trait parameters	*/ 
        prnt_trait_pars(p_m);  	
    }
    prnt_unlnkd_freqs();
    prnt_unlnkd_pars();
    fprintf(stdout,"\n");

    if (n_err)
        return proc_end();

    NLoci = 0;			        /* counts loci and allocates 	*/
    NMarks = 0;                         /* recombination table          */
    NTraits = 0;
    NUnlnkd = 0;
    entry = FALSE; 
    while (p_m = nxt_prntmap(&entry, AVG)) 
    {
        NMarks += p_m->no_mark;
        NTraits += p_m->no_traits;
    }
    entry = FALSE;	
    while (p_u = nxt_unlnkd_trait(&entry))     
    {
        ++NTraits;
        ++NUnlnkd;
    }
    NLoci = NMarks + NTraits;

    alloc_recomb(); 
                                	/* for F and M recomb tables,   */
                                        /* adds entry for each locus    */
                                        /* to table	                */
    for (sex = FEM; sex <= MAL; ++sex)
    {
        loc_no = 0;
        entry = FALSE;
        while (p_m = nxt_prntmap(&entry, sex)) 
        {
            entry2 = FALSE; 
            while (p_l = nxt_prntmap_locus(&entry2, p_m)) 
            {
                mak_recomb(p_l, loc_no, sex);
                ++loc_no;
            }
        }
                         		/* for each unlinked trait	*/
        entry = FALSE;                  /* adds entry to recomb table   */
        while (p_u = nxt_unlnkd_trait(&entry))     
        {
            mak_no_recomb(p_u, loc_no, sex);
            ++loc_no;
        }
    }

    if (NTraits)
    {
        alloc_trait();		        /* for each trait		*/
	                                /* stores trait computational   */
        loc_no = 0;                     /* parameters                   */  
        t_no = 0;          
        entry = FALSE;
        while (p_m = nxt_prntmap(&entry, AVG))
        {
            entry2 = FALSE;
            while (p_l = nxt_prntmap_locus(&entry2, p_m))
            {
                if (p_l->locus_typ == TD) 
                {
                    mak_trait(p_l->id, loc_no, t_no);
                    ++t_no;
                }
                ++loc_no;
            }
        } 
        entry = FALSE;         
        while (p_u = nxt_unlnkd_trait(&entry))     
        {
            mak_trait(p_u->id, loc_no, t_no);
            ++t_no;
            ++loc_no;
        }
    }
					/* for each chromosome	 	*/
    loc_no = 0;
    entry = FALSE;
    while (p_m = nxt_prntmap(&entry, AVG))
    {			                /* for each locus		*/
        entry2 = FALSE;                 /* stores allele frequencies	*/
        while (p_l = nxt_prntmap_locus(&entry2, p_m))
        {
            mak_freq(p_m, p_l, p_l->id, loc_no); 
            ++loc_no;
        }
    } 
    entry = FALSE;         
    while (p_u = nxt_unlnkd_trait(&entry))     
    {
        mak_freq(NULL, NULL, p_u->id, loc_no);
        ++loc_no;
    }

    max_alle = 0;
    loc_no = 0;
    entry = FALSE;
    while (p_m = nxt_prntmap(&entry, AVG))
    {			               /* finds the maximum no. of      */
        entry2 = FALSE;                /* alleles       		*/
        while (p_l = nxt_prntmap_locus(&entry2, p_m))
    {
            if ((PAlleleFreq + loc_no)->nAlleles > max_alle)
                max_alle = (PAlleleFreq + loc_no)->nAlleles;
            ++loc_no;
        }
    } 
    WidAlle = intWid(0, max_alle);

    return TRUE;
}

/*-------------------- functions follow --------------------------------*/

SIM_PTR nxt_sim_chrom(int * p_entry)
{					
    static SIM_PTR p_sav;               /*returns a pointer to the next */
                                        /*sim node                      */
    if (! *p_entry)
    {
        *p_entry = TRUE;
        return (p_sav = first_sim_node);
    }
    else
        return (p_sav = p_sav->p_next);
}

DIST_PTR nxt_mapd_trait(int * p_entry, int sx, SIM_PTR p_sim)
{				  	/* returns a pointer to the     */
					/* entry for the next trait in  */
                                        /* the distance list            */
                                        /* uses defaults if specified   */
                                        /* sex not present              */
    int i, j;
    static int no_loci;                                                   
    MAP_PTR p, p_map; 
    static DIST_PTR p_sav;

    if (!*p_entry)                      /* tests for next chromosome    */
    {
        *p_entry = TRUE;
        if (p_sim->no_traits == 0)
            return NULL;

        p_map = NULL;

        for (i = 1; !p_map && i <= 4; ++i)
        {
            if (i == 1)			/* order to check for map presnt*/
            j = sx;
            else if (i == 2)
            j = AVG;
            else if (i == 3)
            j = FEM;
            else if (i == 4)
            j = MAL;
        
            if (p = first_trait_map_node[j])
            {
                while (p->p_next && p->chrom_no < p_sim->chrom_no)
                    p = p->p_next;
                if (p->chrom_no == p_sim->chrom_no)
                {
                    no_loci = p->no_loci;
                    p_map = p;
                }
            }
        }
        if (!p_map)
            return NULL;
        else
            return (p_sav = p_map->p_dist);
    }
    if (--no_loci)                      /* returns ptr to next trait    */
        return ++p_sav;                 /* (if any)                     */
    else
        return NULL;
}

DIST_PTR nxt_mapd_mark(int * p_entry, int sx, SIM_PTR p_sim)
{					/* returns a pointer to the 	*/
					/* entry for the next marker in	*/
					/* the distance list		*/
                                        /* uses defaults if specified   */
                                        /* sex not present              */
    int i, j;
    static int no_loci;
    MAP_PTR p, p_map; 
    static DIST_PTR p_sav;

    if (!*p_entry)
    {
        *p_entry = TRUE;

        if (p_sim->no_mark == 1)
            return NULL;
        
        p_map = NULL;

        for (i = 1; !p_map && i <= 4; ++i)
        {
            if (i == 1)			/* order to check for map presnt*/
            j = sx;
            else if (i == 2)
            j = AVG;
            else if (i == 3)
            j = FEM;
            else if (i == 4)
            j = MAL;
        
            if (p = first_mark_map_node[j])
            {
                while (p->p_next && p->chrom_no < p_sim->chrom_no)
                    p = p->p_next;
                if (p->chrom_no == p_sim->chrom_no)
                {
                    no_loci = p->no_loci;
                    p_map = p;
                }
            }
        }
        if (!p_map)
            return NULL;           
        else
            return (p_sav = p_map->p_dist);
    }
    if (--no_loci)
        return ++p_sav;
    else
        return NULL;
}
    
UNLNKD_PTR nxt_unlnkd_trait(int * p_entry)
{
    static UNLNKD_PTR p_sav;

    if (!*p_entry)
    {
        *p_entry = TRUE;				
        return (p_sav = first_unlnkd_node);     /* returns ptr to first */
    }                                        	/* unlinked trait       */
    else
        return (p_sav = p_sav->p_next);      	/* returns ptr to next  */
}

int nxt_mark(int * p_entry, SIM_PTR p_sim)
{
    static int mark_no;

    if (!*p_entry)
    {
        *p_entry = TRUE;
        return (mark_no = 1);
    }
    else
    {
        if (mark_no >= p_sim->no_mark)
            return 0;
        else
            return ++mark_no;
    }
}

TLIST_PTR nxt_trait(int * p_entry)
{
    static TLIST_PTR p_sav;

    if (!*p_entry)
    {
        *p_entry = TRUE;
        return (p_sav = first_tlist_node);
    }
    else
        return (p_sav = p_sav->p_next);
}

PRNTLOC_PTR nxt_prntmap_locus(int * p_entry, PRNTMAP_PTR p_pmap)
{
    static PRNTLOC_PTR p_loc;

    if (!*p_entry)
    {
        *p_entry = TRUE;
        p_loc = p_pmap->p_locus;
    }
    else
        p_loc = p_loc->p_next;
    
    return p_loc;
}

PRNTMAP_PTR nxt_prntmap(int * p_entry, int sx)
{				
    static PRNTMAP_PTR p_sav;         	/*returns a pointer to the next */
                                      	/*printmap node                 */
    if (! *p_entry)
    {
        *p_entry = TRUE;
        return (p_sav = first_prntmap_node[sx]);
    }
    else
        return (p_sav = p_sav->p_next);
}

int test_sim(void)
{
    if (!first_sim_node && !first_unlnkd_node)
    {                                   /* tests for simulate statement */
        err_typ = ERD_SIM;
        return FALSE;
    }
    if (first_sim_node)                 /* checks for both names and    */
                                        /* unnamed chromosomes          */ 
        if (first_sim_node->chrom_no == 0 && first_sim_node->p_next)
        {
            err_typ = ERD_CHR;
            return FALSE;
        }
     return TRUE;
}

int test_mark_map(SIM_PTR p_sim, int * ss_map)
{					/*tests for marker map present	*/
					/*for this chromosome and with	*/
					/*correct no. of markers	*/
    int sx;
    int found_map = FALSE;
    MAP_PTR p_map;

    if (p_sim->no_mark == 1)
        return TRUE;
           
    for (sx = FEM; sx <= AVG; ++sx)
        if (p_map = first_mark_map_node[sx])
        {
            while (p_map->p_next && p_map->chrom_no < p_sim->chrom_no)
                p_map = p_map->p_next;
            if (p_map->chrom_no == p_sim->chrom_no)
            {
                found_map = TRUE;
                if (sx != AVG)
                    *ss_map = TRUE;        
                if (p_map->no_loci != p_sim->no_mark)
                {
                    if (sx == AVG)
                        err_typ = ERD_NAM;
                    else if (sx == MAL)
                        err_typ = ERD_NMM;
                    else if (sx == FEM)
                        err_typ = ERD_NFM;
                    on_chrom (p_sim->chrom_no);
                }
            }
        }
    if (!found_map)
    {
        err_typ = ERD_MMM;
        for_chrom (p_sim->chrom_no);
    }
    if (err_typ)
        return FALSE;
    else
        return TRUE;
}

int test_trait_map(SIM_PTR p_sim, int * ss_map)
{                                       /*tests for trait map present   */
                                        /*for this chromosome and with  */
                                        /*correct no. of traits and     */
                                        /*marker no's within range      */
                                        /*checks sex-specific maps      */
    int j, sx;
    MAP_PTR p_map, p_sm[3] = {NULL, NULL, NULL};                     
    DIST_PTR p_d1, p_d2;                  

    if (!p_sim->no_traits)           
        return TRUE;
    
    for (sx = FEM; sx <= AVG; ++sx)
        if (p_map = first_trait_map_node[sx])
        {
            while (p_map->p_next && p_map->chrom_no < p_sim->chrom_no)
                p_map = p_map->p_next;
                                        /*checks no. traits mapped for	*/
                                        /*chrom same as no. to be sim   */
            if (p_map->chrom_no == p_sim->chrom_no)
            {
                if (sx != AVG)
                    *ss_map = TRUE;        
                p_sm[sx] = p_map;
                if (p_map->no_loci != p_sim->no_traits)
                {
                    if (sx == AVG) 
                        err_typ = ERD_NAT;
                    else if (sx == MAL) 
                        err_typ = ERD_NMT;
                    else if (sx == FEM) 
                        err_typ = ERD_NFT;
                    on_chrom (p_sim->chrom_no);
                }
            }
        }
                                       	/*checks for at least one trait	*/
                                        /*map present                   */
    if (!p_sm[AVG] && !p_sm[MAL] && ! p_sm[FEM])
    {
        err_typ = ERD_MTM;
        for_chrom (p_sim->chrom_no);
    }
    if (err_typ)
        return FALSE;
                                    	/*checks sex-specific maps for	*/
    for (sx = FEM; sx <= AVG; ++sx)     /*consistency                   */
    {
        if (p_sm[sx])
        {
            p_d1 = p_sm[sx]->p_dist;
            p_d2 = NULL;
            if (p_sm[(sx + 1) % (AVG + 1)])
                p_d2 = p_sm[(sx + 1) % (AVG + 1)]->p_dist;

            for (j = 0; j < p_sm[sx]->no_loci; ++j)
            {				/*checks for valid marker no.	*/
                                        /*in trait map                  */
                if (p_d1->mark_id > p_sim->no_mark)
                {
                    err_typ = ERD_MMX;
                    for_chrom (p_sim->chrom_no);
                }
                else if (p_d2) 		/*checks for trait no. the same	*/
                {                       /*in sex-specific maps          */
                    if (p_d1->id != p_d2->id)
                    { 
                        err_typ = ERD_TID;
                        for_chrom (p_sim->chrom_no);
                    }
                    else if (p_d1->mark_id != p_d2->mark_id)
                    { 			/*checks for trait locus the 	*/
	                                /*same in sex-specific maps     */
                        err_typ = ERD_MID;
                        for_chrom (p_sim->chrom_no);
                    }
                    ++p_d2;
                }
                ++p_d1;
            } 
        }
    }
    if (err_typ)
        return FALSE;
    else 
        return TRUE;
}

int test_trait_list(int linked, int t_no)
{
    TLIST_PTR cur_node, new_node;

    if (first_tlist_node)
    {
        cur_node = first_tlist_node;
        while (cur_node->p_next && cur_node->p_next->id <= t_no)
            cur_node = cur_node->p_next;
        				
        if (cur_node->id == t_no)     	/* checks for trait already in  */
                                        /* list                         */
        {
            sprintf(errnumb, "%d", t_no);
            err_txt = errnumb;
            if (linked && cur_node->lnkd)
                err_typ = ERD_LLL;
            else
                err_typ = ERD_LUL;
        }
    }
    if (err_typ)
        return FALSE;

    new_node = malloc(sizeof(TLIST));
    alloc_chk(WHERE, new_node);

    new_node->id = t_no;
    new_node->lnkd = linked;

    if (!first_tlist_node || first_tlist_node->id > t_no)
    {
        new_node->p_next = first_tlist_node;
        first_tlist_node = new_node;
    }
    else
    {
        new_node->p_next = cur_node->p_next;
        cur_node->p_next = new_node;
    }
    return TRUE;
}

int test_mark_loci(SIM_PTR p_sim, int m_no)
{
    LOCUS_PTR p_locus;
    FREQ_PTR p_freq;

    p_locus = first_mark_locus_node;

    if (!p_locus)
        err_typ = ERD_NML;
    else
    { 
        while (p_locus->p_next && (p_locus->chrom_no < p_sim->chrom_no
                            || (p_locus->chrom_no == p_sim->chrom_no
                                  && p_locus->id < m_no)))
            p_locus = p_locus->p_next;
    
        if (p_locus->chrom_no != p_sim->chrom_no 
                 || (p_locus->chrom_no == p_sim->chrom_no 
                               && p_locus->id != m_no))
            err_typ = ERD_NML;
    }
    if (err_typ)
    {
        for_chrom_mark(p_sim->chrom_no, m_no);
					/*stores default values  	*/
        p_freq = alloc_dflt_frq();
        stor_dflt_frq(p_freq);
        p_locus = alloc_dflt_locus_node(p_sim->chrom_no, m_no, MD);
        stor_dflt_locus_data(p_sim->chrom_no, m_no, p_locus, p_freq);

        return FALSE;
    }
    else
        return TRUE;
}
int test_trait_vars(int t_no)
{
    int no_addi = FALSE, no_resi = FALSE;
    double sum = 0.;
    VARI_PTR p_addi, p_resi;

    p_addi = first_addi_node;
 
    p_resi = first_resi_node;

    if (!p_addi)
        no_addi = TRUE; 
    else
    {
        while (p_addi->p_next && p_addi->id < t_no)
            p_addi = p_addi->p_next;
        if (p_addi->id == t_no)
            sum = p_addi->vari;
        else
            no_addi = TRUE;
    }
    if (!p_resi)
        no_resi = TRUE;
    else
    {
        while (p_resi->p_next && p_resi->id < t_no)
            p_resi = p_resi->p_next;
        if (p_resi->id == t_no)
            sum = sum + p_resi->vari;
        else
            no_resi = TRUE;
    }
					/*stores default values 	*/
    if (no_addi)
    {
        p_addi = alloc_vari_node(t_no, AV);
        stor_dflt_vari_data(t_no, p_addi);
    }
    if (no_resi)
    {
        p_resi = alloc_vari_node(t_no, RV);
        stor_dflt_vari_data(t_no, p_resi);
    }
    if (no_addi && no_resi)
        err_typ = ERD_NAR;
    else if (sum == 0.)
        err_typ = ERD_SVR;
    else if (no_addi)
        err_typ = ERD_TAV;
    else if (no_resi)
        err_typ = ERD_TRV;

    if (err_typ)
    {
        sprintf(errnumb, "%d", t_no);
        err_txt = errnumb;
        return FALSE;
    }
    else 
        return TRUE;
}

int test_trait_genotyp(int t_no)
{
    MEANS_PTR p_means;

    p_means = first_means_node;

    if (!p_means)
        err_typ = ERD_NGM;
    else
    {
        while (p_means->p_next && p_means->id < t_no)
            p_means = p_means->p_next;

        if (p_means->id != t_no)
            err_typ = ERD_NGM;
    }
    if (err_typ)
    {
        sprintf(errnumb, "%d", t_no);
        err_txt = errnumb;
					/*stores default values  	*/
        p_means = alloc_means_node(t_no);
        stor_dflt_mean_data(t_no, p_means);

        return FALSE;
    }
    else
        return TRUE;
}

int test_trait_loci(int t_no)
{
    LOCUS_PTR p_locus;
    FREQ_PTR p_freq;

    p_locus = first_trait_locus_node;

    if (!p_locus)
        err_typ = ERD_NTL;
    else
    { 
        while (p_locus->p_next && p_locus->id < t_no)
            p_locus = p_locus->p_next;
    
        if (p_locus->id != t_no)
            err_typ = ERD_NTL;
    }
    if (err_typ)
    {
        sprintf(errnumb, "%d", t_no);
        err_txt = errnumb;
					/*stores default values		*/
        p_freq = alloc_dflt_frq();
        stor_dflt_frq(p_freq);
        p_locus = alloc_dflt_locus_node(NO_CH_NO, t_no, TD);
        stor_dflt_locus_data(NO_CH_NO, t_no, p_locus, p_freq);

        return FALSE;
    }
    else
        return TRUE;
}

void for_chrom(int chr_no)
{
    if (chr_no > 0)
        sprintf(errnumb, "for chromosome %d", chr_no);
    else
        sprintf(errnumb, "");

    err_txt = errnumb;
    return;
}

void on_chrom(int chr_no)
{
    if (chr_no > 0)
        sprintf(errnumb, "on chromosome %d", chr_no);
    else
        sprintf(errnumb, "");

    err_txt = errnumb;
    return;
}

void for_chrom_mark(int chr_no, int m_no)
{
    if (chr_no > 0)
        sprintf(errnumb, "chromosome %d marker %d", chr_no, m_no);
    else
        sprintf(errnumb, "marker %d", m_no);

    err_txt = errnumb;
    return;
}

PRNTMAP_PTR add_prntmap(SIM_PTR p_sim, int sx)
{
    PRNTMAP_PTR p_pmap, p_cur;
						
    p_pmap = malloc(sizeof(PRNTMAP));	/* makes prntmap node for this	*/
    alloc_chk(WHERE, p_pmap);		/* chromosome                   */

    p_pmap->chrom_no = p_sim->chrom_no;
    p_pmap->no_mark = 0;
    p_pmap->no_traits = 0;
    p_pmap->p_locus = NULL;
    p_pmap->p_next = NULL;

    if (!first_prntmap_node[sx])       /* puts node as first in list 	*/
        first_prntmap_node[sx] = p_pmap;	
    else
    {
        p_cur = first_prntmap_node[sx];	/* or adds node to end of list	*/
        while (p_cur->p_next)		
            p_cur = p_cur->p_next;
        p_cur->p_next = p_pmap;
    }
    return p_pmap;
}

void add_only_mark_locus(PRNTMAP_PTR p_pmap)
{
    PRNTLOC_PTR p_new;

    p_new = malloc(sizeof(PRNTLOC));	/* makes prntmap_locus node for	*/
    alloc_chk(WHERE, p_new);		/* the unmapped marker          */
			
    p_new->id = 1;
    p_new->mark_id = 0;
    p_new->locus_typ = MD;
    p_new->dist = 0.;
    p_new->loc_dist = 0.;
    p_new->p_prev = NULL;
    p_new->p_next = NULL;

    ++p_pmap->no_mark;
    p_pmap->p_locus = p_new;

    return;
}

void add_mark_locus(PRNTMAP_PTR p_pmap, DIST_PTR p_dist)
{
    PRNTLOC_PTR p_new, p_cur;

    p_new = malloc(sizeof(PRNTLOC));	/* makes prntmap_locus node for	*/
    alloc_chk(WHERE, p_new);		/* this marker                  */

    p_new->id = p_dist->id;
    p_new->mark_id = p_dist->mark_id;
    p_new->locus_typ = MD;
    p_new->dist = p_dist->dist;
    p_new->loc_dist = 0.;
    p_new->p_prev = NULL;
    p_new->p_next = NULL;

    if(!p_pmap->p_locus)		/* puts node at start of list	*/
        p_pmap->p_locus = p_new;	
    else
    {
        p_cur = p_pmap->p_locus;	/* or adds node to end  of list	*/
        while (p_cur->p_next)	
            p_cur = p_cur->p_next;
        p_cur->p_next = p_new;
        p_new->p_prev = p_cur;
    }
    ++p_pmap->no_mark;

    return;
}

void add_trait_locus(PRNTMAP_PTR p_pmap, DIST_PTR p_dist)
{
    PRNTLOC_PTR p_new, p_cur;

    p_new = malloc(sizeof(PRNTLOC));	/* makes prntmap_locus node  	*/
    alloc_chk(WHERE, p_new);		/* for this trait               */

    p_new->id = p_dist->id;
    p_new->mark_id = p_dist->mark_id;
    p_new->locus_typ = TD;
    p_new->dist = p_dist->dist;
    p_new->loc_dist = 0;
    p_new->p_prev = NULL;
    p_new->p_next = NULL;

    if (p_dist->mark_id == 0)
    {
        if (p_pmap->p_locus->locus_typ == MD
             || (p_pmap->p_locus->locus_typ == TD
                  && p_pmap->p_locus->dist < p_dist->dist))
        {
            p_new->p_next = p_pmap->p_locus;
            p_pmap->p_locus->p_prev = p_new;
            p_pmap->p_locus = p_new;
        }
        else
        {
            p_cur = p_pmap->p_locus;        
            while (p_cur->p_next
                   && p_cur->p_next->locus_typ == TD
                               && p_cur->p_next->dist > p_dist->dist)
                p_cur = p_cur->p_next;
            
            p_cur->p_next->p_prev = p_new;
            p_new->p_next = p_cur->p_next;
            p_new->p_prev = p_cur;
            p_cur->p_next = p_new;
        }
    }
    else
    {
        p_cur = p_pmap->p_locus;
        while (p_cur->p_next 
               && (p_cur->p_next->locus_typ == TD 
                   || (p_cur->p_next->locus_typ == MD 
                                && p_cur->p_next->id <= p_dist->mark_id)))
            p_cur = p_cur->p_next;
       
        if (p_cur->p_next)
            p_cur->p_next->p_prev = p_new;
        p_new->p_next = p_cur->p_next;
        p_new->p_prev = p_cur;
        p_cur->p_next = p_new;
    }

    ++p_pmap->no_traits;

    return;
}

int test_trait_dist(PRNTMAP_PTR p_pmap, PRNTLOC_PTR p_loc)
			
{					/*checks marker to trait dist.	*/
					/*Should not be > marker to 	*/
					/*marker dist 			*/ 
    PRNTLOC_PTR p_cur;

    if (p_loc == p_pmap->p_locus)	/*checks for 1st locus		*/
        return TRUE;

    else if (p_loc->locus_typ == MD)	/*checks for marker locus	*/
        return TRUE;

    else if (p_loc->mark_id == 0)	/*checks for no preceding 	*/
        return TRUE;			/*marker locus			*/
				
    else if (!p_loc->p_next)		/*checks for no following	*/
        return TRUE;			/*marker locus			*/

    else
    {
        p_cur = p_loc->p_next;

        while (p_cur->p_next &&  p_cur->locus_typ != MD)
            p_cur = p_cur->p_next;

        if (p_cur->locus_typ == MD	/*tests marker to trait dist	*/
                   && p_cur->dist < p_loc->dist)
        {
            err_typ = ERD_MTD;
            sprintf(errnumb, "%d", p_loc->id);
            err_txt = errnumb;
        }
    }
    if (err_typ)
        return FALSE;
    else
        return TRUE;
}

void fix_locus_dist(PRNTMAP_PTR p_pmap, PRNTLOC_PTR p_loc)
{					/*finds dist between preceding	*/
					/*locus & current locus  	*/

    if (p_loc == p_pmap->p_locus)		/*if first locus,     	*/
        p_loc->loc_dist = 0.;			/*dist is 0.		*/

    else if (p_loc->locus_typ == TD		/*if trait precedes    	*/
             && p_loc->mark_id == 0)		/*1st marker,          	*/
        p_loc->loc_dist = p_loc->p_prev->dist	/*dist is preceding	*/
                          - p_loc->dist;	/*trait to marker dist	*/
						/*less trait to marker	*/
						/*dist			*/	

    else if (p_loc->locus_typ == MD		/*if marker is 1st 	*/
             && p_loc->id == 1)			/*marker,		*/ 
        p_loc->loc_dist = p_loc->p_prev->dist;	/*dist is preceding	*/
						/*trait to marker dist	*/
				
    else if (p_loc->p_prev->locus_typ == MD)	/*if marker or trait 	*/
        p_loc->loc_dist = p_loc->dist;		/*follows marker, 	*/
						/*dist is ok		*/

     else if (p_loc->locus_typ == MD 		/*if marker follows	*/
             && p_loc->p_prev->locus_typ == TD) /*trait,		*/
        p_loc->loc_dist = p_loc->dist 		/*dist is preceding	*/ 
                        - p_loc->p_prev->dist;	/*marker to marker dist	*/
						/*less preceding marker	*/
						/*to trait dist		*/

    else if (p_loc->locus_typ == TD 		/*if trait follows 	*/
             && p_loc->p_prev->locus_typ == TD) /*trait,		*/
        p_loc->loc_dist = p_loc->dist		/*dist is preceding 	*/
                        - p_loc->p_prev->dist;  /*marker to trait dist	*/
						/*less preceding marker	*/
						/*to preceding trait	*/
						/*dist			*/	
    return;
}

void prnt_map(PRNTMAP_PTR p_pmap, int sex)
{					/*prints distances between	*/
					/*loci on the chromosome	*/
    PRNTLOC_PTR p_cur;
    int no_loci;
    int spac = 7, max_loci = 11;
    char trait_line[81],      * p_tr_ids  = trait_line;
    char trait_tic_marks[81], * p_tr_tics = trait_tic_marks;
    char dist_line[81],       * p_dists   = dist_line;
    char mark_tic_marks[81],  * p_mk_tics = mark_tic_marks;
    char mark_line[81],       * p_mk_ids  = mark_line;

    if (p_pmap->no_mark <= 1 && p_pmap->no_traits <= 0)
    {
        if (p_pmap->chrom_no != 0)
            fprintf(stdout, "\n No map for chromosome %d\n\n\n",
                                p_pmap->chrom_no);
        else
            fprintf(stdout, "\n No chromosome map\n\n\n");
        
        return;
    }

    if (p_pmap->chrom_no != 0)
    {
        fprintf(stdout, "\n Chromosome %d", p_pmap->chrom_no);
        switch (sex)
        {
            case MAL:
                fprintf(stdout, " male map distances");
                break;
            case FEM:
                fprintf(stdout, " female map distances");
                break;
            default:
                fprintf(stdout, " distances");
                break;
        }
    }
    else
    {
        switch (sex)
        {
            case MAL:
                fprintf(stdout, "\n Male map distances");
                break;
            case FEM:
                fprintf(stdout, "\n Female map distances");
                break;
            default:
                fprintf(stdout, "\n Distances");
                break;
        }
    }
    fprintf(stdout, " (cM):\n\n");

    p_cur = p_pmap->p_locus;

    while (p_cur && p_cur->p_next)
    {					/*initializes the lines		*/
        if (p_cur->locus_typ == TD)
        {			
            sprintf(p_tr_ids,  "T%d  ",  p_cur->id);
            sprintf(p_tr_tics, " +");
            sprintf(p_mk_tics, " -");
            sprintf(p_mk_ids,  "    ");
        }
        else
        {
            sprintf(p_tr_ids,  "    ");
            sprintf(p_tr_tics, " -");
            sprintf(p_mk_tics, " +");
            sprintf(p_mk_ids,  "M%d  ", p_cur->id);
        }
        sprintf(p_dists, "   ");

        p_cur =  p_cur->p_next;
        no_loci = 1;

        while (p_cur && no_loci < max_loci)
        {				/*adds next locus and distance	*/
            if (p_cur->locus_typ == TD)
            {			
                sprintf(p_tr_ids+no_loci*spac-3,  "   T%d  ",  p_cur->id);
                sprintf(p_tr_tics+no_loci*spac-5, "------+  ");
                sprintf(p_mk_tics+no_loci*spac-5, "-------  ");
                sprintf(p_mk_ids+no_loci*spac-3,  "       ");
            }
            else
            {
                sprintf(p_tr_ids+no_loci*spac-3,  "       ");
                sprintf(p_tr_tics+no_loci*spac-5, "-------  ");
                sprintf(p_mk_tics+no_loci*spac-5, "------+  ");
                sprintf(p_mk_ids+no_loci*spac-3,  "   M%d  ", p_cur->id);
            }
            sprintf(p_dists+no_loci*spac-5, "%5.1f    ", p_cur->loc_dist);

            ++no_loci;
            p_cur = p_cur->p_next;
        }                               /*done with this map line	*/
					/*prints it out			*/
                                        /*first makes all lines the	*/
					/*same length			*/

        *(p_tr_ids+no_loci*spac-3) = '\0';
        *(p_tr_tics+no_loci*spac-3) = '\0';
        *(p_tr_tics+no_loci*spac-3) = '\0';
        *(p_dists +no_loci*spac-3) = '\0';
        *(p_mk_tics+no_loci*spac-3) = '\0';
        *(p_mk_ids+no_loci*spac-3) = '\0';
        fprintf(stdout, "%s\n", p_tr_ids);
        fprintf(stdout, "%s\n", p_tr_tics);
        fprintf(stdout, "%s\n", p_dists); 
        fprintf(stdout, "%s\n", p_mk_tics);
        fprintf(stdout, "%s\n\n", p_mk_ids);

        if (p_cur)			/*backs up one locus		*/
            p_cur = p_cur->p_prev;	/*for start of next line	*/
    }
    fflush(stdout);
    return;
}

void prnt_mark_freqs(SIM_PTR p_sim)
{
    int cols_used = 0;
    int tot_cols = 15;
    int ix;
    char prnt_lin[81] = "";
    LOCUS_PTR cur_mark;

    if (!first_mark_locus_node)
        return;

    cur_mark = first_mark_locus_node;
    while (cur_mark->p_next && cur_mark->chrom_no < p_sim->chrom_no)
        cur_mark = cur_mark->p_next;

    if (cur_mark->chrom_no != p_sim->chrom_no)
        return;

    if (p_sim->chrom_no !=0)
        fprintf(stdout, " Marker allele frequencies for"
                        " chromosome %d:\n\n", p_sim->chrom_no);
    else
        fprintf(stdout, " Marker allele frequencies:\n\n");

					/* prints allele frequencies	*/
					/* for next marker 		*/
    while (cur_mark && cur_mark->chrom_no == p_sim->chrom_no
                    && cur_mark->id <= p_sim->no_mark)
    {
        if (cols_used > 0)
            if (cur_mark->no_alleles > tot_cols - cols_used - 2)
            {
                fprintf(stdout, "%s\n", prnt_lin);
                cols_used = 0;
            }
            else
            {
                sprintf(prnt_lin+5*cols_used, "     ");
                ++cols_used;
            }
        sprintf(prnt_lin+5*cols_used, " M%d  ", cur_mark->id);
        ++cols_used;
        ix = 0;

        while (ix < cur_mark->no_alleles)
        {
            if (cols_used >= tot_cols)
            {
                fprintf(stdout, "%s\n", prnt_lin);
                cols_used = 0;
            }
            sprintf(prnt_lin+5*cols_used, "%5.3f", 
                           *(cur_mark->p_freq_node->p_freq+ix)); 
            *(prnt_lin+5*cols_used) = ' ';
            ++ix;
            ++cols_used;
        }
        cur_mark = cur_mark->p_next;
    }
    if (cols_used > 0) 
        fprintf(stdout, "%s\n", prnt_lin);
    fprintf(stdout, "\n");
    fflush(stdout);
}

void prnt_trait_freqs(PRNTMAP_PTR p_pmap)
{
    PRNTLOC_PTR to_do;
    char prnt_lin[81];
    int i;

    if (!p_pmap->no_traits)
        return;
    
    fprintf(stdout, " Trait allele frequencies:\n\n");

    to_do = p_pmap->p_locus;
    
    for (i = 0; i < p_pmap->no_traits; ++i)
    {                                           /*finds next trait no   */
                                                /*in map                */
        while (to_do->locus_typ == MD  &&  to_do->p_next)
            to_do = to_do->p_next; 
                                                /*formats the freqs     */
        fmt_trait_freqs(prnt_lin, i, to_do->id);
         
        if (i + 1 == p_pmap->no_traits || !((i + 1) % 4))
            fprintf(stdout, "%s\n", prnt_lin);	/* prints the line      */
        else
            sprintf(prnt_lin+(i%4)*20+15, "     ");
    
        if (to_do->p_next)
            to_do = to_do->p_next;
    }
    fprintf(stdout, "\n");
    fflush(stdout);
}
    
void prnt_trait_pars(PRNTMAP_PTR p_pmap)
{
    PRNTLOC_PTR to_do;
    char prnt_lin[81];
    int i;

    if (!p_pmap->no_traits)
        return;
    
    fprintf(stdout, " Trait genotype means                  "
                    " Additive, residual variances\n\n");

    to_do = p_pmap->p_locus;
    
    for (i = 0; i < p_pmap->no_traits; ++i)
    {
        while (to_do->locus_typ == MD 
                      && to_do->p_next)   	/*finds next trait no   */
            to_do = to_do->p_next;              /*in map	        */
                                                /*formats the means and */
        fmt_trait_pars(prnt_lin, to_do->id);    /*variances             */

        fprintf(stdout, "%s\n", prnt_lin);	/* prints the line      */
    
        if (to_do->p_next)
            to_do = to_do->p_next;
    }
    fprintf(stdout, "\n");
    fflush(stdout);
}

void prnt_unlnkd_freqs(void)
{
    UNLNKD_PTR to_do;
    char prnt_lin[81];
    int i;

    if (!first_unlnkd_node)
        return;
    
    if (!first_unlnkd_node->p_next)
        fprintf(stdout, "\n Unlinked trait:\n\n");
    else
        fprintf(stdout, "\n Unlinked traits:\n\n");

    fprintf(stdout, " Allele frequencies:\n\n");

    to_do = first_unlnkd_node;
    i = 0;
    
    while (to_do)                               /*formats the freqs for */
    {                                           /*the next unlinkd trait*/
        fmt_trait_freqs(prnt_lin, i, to_do->id);

        if (!to_do->p_next || !((i + 1) % 4))
            fprintf(stdout, "%s\n", prnt_lin);	/* prints the line      */
        else
            sprintf(prnt_lin+(i%4)*20+15, "     ");
    
        to_do = to_do->p_next;
        ++i;
    }
    fprintf(stdout, "\n");
    fflush(stdout);
}

void prnt_unlnkd_pars(void)
{
    UNLNKD_PTR to_do;
    MEANS_PTR cur_means;
    VARI_PTR cur_addi, cur_resi;
    char prnt_lin[81];

    if (!first_unlnkd_node)
        return;
    
    fprintf(stdout, " Genotype means                        "
                    " Additive, residual variances\n\n");

    to_do = first_unlnkd_node;
    
    while (to_do)                               /*formats the means and */
    {                                           /*variances for the next*/
                                                /*unlinked trait        */
        fmt_trait_pars(prnt_lin, to_do->id);

        fprintf(stdout, "%s\n", prnt_lin);	/* prints the line      */
    
        to_do = to_do->p_next;
    }
    fprintf(stdout, "\n");
    fflush(stdout);

    return;
}

void fmt_trait_freqs(char * p_lin, int n, int t_no)
{
    LOCUS_PTR cur_trait;
						/*finds locus node      */
    cur_trait = first_trait_locus_node;		/*for this trait        */
    while (cur_trait->p_next && cur_trait->id < t_no)
        cur_trait = cur_trait->p_next;
						/*adds params for the   */
						/*trait to line         */
    sprintf(p_lin+(n%4)*20, " T%d  ", cur_trait->id);
    sprintf(p_lin+(n%4)*20+5, "%5.3f", 
                                 *(cur_trait->p_freq_node->p_freq));
    *(p_lin+(n%4)*20+5) = ' ';
    sprintf(p_lin+(n%4)*20+10, "%5.3f",
                                 *(cur_trait->p_freq_node->p_freq+1));
    *(p_lin+(n%4)*20+10) = ' ';

    return;
}

void fmt_trait_pars(char * p_lin, int t_no)
{
    MEANS_PTR cur_means;
    VARI_PTR cur_vari;
    int deci;
    double mini, maxi;
						/*finds means node for  */
    cur_means = first_means_node;  		/*this trait            */
    while (cur_means->p_next && cur_means->id < t_no)
        cur_means = cur_means->p_next;
                                                /*determines print      */
    mini = *(cur_means->mean);                  /*format for parameters */
    if (*(cur_means->mean+1) < mini)
        mini = *(cur_means->mean+1);
    if (*(cur_means->mean+2) < mini)
        mini = *(cur_means->mean+2);
    maxi = *(cur_means->mean);
    if (*(cur_means->mean+1) > maxi)
        maxi = *(cur_means->mean+1);
    if (*(cur_means->mean+2) > maxi)
        maxi = *(cur_means->mean+2);
    deci = decWid(mini, maxi, 4);
    if (deci > 7)
        deci = 7; 

    sprintf(p_lin, " T%d  ", cur_means->id);
						/*adds means for the    */
						/*trait to line         */
    sprintf(p_lin +5, " %9.*f", deci, *(cur_means->mean));
    sprintf(p_lin+15, " %9.*f", deci, *(cur_means->mean+1));
    sprintf(p_lin+25, " %9.*f", deci, *(cur_means->mean+2));
    sprintf(p_lin+35, "     ");

						/*finds additive vari   */
    cur_vari = first_addi_node;    		/*for this trait        */
    while (cur_vari->p_next && cur_vari->id < t_no)
        cur_vari = cur_vari->p_next;            /*and adds to line      */
    sprintf(p_lin+40, " %9.*f", deci, cur_vari->vari);
						/*finds residual vari   */
    cur_vari = first_resi_node;    		/*for this trait        */
    while (cur_vari->p_next && cur_vari->id < t_no)
        cur_vari = cur_vari->p_next;            /*and adds to line      */
    sprintf(p_lin+50, " %9.*f", deci, cur_vari->vari); 

    return;
}

void alloc_recomb(void)
{
    PRecombMap = malloc(NLoci * sizeof(RMAP));
    alloc_chk(WHERE, PRecombMap);

    return;
}

void mak_recomb(PRNTLOC_PTR p_ploc, int loc, int sx)
{
    RMAP_PTR p_recomb;
    PRNTLOC_PTR p_bck, p_nxt;
    int i;
    double d, r_a, r_b, r_star;		
                       			/*initializes recomb map for    */
    p_recomb = PRecombMap + loc;        /*for markers & lnked traits    */
    p_recomb->id = p_ploc->id;                
    p_recomb->iA = p_recomb->iB = 0;
    p_recomb->probA[sx] = p_recomb->probB[sx] = 0.;

    if (p_ploc->locus_typ == MD)       	/* **marker**                   */
    {
        p_bck = p_ploc;
        i = loc;
                                        /*looks for preceding marker    */
        while ((p_bck = p_bck->p_prev) && p_bck->locus_typ != MD)
            --i;

        if (!p_bck || p_bck->locus_typ != MD)
            p_recomb->recombTyp = FM;  	/*1st marker on chromosome      */

        else
        {
            p_recomb->recombTyp = OM;   /*not 1st marker on chrom       */
            p_recomb->iA = i - 1;
            p_recomb->probA[sx] = dist2rf(p_ploc->dist);
        }
    }
    else                                /* **trait**                    */
        if (!p_ploc->p_prev)
        {                               /*trait is 1st locus on chrmosom*/
            p_recomb->recombTyp = FT;
            p_recomb->probB[sx] = dist2rf(p_ploc->dist);
 
            p_nxt = p_ploc;
            i = loc;
                                        /*finds following marker        */
            while ((p_nxt = p_nxt->p_next) && p_nxt->locus_typ != MD)
                ++i;

            p_recomb->iB = i + 1;
        }
        else                          	/*trait is preceded by marker   */
        {                               /*or trait                      */ 
            p_recomb->iA = loc - 1;

            p_nxt = p_ploc;
            i = loc;
            d = 0.;
                                        /*looks for following marker    */
            while ((p_nxt = p_nxt->p_next) && p_nxt->locus_typ != MD)
            {
                ++i;
                d += p_nxt->loc_dist;
            }
            if (p_nxt && p_nxt->locus_typ == MD)
            {                           /*trait has following marker    */
                p_recomb->recombTyp = IT;
                p_recomb->iB = i + 1;
					/* probA: prob recomb if kA==kB	*/
                                        /* probB: prob recomb with B	*/
                         		/*        if kA!=kB            	*/
                d += p_nxt->loc_dist;
                r_a = dist2rf(p_ploc->loc_dist);
                r_b = dist2rf(d);
                r_star = dist2rf(p_ploc->loc_dist + d);
                p_recomb->probA[sx] = r_a * r_b / (1 - r_star);
                if (r_star != 0.)  
                    p_recomb->probB[sx] = (1 - r_a) * r_b / r_star;
                else
                    p_recomb->probB[sx] = 0.;
            }
            else                   	/*trait has no following marker */
            {
                p_recomb->recombTyp = ET;
                p_recomb->probA[sx] = dist2rf(p_ploc->loc_dist);
            }
        }
    return;
} 

void mak_no_recomb(UNLNKD_PTR p_nolnk, int loc, int sx) 
{
    RMAP_PTR p_recomb;
                                      	/*initializes recomb map for    */
    p_recomb = PRecombMap + loc;        /*unlinked traits               */
    p_recomb->id = p_nolnk->id;            
    p_recomb->recombTyp = UT;
    p_recomb->iA = p_recomb->iB = 0;
    p_recomb->probA[sx] = p_recomb->probB[sx] = 0.;

    return;
}

void alloc_trait(void)
{
    PTraitPars = malloc(NTraits * sizeof(TPARS));
    alloc_chk(WHERE, PTraitPars);

    PTraitOut = malloc(NTraits * sizeof(TOUT));
    alloc_chk(WHERE, PTraitOut);

    return;
}

void mak_trait(int id, int loc, int tno)
{
    TPARS_PTR p_trait;
    TOUT_PTR p_tout;
    MEANS_PTR p_means;
    VARI_PTR p_vari;
    TLIST_PTR p_tlist;

    double mini, maxi;

    p_trait = PTraitPars + tno;
    p_trait->id = id;
    
    p_means = first_means_node;
    while (p_means->p_next && p_means->id < id)
        p_means = p_means->p_next;
    p_trait->means[0] = p_means->mean[0];
    p_trait->means[1] = p_means->mean[1];
    p_trait->means[2] = p_means->mean[2];
 
    mini = p_trait->means[0];
    if (p_trait->means[1] < mini)
        mini = p_trait->means[1];
    if (p_trait->means[2] < mini)
        mini = p_trait->means[2];
    maxi = p_trait->means[0];
    if (p_trait->means[1] > maxi)
        maxi = p_trait->means[1];
    if (p_trait->means[2] > maxi)
        maxi = p_trait->means[2];

    p_vari = first_addi_node;
    while (p_vari->p_next && p_vari->id < id)
        p_vari = p_vari->p_next;
    p_trait->addi_term_found = sqrt(p_vari->vari);
    p_trait->addi_term = sqrt(p_vari->vari / 2.);

    p_vari = first_resi_node;
    while (p_vari->p_next && p_vari->id < id)
        p_vari = p_vari->p_next;
    p_trait->resi_term = sqrt(p_vari->vari);

    p_tlist = first_tlist_node;
    p_tout = PTraitOut;
    while (p_tlist->p_next && p_tlist->id < id)
    {
        p_tlist = p_tlist->p_next;
        ++p_tout;
    }
    p_tout->id = id;
    p_tout->dec = decWid(mini, maxi, SIGFIG);
    p_tout->wid = fldWid(mini, maxi, p_tout->dec) + 1;
    p_tout->ixG = loc;
    p_tout->ixPh = tno;

    return;
}

void mak_freq(PRNTMAP_PTR p_pmap, PRNTLOC_PTR p_ploc, int id, int loc)
{				
                           		/*stores an array of allele	*/
    int i;				/*frequencies for the locus	*/
    double * p_c; 	
    LOCUS_PTR p_locus;

    if (!PAlleleFreq)
    {
        PAlleleFreq = malloc(NLoci * sizeof(ALLELE));
        alloc_chk(WHERE, PAlleleFreq);
    }
    if (p_ploc && p_ploc->locus_typ == MD)
    {
        p_locus = first_mark_locus_node;
        while (p_locus->p_next 
                    && (p_locus->chrom_no < p_pmap->chrom_no
                        || p_locus->chrom_no == p_pmap->chrom_no
                           && p_locus->id < p_ploc->id))
            p_locus = p_locus->p_next;
    }
    else 
    {
        p_locus = first_trait_locus_node;
        while (p_locus->p_next && p_locus->id < id)
            p_locus = p_locus->p_next;
    }

    (PAlleleFreq + loc)->nAlleles = p_locus->no_alleles;
    (PAlleleFreq + loc)->pCumFreq 
                        = malloc(p_locus->no_alleles * sizeof(double));
    p_c = (PAlleleFreq + loc)->pCumFreq;
    alloc_chk(WHERE, p_c);
                                        /*cummulative frequencies	*/
                                        /*are stored                    */
    *(p_c) = *(p_locus->p_freq_node->p_freq);
    for (i = 1; i < p_locus->no_alleles; ++i)
        *(p_c + i) = *(p_c + i - 1) + *(p_locus->p_freq_node->p_freq + i); 

    return;
}

