#define PROGNAM "PolyEM unibig"
#define PNAM "unibig"

#include "polyem.h"

/*
 * univar program for larger and/or looped pedigrees -- no huge matrices,
 * or peeling ;  modified from univar -- Sept 1993 (EAT)
 *
 * modified, June 1996, by EAT, to allow reading in of predetermined
 * eigenvalues as an option specified by the constant ICH modified, June
 * 1996, by Chi-hong Tseng, to determine eigenvalues of A-matrix by
 * components 
 *
 * modified, Aug 1997 for Version 2.0 revisions:   
 * use of parameter file for user-friendly specification of modelling 
 * parameters;
 * use of dynamic storage for all major arrays, with memory allocated just 
 * prior * to use, and freed as soon as no longer needed.  (MAJ)
 *
 ************************************************************************/
									
void tau_fit(double *e, double *evg)
/* routine for residual variance   */
{
    double temp;
    temp = SigErr * (np - SigErr * invtrc(evg));
    SigErr = (innerprod(e, e, np) + temp) / np;
}

void sig_fit(double *evg, double *b, Term * ginv, int nt)
{
    double temp3;

    /* NOTE: there are two alternative forms of the EM-equations for */
    /* additive variance; sig_fit can use either form depending on */
    /* "what is missing" */

    temp3 = quarz(b, b, ginv, nt);
    SigAdd = (temp3 + SigAdd * (SigErr * invtrc(evg) + (pedsize - np)))
                                                               / pedsize;
}

/* Note within this routine "xx" are the residuals from any fixed effects;
 * i.e. "y" and NOT original data x 
 *
 * Re-written to allow for looped pedigrees; June 1997 (SCH) 
 *
 * IMPORTANT: Assumes that ginv contains the *lower left* non-zero terms 
 * of the full G-inverse.  
 * The henderson() routine has been modified to do this  (originally 
 * it produced the *upper right* terms).  This, of course, only affects
 * the order in which the terms are stored. Having the lower left terms
 * means that itermo() can efficiently work from the bottom of the
 * pedigree upwards, which improves initial convergence if data is mainly
 * present at the bottom of the pedigree. 
 */

void itermo(double *xx, double *b, Term * ginv, int nt)
{
    int i, it, j, k, k1, n;
    double z, alpha, zz, *t1, d;

    t1 = malloc(sizeof(double) * pedsize);
    alloc_chk(WHERE, t1);
    alpha = SigErr / SigAdd;

    for (it = 0; it < ItMo; it++)
    {
	for (i = 0; i < pedsize; i++)
	    t1[i] = xx[qdex[i]];
	n = 0;
	d = 0.0;
	k = nt;
	do
	{
	    i = ginv[k].r1;
	    zz = ginv[k--].value * alpha;
	    k1 = k;
	    while (k > 0 && ginv[k].r1 == i)
	    {
 		j = ginv[k].r2;
 		z = ginv[k--].value;
 		t1[i] -= b[j] * z * alpha;
	    }
	    if (qdex[i])
		zz += 1.0;
	    t1[i] /= zz;

            /* Compute difference in new breeding value to assess */
            /* convergence */
	    zz = b[i] - t1[i];
	    d += zz * zz;
	    n++;
	    b[i] = t1[i];
	    k = k1;
	    zz = b[i] * alpha;
	    while (k > 0 && ginv[k].r1 == i)
            {
		j = ginv[k].r2;
		z = ginv[k--].value;
		t1[j] -= zz * z;
            }
	}
	while (k > 0);
	d = d / (SigAdd * (double) n);
	if (d <= Diff)
	    break;
    }
    free(t1);
    t1 = NULL;
}

void breeding_values(double *y, double *a, double *b, double *e, 
                                                     Term * ginv, int nt)
{
    int i;

    itermo(y, b, ginv, nt);
    for (i = 0; i < pedsize; ++i)
	a[qdex[i]] = b[i];
    for (i = 1; i <= np; ++i)
	e[i] = y[i] - a[i];
}

void em_iter(int k, double *a, double *b, double *epsi, double *y,  
                       Term * ginv, int nt, double *evgen, double *ivgen)

{						        /* EM iteration	*/
    /* iterate for expected breeding values */
    int i;
    double temp;

    out_fixed(y, 0, 0);
    breeding_values(y, a, b, epsi, ginv, nt);

    /* now evaluate likelihood */
    if ((k / ItWt) * ItWt == k)
	lik_big(k, ivgen, y, a, b, ginv, nt);

    /* now implement EM equations for new SigAdd, SigErr */
    sig_fit(evgen, b, ginv, nt);
    tau_fit(epsi, evgen);

    /* now implement EM equations for new mu, SigErr */
    if (NFix == 0)
    {
	out_fixed(y, 0, 0);
  	breeding_values(y, a, b, epsi, ginv, nt);
	mu[0] = 0.0;
	for (i = 1; i <= np; ++i)
	{
	    y[i] = (x[0][i] - a[i]);
	    mu[0] += y[i];
	}
	mu[0] = mu[0] / np;
	Mufti[0][0][1] = mu[0];
	tau_fit(epsi, evgen);
    }
    /* or for new fixed effects */
    else
	for (i = 1; i <= NFix; ++i)
	{
	    out_fixed(y, 0, 0);
  	    breeding_values(y, a, b, epsi, ginv, nt);
	    out_bar_one(a, y, x[0], i, 0);
	    temp = fixeffit(y, i, 0);
	    mu[0] = temp;
	    Mufti[0][0][1] = temp;
	    tau_fit(epsi, evgen);
	}
}
void setvals(double *b)
{
    int i;

    /* mu[0] is set to the initial value for the overall mean */
    mu[0] = Mufti[0][0][1];
 
    /* set initial breeding values */
    for (i = 0; i < pedsize; ++i)
	b[i] = 0.5 * (x[0][qdex[i]] - mu[0]);
    x[0][0] = 0.0;
    b[0] = 0.0;
}

void alloc_vect(double **a, double **b, double **epsi, double **y)
{
    *a = malloc((np + 1) * sizeof(double));
    alloc_chk(WHERE, *a);  
    *b = malloc(pedsize * sizeof(double));
    alloc_chk(WHERE, *b);
    *epsi = malloc((np + 1) * sizeof(double));
    alloc_chk(WHERE, *epsi);
    *y = malloc((np + 1) * sizeof(double));
    alloc_chk(WHERE, *y);
}

free_vect(double *a, double *b, double *epsi, double *y,
                            Term *ginverse, double *evgen, double *ivgen)
{
    free((void *) a);
    a = NULL;
    free((void *) b);
    b = NULL;
    free((void *) epsi);
    epsi = NULL;
    free((void *) y);
    y = NULL;
    free((void *) ginverse);
    ginverse = NULL;
    free((void *) evgen);
    evgen = NULL;
    free((void *) ivgen);
    ivgen = NULL;
}

int main(int argc, char *argv[])
{
    int i, j, k, cpnp, *cdex;
    double *a, *epsi, *b, *y;
    double *evgen, *ivgen, *evgen1;
    int gsize;
    Ind **cpnghd;
    double cc;
    Term *ginverse;
    Indlist *peopl;
    int **dodo = NULL;

    atexit(print_tail);
    print_head(PROGNAM);
 
    /* process all of the input parameter statements */

    if (!proc_pars(argc, argv, PNAM))
	exit(EXIT_FAILURE);

    if (!proc_poly())
	exit(EXIT_FAILURE);

    if (!proc_fils())
	exit(EXIT_FAILURE);

    /* reads and checks the pedigree file for pedigree errors */

    if (!chk_pedfil())
        give_up(ERR_qit);	

    /* if the pedigree file was not in chronological order, */
    /* or if there are changes, writes a new one */

    if (!mak_opedfil())
        give_up(ERR_qit);	

    /* reopen the pedigree file for reading the data */

    if (!(pfile = open_ped()))
        give_up(ERR_qit);	

    alloc_ped();

    /* input the pedigree, and data observations, construct the */
    /* pedigree neighbourhoods, and sort the data effects */

    input_data(pfile, 0, dodo);
    fclose(pfile);

    free_ped();

    /* the no. of observed individuals is known; check against limit */

    if (!check_pars(np, max_cpsize))
	exit(EXIT_FAILURE);
    
    ginverse = malloc(pedsize * 4 * sizeof(Term));
    alloc_chk(WHERE, ginverse);

    /* read or find non-zero terms of g-inverse for all pedigree */
    /* individuals */

    gsize = henderson(ginverse);
    ginverse = realloc(ginverse, (gsize + 1) * sizeof(Term));
    alloc_chk(WHERE, ginverse);

    /* for testing; if G-inverse is to be printed, exit */

    if (ChkGnv)
    {   printf(" G-inverse: \n\n");
        for (i = 1; i <= gsize; ++i)
	{
            printf(" %5d %5d %5d", i, ginverse[i].r1, ginverse[i].r2);
            print_val(ginverse[i].value);
            printf("\n");
        }
        if (!ChkEig)
            exit(EXIT_SUCCESS);
        else
	    printf("\n");
    }

    /* find or read eigenvalues of G-inverse for OBSERVED individuals */

    /* cdex : component index     cpnghd : component nghd */

    if (ICh == 1)
    {
	cdex = malloc(pedsize * sizeof(int));
        alloc_chk(WHERE, cdex);
	cpnghd = malloc(max_cpsize * sizeof(Ind *));
        alloc_chk(WHERE, cpnghd);
	evgen1 = malloc((max_cpsize + 1) * sizeof(double));
        alloc_chk(WHERE, evgen1);
	evgen = malloc((np + 1) * sizeof(double));
        alloc_chk(WHERE, evgen);

	j = 1;
	for (k = 1; k < n_compnts + 1; k++)
	{
	    i = 0;
            cpnp = 0;                   /* no. NOT observed in component*/
	    peopl = cmplist[k].people;
	    while (peopl->next != NULL)
	    {
		peopl = peopl->next;
		cpnghd[i] = peopl->current;
		if (qdex[cpnghd[i]->index] == 0)
		    cpnp++;
		cdex[cpnghd[i]->index] = i;
		i++;
	    }
            if (i > cpnp)
 	        bigcheig(evgen1, i - cpnp, i, cpnghd, cdex);
            else if (i == cpnp)
                give_warn(ERR_noo, i, cpnghd[0]->self);

	    while (i > cpnp)
	    {
		evgen[j] = evgen1[i - cpnp];
		i--;
		j++;
	    }
	}
	free((void *) evgen1);
        evgen1 = NULL;
	free((void *) cdex);
        cdex = NULL;
	free((void *) cpnghd);
        cpnghd = NULL;
        printf(" Eigenvalues have been computed\n\n\n");
    } 
    else if (ICh == 0)
    {   
        evgen = peigen;
        printf(" Eigenvalues have been read from parameter file\n\n\n");
    }
    else if (ICh == -1)
    {
        if (!input_eigen(np))
            exit(EXIT_FAILURE);
        else
        {
            evgen = peigen;
            printf(" Eigenvalues have been read from eigenvalues"
                                                           " file\n\n\n");
        }
    }

    free_cmplists();
    free_nghds();

    /* For testing bigcheig and reduce routines;  print eigenvalues */
    /* and exit */

    if (ChkEig)
    {   
        printf(" eigenvalues of G-observed:\n\n");
        for (i = 1; i <= np; ++i)
        {
            print_val5(evgen[i]);
            if (i%10 == 0)
                printf("\n");  
        }
	printf("\n\n");
        exit(EXIT_SUCCESS);
    }

    ivgen = malloc((np + 1) * sizeof(double));
    alloc_chk(WHERE, ivgen);

    for (i = 1; i <= np; ++i)
	ivgen[i] = 1.0 / evgen[i];

    if (ChkTrc) 
    {
        cc = 0.0;
        for (i = 1; i <= np; ++i)
	    cc += log(evgen[i]);
        printf(" check log-determinant of G =");
        print_val(cc);

        cc = 0.0;
        for (i = 1; i <= np; ++i)
	    cc += ivgen[i];
        printf("\n check trace of G-inverse =  ");
        print_val(cc);
        printf("\n\n");
        fflush(stdout);
    }

    /* allocate space for vectors */

    alloc_vect(&a, &b, &epsi, &y);
   
    y[0] = 0.0;

    /* set initial values */

    setvals(b);

    /* now ready for EM iteration  */

    for (k = 0; k <= ItEM; ++k)
	em_iter(k, a, b, epsi, y, ginverse, gsize, evgen, ivgen);

    out_fixed(y, 0, 0);
    breeding_values(y, a, b, epsi, ginverse, gsize);
    lik_big(k, ivgen, y, a, b, ginverse, gsize);

    free_input_data();
    free_vect(a, b, epsi, y, ginverse, evgen, ivgen);

    return 0;
}
