#include "quant.h"

   
#define GG(I,J)      gg[(cpsize+1) * (I+1) + (J+1)]

/************************************************************************
   reduces the G-matrix for the observed individuals to a symmetric matrix
   with the same eigen values, returning the reduced matrix, the extracted
   eigenvalues, and "imm" the number of immediate eigenvalues;
   called from bigcheig() -- for further info see bigcheig() in cheigen.c 
	April 1992						
  
   Modified June 1996 by Chi-hong Tseng (ChT) to find eigenvalues by
	component, to enable large data sets of several pedigrees to
	be handled.
  
****************************************************************************/

int column_test(double *gg, int i, int j, int *move_to, int cpsize)
/*
 * to test identity of columns i and j of current matrix  
 * testing only non-diagonal terms, and non-blanked terms    
 */
{
    int l, test;
    test = 1;

    if (GG(i, i) != GG(j, j))
	return (0);

    for (l = 0; l < cpsize && test != 0; ++l)
    {
	if ((i != l) && (j != l) && (move_to[l] != 0))
	    if (GG(i, l) != GG(j, l))
		test = 0;
    }

    return (test);
}

int reduce(double *gg, double *evec, int *move_to, int *cdex, 
                                int cpsize, Ind ** cpnghd)
{
    int imm;
    int *done;
    Marlist *marriage_ptr;
    int i, j, l, spouse, mum, dad, sib;
    Indlist *head;
    int count;
    double a1, cc;

    done = malloc(cpsize * sizeof(int));
    alloc_chk(WHERE, done);
    imm = 0;
    for (i = 0; i < cpsize; ++i)
    {
	move_to[i] = i + 1;
	done[i] = 0;
    }

    /* first white out the rows/columns for unobserved individuals     */
    for (i = 0; i < cpsize; ++i)
	if (qdex[cpnghd[i]->index] == 0)
	    move_to[i] = 0;

    /* for each individual, is it in G-obs and still unprocessed anyhow? */
    for (i = 0; i < cpsize; ++i)
    {
	if ((done[i] == 0) && (qdex[cpnghd[i]->index] != 0))

	    /* for each individual; is it a founder?       	 */
	{
	    mum = cpnghd[i]->ma->index;
	    if (mum == INVALID_INDEX)
		/* if so test for reduction with spouses */
	    {
		marriage_ptr = cpnghd[i]->marriages;
		while (marriage_ptr != NULL)
		{
		    spouse = cdex[marriage_ptr->current->spouse->index];
		    j = move_to[spouse];
		    if (j != 0 && done[spouse] != 1)
		    {
			if (column_test(gg, i, spouse, move_to, cpsize) == 1)
			{
			    move_to[spouse] = 0;
			    done[spouse] = 1;
			    ++imm;
			    evec[imm] = GG(i, i) - GG(i, spouse);
			    GG(i, i) = GG(i, i) + GG(i, spouse);
			    for (l = 0; l < cpsize; ++l)
				if (l != i)
				{
				    GG(i, l) = GG(i, l) * sqrt(2.0);
				    GG(l, i) = GG(i, l);
				}
			    done[i] = 1;
			}
		    }
		    marriage_ptr = marriage_ptr->next;
		}

	    }
	    /* if not, test for reduction with parents' kids  */
	    else
	    {
		count = 0;
		a1 = -1.0;
		marriage_ptr = nghd[mum].marriages;
		while (marriage_ptr != NULL)
		{
		    head = marriage_ptr->current->kids;
		    while (head != NULL)
		    {
			sib = cdex[head->current->index];
			j = move_to[sib];
			if ((j != 0) && (sib > i) 
                                     && (column_test(gg, i, sib, move_to, cpsize)) == 1)
			{
			    if (count == 0)
				a1 = GG(i, sib);
			    if (GG(i, sib) == a1)
			    {
				move_to[sib] = 0;
				++count;
			    }
			}
			head = (head->next);
		    }
		    marriage_ptr = marriage_ptr->next;
		}

		dad = cpnghd[i]->pa->index;
		marriage_ptr = nghd[dad].marriages;
		while (marriage_ptr != NULL)
		{
		    head = marriage_ptr->current->kids;
		    while (head != NULL)
		    {
			sib = cdex[head->current->index];
			j = move_to[sib];
			if ((j != 0) && (sib > i) 
                                     && (column_test(gg, i, sib, move_to, cpsize)) == 1)
			{
			    if (count == 0)
				a1 = GG(i, sib);
			    if (GG(i, sib) == a1)
			    {
				move_to[sib] = 0;
				++count;
			    }
			}
			head = (head->next);
		    }
		    marriage_ptr = marriage_ptr->next;
		}

		if (count > 0)		/* number of sibs = count+1 */
		{
		    for (l = 1; l <= count; ++l)
		    {
			++imm;
			evec[imm] = GG(i, i) - a1;	/* a0	*/
		    }	
		    GG(i, i) = GG(i, i) + count * a1;	/* a0+(count+1) * a1 */
		    cc = count + 1.0;
		    for (l = 0; l < cpsize; ++l)
			if (l != i)
			{
			    GG(i, l) = GG(i, l) * sqrt(cc);
			    GG(l, i) = GG(i, l);
			}
		}
		done[i] = 1;
	    }
	}
    }
    free((void *) done);
    done = NULL;

    return (imm);
}
